// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ManifoldParameterization.hpp"

using namespace LORM;

template<typename T_scalar>
ManifoldParameterization<T_scalar>::ManifoldParameterization( 
  const std::shared_ptr<const Manifold> & sp_manifold ) :
  sp_manifold_(sp_manifold) 
{}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> ManifoldParameterization<T_scalar>::evalParameterization( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  assert( localPointCoords.rows() == sp_manifold_->getDimension() );
  return computeParameterization( localPointCoords );      
}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> ManifoldParameterization<T_scalar>::evalInverseParameterization(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const 
{
  assert( pointCoords.rows() == sp_manifold_->getAmbientSpaceDimension() );
  return computeInverseParameterization( pointCoords );
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> ManifoldParameterization<T_scalar>::evalJacobiMatrix(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  assert( localPointCoords.rows() == sp_manifold_->getDimension() );
  return computeJacobiMatrix( localPointCoords );
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> ManifoldParameterization<T_scalar>::evalInverseRiemannianMatrix(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  assert( localPointCoords.rows() == sp_manifold_->getDimension() ); 
  return computeInverseRiemannianMatrix( localPointCoords ); 
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> ManifoldParameterization<T_scalar>::evalChristoffelMatrixLinearCombination(	  
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  assert( localCoefficients.rows() == sp_manifold_->getDimension() );
  assert( localPointCoords.rows() == sp_manifold_->getDimension() );
  return computeChristoffelMatrixLinearCombination( localCoefficients, localPointCoords ); 
}

//Macro for explicit instantiations of ManifoldParameterization<T_scalar>
#define LORM_INSTANTIATE_MANIFOLDPARAMETERIZATION(T_scalar)	\
  template class ManifoldParameterization<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_MANIFOLDPARAMETERIZATION( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_MANIFOLDPARAMETERIZATION(double)
#endif

#undef LORM_INSTANTIATE_MANIFOLDPARAMETERIZATION
