// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _MANIFOLD_PARAMETERIZATION_HPP_
#define _MANIFOLD_PARAMETERIZATION_HPP_

#include "Manifold.hpp"
#include "ManifoldPointArray.hpp"

namespace LORM {

  /** \class ManifoldParameterization
   *
   *  \brief The abstract base class template for a parameterization of a
   *  manifold. 
   *
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   *
   *  \tparam T_scalar Numerical Type, e.g. float, double, long double
   *
   */

  template<typename T_scalar>
  class ManifoldParameterization {
  public:
    /// Constructor. One must provide a valid shared pointer to an instance of
    /// Manifold.
    explicit ManifoldParameterization( const std::shared_ptr<const Manifold> & sp_manifold );
    /// Destructor.
    virtual ~ManifoldParameterization() {};

    /// Returns the shared pointer to the manifold.
    const std::shared_ptr<const Manifold> & getManifold() const { return sp_manifold_; };
 
    /// Returns the Cartesian coordinates determined by the local Coordinates of
    /// the parameterization.
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> evalParameterization( 
              const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;

    /// Returns the local Coordinates of the parameterization determined by the
    /// Cartesian coordinates.
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> evalInverseParameterization(
	      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const;

    /// Returns the Jacobi matrix of the parameterization.
    Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> evalJacobiMatrix(
              const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;

    /// Returns the inverse of the Riemannian matrix at the point determined by
    /// its local coordinates.
    Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> evalInverseRiemannianMatrix(
	      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;
    
    /// Returns the linear combination of the Christoffel matrices determined by
    /// \c localCoefficients, where the point is given in local coordinates.
    Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> evalChristoffelMatrixLinearCombination(
      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,	  
      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;

  protected:
    /// Must implement the computations for evalParameterization(). 
    virtual Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeParameterization( 
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const =0;
    
    /// Must implement the computations for evalInverseParameterization().     
    virtual Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeInverseParameterization(                 
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const =0;
    
    /// Must implement the computations for evalJacobiMatrix(). 
    virtual Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeJacobiMatrix( 	         
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const =0;

    /// Must implement the computations for evalInverseRiemannianMatrix(). 
    virtual Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeInverseRiemannianMatrix( 	
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const =0;

    /// Must implement the computations for evalChristoffelMatrixLinearCombination().
    virtual Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination(
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
		const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const =0;

  private:
    /// \note Non-copyable.
    ManifoldParameterization( const ManifoldParameterization<T_scalar> & ) {};
    /// \note Non-copy-assignable.
    const ManifoldParameterization<T_scalar> & operator=( const ManifoldParameterization<T_scalar> & )
    { return *this; };
    
    const std::shared_ptr<const Manifold> sp_manifold_;
  };
  

  template<typename T_scalar>
  void computeManifoldPointArrayFromLocalCoordinatesArray(
    ManifoldPointArray<T_scalar> * p_pointArray, 
    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray,
    const ManifoldParameterization<T_scalar> & parameterization) 
  {
    auto & sp_manifold = parameterization.getManifold();
    assert( p_pointArray->getManifold().get() == sp_manifold.get() );
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> pointCoordsArray;
    pointCoordsArray.resize( sp_manifold->getAmbientSpaceDimension(), localPointCoordsArray.cols() );
    for( int i=0; i<localPointCoordsArray.cols(); ++i) 
      pointCoordsArray.col(i) =  
	parameterization.evalParameterization( localPointCoordsArray.col(i) );
    p_pointArray->setCoordinates( pointCoordsArray );
  }
  
  template<typename T_scalar>
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>
  computeLocalCoordinatesArrayFromPointCoordinatesArray( 
    const  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray,
    const ManifoldParameterization<T_scalar> & parameterization ) 
  {
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> localPointCoordsArray;
    localPointCoordsArray.resize( parameterization.getManifold()->getDimension(), pointCoordsArray.cols() );
    for( int i=0; i<pointCoordsArray.cols(); ++i ) 
      localPointCoordsArray.col(i) = 
	parameterization.evalInverseParameterization( pointCoordsArray.col(i) );
    return localPointCoordsArray;
  }

  template<typename T_scalar>
  void computeLocalCoordinatesArrayFromManifoldPointArray( 
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_localPointCoordsArray,
    const ManifoldPointArray<T_scalar> & pointArray,
    const ManifoldParameterization<T_scalar> & parameterization  ) 
  {
    auto & sp_manifold = parameterization.getManifold();
    assert( pointArray.getManifold().get() == sp_manifold.get() );
    computeLocalCoordinatesArrayFromPointCoordinates( p_localPointCoordsArray, pointArray.getCoordinates() );
  }
  
  template<typename T_scalar>
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeTangentCoordinatesFromLocalCoordinates( 
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localTangentCoords,
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords,
    const ManifoldParameterization<T_scalar> & parameterization ) 
  {
    // v = D_h * v_h
    auto jacobiMatrix = parameterization.evalJacobiMatrix( localPointCoords );
    return jacobiMatrix * localTangentCoords.matrix();
  }
  
  template<typename T_scalar>
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeLocalCoordinatesFromTangentCoordinates( 
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords,
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords,
    const ManifoldParameterization<T_scalar> & parameterization ) 
  {
    // v_h = G_h^(-1) * D_h^T * v
    auto jacobiMatrix = parameterization.evalJacobiMatrix( localPointCoords );
    auto invRiemannMatrix = parameterization.evalInverseRiemannianMatrix( localPointCoords );
    return invRiemannMatrix * jacobiMatrix.transpose() * tangentCoords.matrix();
  }
  
} // end namespace LORM

#endif // _MANIFOLD_PARAMETERIZATION_HPP_
