// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _MANIFOLD_POINT_ARRAY_HPP_
#define _MANIFOLD_POINT_ARRAY_HPP_

#include "Manifold.hpp"
#include "../Utils/SignalOnChange.hpp"

namespace LORM {

  //forward declaration
  template<typename T_scalar>
  class ManifoldTangentVectorArray;

  /** \class ManifoldPointArray
   *
   *  \brief The abstract base class template for an array of points on a Riemannian manifold.
   *
   *  \tparam T_scalar Numerical Type, e.g. float, double, long double
   *
   */

  template<typename T_scalar>
  class ManifoldPointArray : public SignalOnChange {
    friend class ManifoldTangentVectorArray<T_scalar>;
  public:
    /// Constructor. One must provide a valid shared pointer to an instance of Manifold.
    explicit ManifoldPointArray( const std::shared_ptr<const Manifold> & sp_manifold );
    
    /// Destructor.
    virtual ~ManifoldPointArray() {};
    
    /// Copy-Constructor.
    ManifoldPointArray( const ManifoldPointArray<T_scalar> & other );
    
    /// Assignment-Operator. \note Assignment is only allowed if the shared
    /// pointers to the corresponding Manifold instances are equal.
    const ManifoldPointArray<T_scalar> & operator= ( const ManifoldPointArray<T_scalar> & rhs );
    
    /// Returns the shared pointer of the manifold.
    const std::shared_ptr<const Manifold> & getManifold() const { return sp_manifold_;};
    
    /// Returns the array of the Cartesian coordinates of the points.
    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  & getCoordinates() const { return coordinates_; };
    
    /// Sets the array of the Cartesian coordinates of the point. The point \c
    /// coordinates are  provided by an array 
    /// (rows == ambient space dimension of the manifold, 
    ///  columns == number of points) 
    void setCoordinates( const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & coordinates);
    
    /// Must implement the creation of a new ManifoldPointArray returned by a shared
    /// pointer, which is copy-constructed by the actual ManifoldPointArray.
    virtual std::shared_ptr<ManifoldPointArray<T_scalar> > clonePointArray () const =0;
    
    /// Must implement the creation of a new ManifoldTangentVectorArray returned by
    /// a shared pointer with ManifoldPointArray as its base point. The shared
    /// pointer is \c NULL if no tangentVectorArray can be created, which should
    /// be the case if there exists no tangent vector implementation for derived
    /// manifold points. \note To use this function properly, the actual
    /// instance of ManifoldPointArray must be created as a shared pointer.
    virtual std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > createTangentVectorArray () const  =0;
    
  protected:
    /// Must implement the projection of a point given in Cartesian coordinates on to the manifold.
    virtual void projectPointOnManifold( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coordinates ) =0;
    
  private:
    const std::shared_ptr<const Manifold> sp_manifold_ ;
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> coordinates_;
  };
  
  /// The standard implementation returns a \c NULL pointer.
  template <typename T_scalar>
  std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > 
  ManifoldPointArray<T_scalar>::createTangentVectorArray() const { 
    return std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > (); 
  }  
  
} // end namespace LORM

#endif // _MANIFOLD_POINT_ARRAY_HPP_
