// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ManifoldTangentVectorArray.hpp"

using namespace LORM;

template<typename T_scalar>
ManifoldTangentVectorArray<T_scalar>::ManifoldTangentVectorArray(
  const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_basePointArray ) :
  SignalOnChange(),
  sp_basePointArray_ (sp_basePointArray), 
  coordinates_ () 
{
  resetCoordinates();
  setConnectionToSignalBasePointArrayChanged();
}


template<typename T_scalar>
ManifoldTangentVectorArray<T_scalar>::ManifoldTangentVectorArray(
  const ManifoldTangentVectorArray<T_scalar> & other ) :
  SignalOnChange ( other ),
  sp_basePointArray_ ( other.sp_basePointArray_ ),
  coordinates_ ( other.coordinates_ ) 
{
  setConnectionToSignalBasePointArrayChanged();
}

template<typename T_scalar>
const ManifoldTangentVectorArray<T_scalar> & ManifoldTangentVectorArray<T_scalar>::operator=(
  const ManifoldTangentVectorArray<T_scalar> & rhs ) 
{
  if (this != &rhs) {
    assert( sp_basePointArray_ == rhs.sp_basePointArray_ ); // for now, we do not allow to assign different base points for the same manifold
    SignalOnChange::operator=(rhs);
    coordinates_ = rhs.coordinates_;
    SignalOnChange::sendSignal();
  }
  return *this;
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::setCoordinates(
  const Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> & coordinates ) 
{
  assert( coordinates_.rows() == sp_basePointArray_->getCoordinates().rows() && 
	  coordinates_.cols() == sp_basePointArray_->getCoordinates().cols() );
  for( int i=0; i<coordinates_.cols(); ++i ) {
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> temp = coordinates.col(i);
    projectVectorOnTangentSpace( &temp, sp_basePointArray_->getCoordinates().col(i) );
    coordinates_.col(i) = temp;
  }
  SignalOnChange::sendSignal(); 
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::setNegative(
  const ManifoldTangentVectorArray<T_scalar> & other ) 
{
  assert( sp_basePointArray_ == other.sp_basePointArray_ );
  coordinates_ = - other.coordinates_;
  SignalOnChange::sendSignal(); 
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::setAlphaTimesXPlusBetaTimesY( 
  T_scalar alpha, const ManifoldTangentVectorArray<T_scalar> & x,
  T_scalar beta, const ManifoldTangentVectorArray<T_scalar> & y) 
{
  assert( sp_basePointArray_ == x.sp_basePointArray_ );
  assert( sp_basePointArray_ == y.sp_basePointArray_ );
  coordinates_ = alpha * x.coordinates_ + beta * y.coordinates_;
  SignalOnChange::sendSignal(); 
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::applyGeodesicStep ( 
  ManifoldPointArray<T_scalar> * p_basePointArray, 
  Eigen::Array<T_scalar, 1, Eigen::Dynamic> stepLengthArray )
{
  assert( p_basePointArray == sp_basePointArray_.get() );	
  assert( coordinates_.cols() == stepLengthArray.cols() );
  for( int i=0; i<coordinates_.cols(); ++i) {
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> point = p_basePointArray->coordinates_.col(i);
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> tangent = coordinates_.col(i);
    computeGeodesicStep( &point , &tangent , stepLengthArray(i));
    p_basePointArray->projectPointOnManifold( &point );
    projectVectorOnTangentSpace( &tangent, point );
    p_basePointArray->coordinates_.col(i) = point;
    coordinates_.col(i) = tangent;	
  }
  boost::signals2::shared_connection_block block(connectionBasePointArrayChanged_);
  p_basePointArray->SignalOnChange::sendSignal();
  SignalOnChange::sendSignal();	
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::applyGeodesicStep(
  ManifoldPointArray<T_scalar> * p_basePointArray, T_scalar stepLength )
{
  applyGeodesicStep( p_basePointArray, stepLength * Eigen::Array<T_scalar, 1, Eigen::Dynamic> ::
		     Ones(coordinates_.cols()) );
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::applyChristoffelMatrixLinearCombinationMultiplication(
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & coefficientsArray)
{
  assert( coefficientsArray.rows() == coordinates_.rows() &&
	  coefficientsArray.cols() == coordinates_.cols() );
  for( int i=0; i<coordinates_.cols(); ++i) {
    auto christoffelMatrix = 
      computeChristoffelMatrixLinearCombination( coefficientsArray.col(i), 
						 sp_basePointArray_->getCoordinates().col(i) );
    coordinates_.col(i) = (christoffelMatrix * coordinates_.col(i).matrix()).array();
  }
}


template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> ManifoldTangentVectorArray<T_scalar>::computeNormalVectorCoordinatesArray () const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> normalCoordsArray;
  normalCoordsArray.resize( coordinates_.rows(), coordinates_.cols() );
  for (int i=0; i<coordinates_.cols(); ++i) 
    normalCoordsArray.col(i) = computeNormalVectorCoordinates( sp_basePointArray_->getCoordinates().col(i), coordinates_.col(i));
  return normalCoordsArray;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> ManifoldTangentVectorArray<T_scalar>::computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const 
{
  assert( coefficients.rows() == basePointCoords.rows() );
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
  christoffelMatrix.resize( basePointCoords.rows(), basePointCoords.rows() );
  for (int i=0; i<basePointCoords.rows(); ++i) {
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> e_i = 
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> :: Zero(basePointCoords.rows());
    e_i(i) = 1;
    projectVectorOnTangentSpace( &e_i, basePointCoords);
    auto normal = computeNormalVectorCoordinates(basePointCoords, e_i);
    christoffelMatrix(i,i) = -(coefficients*normal).sum();
    for (int j=i+1; j<coordinates_.rows(); ++j) {
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> e_j = 
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> :: Zero(basePointCoords.rows());
      e_j(j) = 1;
      projectVectorOnTangentSpace( &e_j, basePointCoords);
      
      normal = computeNormalVectorCoordinates(basePointCoords, e_i + e_j);
      christoffelMatrix(i,j) = -(coefficients*normal).sum();
      normal = computeNormalVectorCoordinates(basePointCoords, e_i - e_j);
      christoffelMatrix(i,j) += (coefficients*normal).sum();
      christoffelMatrix(i,j) /= (T_scalar) 4;
      christoffelMatrix(j,i) = christoffelMatrix(i,j);
    }
  }
  return christoffelMatrix;
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::setConnectionToSignalBasePointArrayChanged ()
{
  connectionBasePointArrayChanged_ = sp_basePointArray_->
    connectSignalChanged( boost::bind( & ManifoldTangentVectorArray<T_scalar>::resetCoordinates, this) ); 
}

template<typename T_scalar>
void ManifoldTangentVectorArray<T_scalar>::resetCoordinates()
{ 
  coordinates_ =  0 / Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>::
    Zero( sp_basePointArray_->getCoordinates().rows(), sp_basePointArray_->getCoordinates().cols() ); 
  SignalOnChange::sendSignal(); 
}

//Macro for explicit instantiations of ManifoldTangentVectorArray<T_scalar>
#define LORM_INSTANTIATE_MANIFOLDTANGENTVECTORARRAY(T_scalar)	\
  template class ManifoldTangentVectorArray<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_MANIFOLDTANGENTVECTORARRAY( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_MANIFOLDTANGENTVECTORARRAY(double)
#endif

#undef LORM_INSTANTIATE_MANIFOLDTANGENTVECTORARRAY
