// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _MANIFOLD_TANGENT_VECTOR_ARRAY_HPP_
#define _MANIFOLD_TANGENT_VECTOR_ARRAY_HPP_

#include "ManifoldPointArray.hpp"
#include "../Utils/SignalOnChange.hpp"

namespace LORM {

  /** \class ManifoldTangentVectorArray
   *
   *  \brief The abstract base class template for an array of tangent vectors on
   *  a Riemannian manifold.
   *
   *  \tparam T_scalar Numerical Type, e.g. float, double, long double
   *
   */

  template<typename T_scalar>
  class ManifoldTangentVectorArray : public SignalOnChange {
  public:
    /// Constructor. One must provide a valid shared pointer to an instance of ManifoldPointArray.
    explicit ManifoldTangentVectorArray ( const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_basePointArray );
    /// Destructor.
    virtual ~ManifoldTangentVectorArray() {};
    
    /// Copy-Constructor.
    ManifoldTangentVectorArray ( const ManifoldTangentVectorArray<T_scalar> & other );
    
    /// Assignment-Operator. \note Assignment is only allowed if the shared
    /// pointers to the corresponding ManifoldPointArray instances are equal.
    const ManifoldTangentVectorArray<T_scalar> & operator=( const ManifoldTangentVectorArray<T_scalar> & rhs );
    
    /// Returns the shared pointer to the base point array.
    const std::shared_ptr<const ManifoldPointArray<T_scalar> > & getBasePointArray () const { 
      return sp_basePointArray_; };
    
    /// Returns the array of the Cartesian coordinates of the tangent vectors.
    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & getCoordinates() const { 
      return coordinates_;  
    };
    
    /// Sets the array of the Cartesian coordinates of the tangent vector. The
    /// tangent \c coordinates are provided by an array
    ///  (rows == ambient space dimension of the manifold, 
    ///   columns == number of points)
    void setCoordinates( const Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> & coordinates );
    
    /// Sets the coordinates to the negatives of the one provided by \c other.
    /// \note As for the assignment operator, the shared pointers to the
    /// corresponding ManifoldPointArray instances must be equal.
    void setNegative( const ManifoldTangentVectorArray<T_scalar> & other );
   
    /// Sets the coordinates to the linear combination of \c alpha * \c x + \c
    /// beta * \c y. \note As for the assignment operator, the shared pointers
    /// to the corresponding ManifoldPointArray instances must be equal.
    void setAlphaTimesXPlusBetaTimesY( T_scalar alpha, const ManifoldTangentVectorArray<T_scalar> & x,
				       T_scalar beta, const ManifoldTangentVectorArray<T_scalar> & y);

    /// Applies geodesic steps of length \c stepLengthArray starting at  \c
    /// basePointArray in directions determined by the tangent vectors. The
    /// pointer \c p_basePointArray must be identically to the one returned by
    /// getBasePointArray().get(), and the stepLengthArray must have the same size.
    void applyGeodesicStep ( ManifoldPointArray<T_scalar> * p_basePointArray, 
			     Eigen::Array<T_scalar, 1, Eigen::Dynamic> stepLengthArray );

    /// Applies geodesic steps of length \c stepLength starting at \c
    /// basePointArray in directions determined by the tangent vectors. The
    /// pointer \c p_basePointArray must be identically to the one returned by
    /// getBasePointArray().
    void applyGeodesicStep ( ManifoldPointArray<T_scalar> * p_basePointArray, 
			     T_scalar stepLength );

    /// Applies the multiplication of the tangent vectors with the linear
    /// combination of the Christoffel matrices determined by \c coefficientsArray.
    void applyChristoffelMatrixLinearCombinationMultiplication(
	  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & coefficientsArray); 

    /// Returns the coordinates array of the normals determined by the points
    /// and the tangent vectors.
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeNormalVectorCoordinatesArray() const;

  protected:
    /// Must implement the projection in Cartesian coordinates of a single
    /// vector \c tangentCoords on the tangent space determined by \c
    /// basePointCoords.
    virtual void projectVectorOnTangentSpace(Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
 				      const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const =0;

    /// Must implement the computation of a single geodesic step of length \c
    /// stepLength starting at \c basePointCoords in direction \c
    /// tangentVectorCoords given in Cartesian coordinates.
    virtual void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
   				      Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentVectorCoords,
   				      T_scalar stepLength ) const =0;

    /// Must implement the computation of the Cartesian coordinates of a single
    /// normal determined by \c basePointCoords and \c tangentCoords.
    virtual Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
   			  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
   			  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const =0;

    /// Must implement the computation of the linear combination of the
    /// christoffel matrices for given \c coefficients at \c basePointCoords.
    /// For implementing this function, one can use the helper member function
    /// computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates().
    virtual Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>
      computeChristoffelMatrixLinearCombination(
         const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const=0;

    /// Helper function, which computes the linear combination of the
    /// Christoffel matrices determined by \c coefficients, where the base point
    /// is given in Cartesian coordinates. \note This routine might be rather
    /// slow, since the computation needs O(\c ambientSpaceDimension^2) calls to
    /// computeNormalVectorCoordinates().
    Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> 
    computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
    	  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
	  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const;

  private:     
    void setConnectionToSignalBasePointArrayChanged ();
    void resetCoordinates(); 

    const std::shared_ptr<const ManifoldPointArray<T_scalar> > sp_basePointArray_;
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> coordinates_;
    boost::signals2::scoped_connection connectionBasePointArrayChanged_; // connection will be destroyed if the object is destroyed
  };

  /// Computes the dot product between \c tangentVectorArray1 and \c
  /// tangentVectorArray2, where the arrays are considered as a single vector.
  /// \note As for the assignment operator, the shared pointers to the
  /// corresponding ManifoldPointArray instances must be equal.
  template<typename T_scalar>
  T_scalar computeDotProduct( const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray1,
			      const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray2 ) {
    assert( tangentVectorArray1.getBasePointArray() == tangentVectorArray2.getBasePointArray());
    return (tangentVectorArray1.getCoordinates() * tangentVectorArray2.getCoordinates()).sum();
  }

  /// Computes the dot product between \c tangentVectorArray1 and \c
  /// tangentVectorArray2 of every single tangent vector of the array.
  /// \note As for the assignment operator, the shared pointers to the
  /// corresponding ManifoldPointArray instances must be equal.
  template<typename T_scalar>
  Eigen::Array<T_scalar, 1, Eigen::Dynamic> computeDotProductVectorized( const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray1,
			      const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray2 ) {
    assert( tangentVectorArray1.getBasePointArray() == tangentVectorArray2.getBasePointArray());
    return (tangentVectorArray1.getCoordinates() * tangentVectorArray2.getCoordinates()).colwise().sum();
  }
  
  /// Computes the norm of \c tangentVectorArray, where the array is considered
  /// as a single vector.
  template<typename T_scalar>
  T_scalar computeNorm( const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray ) {
    return tangentVectorArray.getCoordinates().matrix().norm();
  }

  /// Computes the norm of every single tangent vector in \c tangentVectorArray.
  template<typename T_scalar>
  Eigen::Array<T_scalar, 1, Eigen::Dynamic> computeNormVectorized( const ManifoldTangentVectorArray<T_scalar> & tangentVectorArray ) {
    return tangentVectorArray.getCoordinates().matrix().colwise().norm();
  }
    
} // end namespace LORM

#endif // _MANIFOLD_TANGENT_VECTOR_ARRAY_HPP_
