// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "EulerAngles.hpp"

using namespace LORM::RotationGroup;

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1>  EulerAngles<T_scalar>::computeParameterization( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointCoords;
  pointCoords.resize(9); 
  Eigen::Map<Eigen::Matrix<T_scalar, 3, 3> > matrix( pointCoords.data(), 3, 3 );
  const T_scalar & phi1 = eulerAngles(0);
  const T_scalar & theta = eulerAngles(1);
  const T_scalar & phi2 = eulerAngles(2);
  matrix(0,0) =    cos(phi1)  * cos(theta) * cos(phi2) - sin(phi1) * sin(phi2);
  matrix(1,0) =    sin(phi1)  * cos(theta) * cos(phi2) + cos(phi1) * sin(phi2);
  matrix(2,0) =  - sin(theta) * cos(phi2);
  matrix(0,1) =  - cos(phi1)  * cos(theta) * sin(phi2) - sin(phi1)  * cos(phi2);
  matrix(1,1) =  - sin(phi1)  * cos(theta) * sin(phi2) + cos(phi1)  * cos(phi2);
  matrix(2,1) =    sin(theta) * sin(phi2);
  matrix(0,2) =    cos(phi1)  * sin(theta);
  matrix(1,2) =    sin(phi1)  * sin(theta);
  matrix(2,2) =    cos(theta);      
  return pointCoords;
}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1>  EulerAngles<T_scalar>::computeInverseParameterization(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const
{
  Eigen::Array<T_scalar, 3, 1> eulerAngles;
  T_scalar & phi1 = eulerAngles(0);
  T_scalar & theta = eulerAngles(1);
  T_scalar & phi2 = eulerAngles(2);
  Eigen::Map<const Eigen::Matrix<T_scalar, 3, 3> > matrix( pointCoords.data(), 3, 3 );
  // theta = arccos( m33 );
  theta = acos( matrix(2,2) );
  // phi1 = 2 arctan( m23/ ( m13-sqrt(m13^2+m23^2)) ) + pi;
  // phi2 = 2 arctan( m32/ (-m31-sqrt(m31^2+m32^2)) ) + pi;
  phi1 = matrix(1,2) / (   matrix(0,2) - sqrt( matrix(0,2)*matrix(0,2) + matrix(1,2)*matrix(1,2) ) );
  phi2 = matrix(2,1) / ( - matrix(2,0) - sqrt( matrix(2,0)*matrix(2,0) + matrix(2,1)*matrix(2,1) ) );
  phi1 = 2 * atan( phi1 ) + M_PI;
  phi2 = 2 * atan( phi2 ) + M_PI;
  if( matrix(1,2) == 0 && matrix(2,1) == 0 && matrix(0,2) == 0 && matrix(2,0) == 0 ) {
    phi2 = 0;
    phi1 = acos( matrix(1,1) );
    return eulerAngles;
  }
  if( matrix(1,2) == 0 && 0 <= matrix(0,2) )
    phi1 = 0;
  if( matrix(2,1) == 0 && matrix(2,0) <= 0 )
    phi2 = 0;
  return eulerAngles;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> EulerAngles<T_scalar>::computeJacobiMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> jacobiMatrix;
  jacobiMatrix.resize(9,3);
  const T_scalar & phi1 = eulerAngles(0);
  const T_scalar & theta = eulerAngles(1);
  const T_scalar & phi2 = eulerAngles(2);
  // dphi1
  Eigen::Map<Eigen::Matrix<T_scalar, 3, 3> > 
    matrix_dphi1( jacobiMatrix.col(0).data(), 3, 3 );       
  matrix_dphi1(0,0) =  - sin(phi1)  * cos(theta) * cos(phi2) - cos(phi1) * sin(phi2);
  matrix_dphi1(1,0) =    cos(phi1)  * cos(theta) * cos(phi2) - sin(phi1) * sin(phi2);
  matrix_dphi1(2,0) =    0;
  matrix_dphi1(0,1) =    sin(phi1)  * cos(theta) * sin(phi2) - cos(phi1)  * cos(phi2);
  matrix_dphi1(1,1) =  - cos(phi1)  * cos(theta) * sin(phi2) - sin(phi1)  * cos(phi2);
  matrix_dphi1(2,1) =    0;
  matrix_dphi1(0,2) =    - sin(phi1)  * sin(theta);
  matrix_dphi1(1,2) =    cos(phi1)  * sin(theta);
  matrix_dphi1(2,2) =    0;      
  // dtheta
  Eigen::Map<Eigen::Matrix<T_scalar, 3, 3> > 
    matrix_dtheta( jacobiMatrix.col(1).data(), 3, 3 );       
  matrix_dtheta(0,0) =  - cos(phi1)  * sin(theta) * cos(phi2);
  matrix_dtheta(1,0) =  - sin(phi1)  * sin(theta) * cos(phi2);
  matrix_dtheta(2,0) =  - cos(theta) * cos(phi2);
  matrix_dtheta(0,1) =    cos(phi1)  * sin(theta) * sin(phi2);
  matrix_dtheta(1,1) =    sin(phi1)  * sin(theta) * sin(phi2);
  matrix_dtheta(2,1) =    cos(theta) * sin(phi2);
  matrix_dtheta(0,2) =    cos(phi1)  * cos(theta);
  matrix_dtheta(1,2) =    sin(phi1)  * cos(theta);
  matrix_dtheta(2,2) =  - sin(theta);      
  // dphi2
  Eigen::Map<Eigen::Matrix<T_scalar, 3, 3> > 
    matrix_dphi2( jacobiMatrix.col(2).data(), 3, 3 );       
  matrix_dphi2(0,0) =  - cos(phi1)  * cos(theta) * sin(phi2) - sin(phi1) * cos(phi2);
  matrix_dphi2(1,0) =  - sin(phi1)  * cos(theta) * sin(phi2) + cos(phi1) * cos(phi2);
  matrix_dphi2(2,0) =    sin(theta) * sin(phi2);
  matrix_dphi2(0,1) =  - cos(phi1)  * cos(theta) * cos(phi2) + sin(phi1) * sin(phi2);
  matrix_dphi2(1,1) =  - sin(phi1)  * cos(theta) * cos(phi2) - cos(phi1) * sin(phi2);
  matrix_dphi2(2,1) =    sin(theta) * cos(phi2);
  matrix_dphi2(0,2) =    0;
  matrix_dphi2(1,2) =    0;
  matrix_dphi2(2,2) =    0;      
  return jacobiMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> EulerAngles<T_scalar>::computeInverseRiemannianMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> invRiemannMatrix;
  invRiemannMatrix.resize(3,3);
  const T_scalar & theta = eulerAngles(1);
  // phi1 column
  invRiemannMatrix(0,0) = 1./2./(sin(theta)*sin(theta));
  invRiemannMatrix(1,0) = 0;
  invRiemannMatrix(2,0) = - cos(theta)/2./(sin(theta)*sin(theta));
  // theta column
  invRiemannMatrix(0,1) = 0;
  invRiemannMatrix(1,1) = 1./2.;
  invRiemannMatrix(2,1) = 0.;
  // phi2 column
  invRiemannMatrix(0,2) = - cos(theta)/2./(sin(theta)*sin(theta));
  invRiemannMatrix(1,2) = 0;
  invRiemannMatrix(2,2) = 1./2./(sin(theta)*sin(theta));
  return invRiemannMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> EulerAngles<T_scalar>:: computeChristoffelMatrixLinearCombination(  
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
  christoffelMatrix.resize(3,3);
  const T_scalar & coef_phi1 = localCoefficients(0);
  const T_scalar & coef_theta = localCoefficients(1);
  const T_scalar & coef_phi2 = localCoefficients(2);
  const T_scalar & theta = eulerAngles(1);
  // phi1 column
  christoffelMatrix(0,0) = 0;
  christoffelMatrix(1,0) = 1./2. * (  coef_phi1 * cos(theta)/sin(theta)
				      - coef_phi2 * 1./sin(theta) );
  christoffelMatrix(2,0) = 1./2. * (  coef_theta * sin(theta) );
  // theta column
  christoffelMatrix(0,1) = 1./2. * (  coef_phi1 * cos(theta)/sin(theta)
				      - coef_phi2 * 1./sin(theta) );
  christoffelMatrix(1,1) = 0;
  christoffelMatrix(2,1) = 1./2. * (  - coef_phi1 * 1./sin(theta)
				      + coef_phi2 * cos(theta)/sin(theta) );
  // phi2 column
  christoffelMatrix(0,2) = 1./2. * (  coef_theta * sin(theta) );
  christoffelMatrix(1,2) = 1./2. * (  - coef_phi1 * 1./sin(theta)
				      + coef_phi2 * cos(theta)/sin(theta) );
  christoffelMatrix(2,2) = 0;     
  return christoffelMatrix;
}

//Macro for explicit instantiations of EulerAngles<T_scalar>
#define LORM_INSTANTIATE_EULERANGLES(T_scalar)	\
  template class EulerAngles<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_EULERANGLES( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_EULERANGLES(double)
#endif

#undef LORM_INSTANTIATE_EULERANGLES
