// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _EULER_ANGELS_COORDINATES_HPP_
#define _EULER_ANGELS_COORDINATES_HPP_

#include "RotationGroup.hpp"
#include "../ManifoldParameterization.hpp"

namespace LORM {

  namespace RotationGroup {

    /** \class EulerAngles
     *
     *  \brief Implements the parameterization of the rotation group SO(3) by
     *  Euler angles (phi1, theta, phi2) in [0,2*pi)x[0,pi]x[0,2*pi) as
     *  ManifoldParameterization.
     *
     **/
    
    template<typename T_scalar>
    class EulerAngles : 
      public ManifoldParameterization<T_scalar> {
    public:
      /// Default-Constructor. An instance of a 3-dimensional RotationGroup is
      /// created.
      explicit EulerAngles() :
	ManifoldParameterization<T_scalar> ( std::make_shared<const RotationGroup> (3) ) {};
      
      /// Constructor. One must provide a valid shared pointer to an instance of
      /// RotationGroup with dimension = 3. 
      explicit EulerAngles(const std::shared_ptr<const RotationGroup> & sp_rotationGroup) :
	ManifoldParameterization<T_scalar> ( sp_rotationGroup ) {
	assert( sp_rotationGroup->getDimension() == 3); };
      
      /// Returns the shared pointer to RotationGroup, which is the underlying manifold.
      const std::shared_ptr<const RotationGroup> getRotationGroup() const { 
	return std::static_pointer_cast<const RotationGroup> (this->getManifold()); };
      
    private:
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeParameterization( 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const;		          
      
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeInverseParameterization(              
         const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeJacobiMatrix( 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeInverseRiemannianMatrix( 
         const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const;  
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination(  
         const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & eulerAngles ) const;
      
    };
    
  } // end namespace RotationGroup

} // end namespace LORM

#endif // _EULER_ANGLES_COORDINATES_HPP_
