// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "RotationGroup.hpp"
#include <Eigen/QR> // for projectionOperation
#include <unsupported/Eigen/MatrixFunctions>  // for exponential map

namespace LORM {
  namespace RotationGroup {
    namespace VectorOperations {

      template<typename T_scalar>
      void projectPointOnManifold ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ) 
      {
	int d = round( sqrt(p_coords->rows()) );
	assert( p_coords->rows() == d*d );
	Eigen::JacobiSVD<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  svd (d,d, Eigen::ComputeFullU | Eigen::ComputeFullV );
	Eigen::Map<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map( p_coords->data(), d, d);
	svd.compute( map );
	map = svd.matrixU() * svd.matrixV().transpose(); // all the singular values are set to one
	if ( map.determinant() < 0 ) // swap the first two columns to get positive determinant
	  map.col(0).swap(map.col(1));       
      }
      
      template<typename T_scalar>
      void projectVectorOnTangentSpace(  
        Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ,
	const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords )  
      {
	assert( p_coords->rows() == basePointCoords.rows() );
	int d = round( sqrt(p_coords->rows()) );
	assert( p_coords->rows() == d*d );
	Eigen::Map<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_t( p_coords->data(), d, d);
	Eigen::Map<const Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_p( basePointCoords.data(), d, d);
	map_t = (map_t - map_p * map_t.transpose() * map_p).eval()/2;
      }
      
      template<typename T_scalar>
      void computeGeodesicStep ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords ,
				 Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,
				 T_scalar stepLength) 
      {
	assert( p_basePointCoords->rows() == p_tangentVectorCoords->rows() );
	int d = round( sqrt(p_basePointCoords->rows()) );
	assert( p_basePointCoords->rows() == d*d );
	Eigen::Map<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_p( p_basePointCoords->data(), d, d);
	Eigen::Map<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_t( p_tangentVectorCoords->data(), d, d);
	Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> exp_apt = 
	  (stepLength * map_p.transpose() * map_t).exp();
	map_p *= exp_apt;
	map_t *= exp_apt;
      }
      
      template<typename T_scalar>
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeNormalVectorCoordinates ( 
         const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentVectorCoords) 
      {
	assert( basePointCoords.rows() == tangentVectorCoords.rows() );
	int d = round( sqrt(basePointCoords.rows()) );
	assert( basePointCoords.rows() == d*d );
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> normal;
	normal.setZero( d*d );
	Eigen::Map<const Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_p( basePointCoords.data(), d, d);
	Eigen::Map<const Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > 
	  map_t( tangentVectorCoords.data(), d, d);
	Eigen::Map<Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> >
	  map_n( normal.data(), d, d);
	map_n = map_t * map_p.transpose() * map_t;
	return normal;
      }

   
    } // end namespace VectorOperations
  } // end namespace RotationGroup
} // end namespace LORM

      
//Macro for explicit instantiations
#define LORM_INSTANTIATE_ROTATIONGROUP(T_scalar)			\
  template void LORM::RotationGroup::VectorOperations::projectPointOnManifold<T_scalar>( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ); \
  template void LORM::RotationGroup::VectorOperations::projectVectorOnTangentSpace<T_scalar>( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords , \
    const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords ); \
  template void LORM::RotationGroup::VectorOperations::computeGeodesicStep<T_scalar> ( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords , \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,T_scalar stepLength); \
  template Eigen::Array<T_scalar, Eigen::Dynamic, 1>  \
     LORM::RotationGroup::VectorOperations::computeNormalVectorCoordinates<T_scalar> ( \
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords, \
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentVectorCoords);
  
#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_ROTATIONGROUP( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_ROTATIONGROUP(double)
#endif

#undef LORM_INSTANTIATE_ROTATIONGROUP

      
