// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _ROTATION_GROUP_HPP_
#define _ROTATION_GROUP_HPP_

#include "../Manifold.hpp"

namespace LORM {

  /** \namespace LORM::RotationGroup
   *
   *  \brief Namespace consisting of classes and templates related to the rotation group SO(n).
   *
   **/

  namespace RotationGroup {
    
    /** \class RotationGroup
     *
     *  \brief Implements the n(n-1)/2-dimensional rotation group SO(n) as
     *  Manifold.
     *
     **/
    
    class RotationGroup : public Manifold {
    public:
      explicit RotationGroup ( int n ) :
	Manifold ( "Rotation Group", n*(n-1)/2, n*n) {};
    };
    
    /** \namespace LORM::RotationGroup::VectorOperations
     *
     * \brief Namespace consisting of helper functions and templates for vector
     * operations on the rotation group SO(n), such as projections and geodesic
     * evaluations.
     * 
     **/
    
    namespace VectorOperations {
      
      /// Projects a point given in Cartesian coordinates on to the rotation group
      /// SO(n) by means of the SVD.
      template<typename T_scalar>
      void projectPointOnManifold ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords );
      
      /// Projects a tangent vector given in Cartesian coordinates on to the
      /// tangent space of the rotation group SO(n) determined by the Cartesian
      /// coordinates of the base point.
      template<typename T_scalar>
      void projectVectorOnTangentSpace( 
             Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ,
	     const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords );
      
      /// Computes the geodesic step on the rotatiton group SO(n) of length \c
      /// stepLength starting at the base point in direction of the tangent
      /// vectors. The base point and the tangent vector are given in Cartesian
      /// coordinates and determine the dimension of the rotation group.
      template<typename T_scalar>
      void computeGeodesicStep ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords ,
				 Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,
				 T_scalar stepLength);
      
      /// Returns the Cartesian coordinates of the normal vector on the rotation
      /// group SO(n), which is determined by the Cartesian coordinates of the
      /// base point and tangent vector.
      template<typename T_scalar>
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeNormalVectorCoordinates ( 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentVectorCoords);
    
    } // end namespace VectorOperations

  } // end namespace RotationGroup    
    
} // end namespace LORM

#endif // _ROTATION_GROUP_HPP_
