// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _ROTATION_GROUP_POINT_ARRAY_HPP_
#define _ROTATION_GROUP_POINT_ARRAY_HPP_

#include "RotationGroup.hpp"
#include "RotationGroupTangentVectorArray.hpp"
#include "../ManifoldPointArray.hpp"

namespace LORM {

  namespace RotationGroup {
    
    // forward declaration
    template<typename T_scalar>
    class RotationGroupTangentVectorArray;
    
    /** \class RotationGroupPointArray
     *
     *  \brief Implements an array of points on the rotation group SO(n) as
     *  ManifoldPointArray.
     *
     **/
    
    template<typename T_scalar> class RotationGroupPointArray : 
      public  ManifoldPointArray<T_scalar>, 
      public  std::enable_shared_from_this<RotationGroupPointArray<T_scalar> > { 
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of RotationGroup.
      explicit RotationGroupPointArray ( std::shared_ptr<const RotationGroup> sp_rotationGroup ) : 
	ManifoldPointArray<T_scalar> ( sp_rotationGroup ) {};
      
      /// Constructor. One must provide the number n for SO(n), so that an
      /// instance of RotationGroup can be constructed.
      explicit RotationGroupPointArray ( int n ) :
	ManifoldPointArray<T_scalar> ( std::make_shared<const RotationGroup> (n) ) {};
      
      /// Implements the creation of a new ManifoldPointArray returned by a shared
      /// pointer, which is copy-constructed by the actual ManifoldPointArray.
      std::shared_ptr<ManifoldPointArray<T_scalar> > clonePointArray() const {
	return std::make_shared<RotationGroupPointArray<T_scalar> > ( *this );
      };
      
      /// Implements the creation of a a new ManifoldTangentVectorArray returned by
      /// a shared pointer with ManifoldPointArray as its base point. 
      /// \note To use this function properly, the actual instance of
      /// RotationGroupPointArray must be created as a shared pointer.
      std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > createTangentVectorArray() const {
	return std::make_shared<RotationGroupTangentVectorArray<T_scalar> >( this->shared_from_this() );
      };
      
    private:
      void projectPointOnManifold( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coordinates ) {
	VectorOperations::projectPointOnManifold( p_coordinates );
      };
    };

  }  // end namespace RotationGroup
    
} // end namespace LORM


#endif // _ROTATION_GROUP_POINT_ARRAY_HPP_
