// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _ROTATION_GROUP_TANGENT_VECTOR_ARRAY_HPP_
#define _ROTATION_GROUP_TANGENT_VECTOR_ARRAY_HPP_

#include "../ManifoldTangentVectorArray.hpp"
#include "RotationGroupPointArray.hpp"

namespace LORM {

  namespace RotationGroup {

    /** \class RotationGroupTangentVectorArray
     *
     *  \brief Implements an array of tangent vectors on the rotation group SO(n)
     *  as ManifoldTangentVectorArray.
     *
     */
    
    // forward declaration
    template<typename T_scalar>
    class RotationGroupPointArray;
    
    template<typename T_scalar>
    class RotationGroupTangentVectorArray : public ManifoldTangentVectorArray<T_scalar> {
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of
      /// RotationGroupPointArray.
      explicit RotationGroupTangentVectorArray ( 
         const std::shared_ptr<const RotationGroupPointArray<T_scalar> > & sp_basePointArray ) :
	ManifoldTangentVectorArray<T_scalar> ( sp_basePointArray ) {};
    private:
      void projectVectorOnTangentSpace(Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
				       const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const {
	VectorOperations::projectVectorOnTangentSpace( p_tangentCoords, basePointCoords );
      };
      void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
				Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentVectorCoords,
				T_scalar stepLength ) const {
	VectorOperations::computeGeodesicStep( p_basePointCoords, p_tangentVectorCoords, stepLength);
      };
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
    	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const {
	return VectorOperations::computeNormalVectorCoordinates( basePointCoords, tangentCoords );
      };
    
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const {
	return this->computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
	  coefficients, basePointCoords);
      };
    };
    
  } // end namespace RotationGroup
    
} // end namespace LORM
  
#endif // _ROTATION_GROUP_TANGENT_VECTOR_ARRAY_HPP_
