// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _EULER_ANGELS_COORDINATES_HPP_
#define _EULER_ANGELS_COORDINATES_HPP_

#include "RotationGroupTimesR.hpp"
#include "../ManifoldParameterization.hpp"

namespace LORM {

  namespace RotationGroup {

    /** \class EulerAnglesTimesR
     *
     *  \brief Implements the parameterization of RotationGroupTimesR for n=3 as
     *  ManifoldParameterization.
     *
     **/
    
    template<typename T_scalar>
    class EulerAnglesTimesR : 
      public ManifoldParameterization<T_scalar> {
    public:
      /// Default-Constructor. An instance of a 4-dimensional manifold SO(3) x R
      /// is created.
      explicit EulerAnglesTimesR() :
	ManifoldParameterization<T_scalar> ( std::make_shared<const RotationGroupTimesR> (3) ) {};
      
      /// Constructor. One must provide a valid shared pointer to an instance of
      /// RotationGroupTimesR with dimension = 4. 
      explicit EulerAnglesTimesR(const std::shared_ptr<const RotationGroupTimesR> & sp_rotationGroupTimesR) :
	ManifoldParameterization<T_scalar> ( sp_rotationGroupTimesR ) {
	assert( sp_rotationGroupTimesR->getDimension() == 4); };
      
      /// Returns the shared pointer to RotationGroupTimesR, which is the underlying manifold.
      const std::shared_ptr<const RotationGroupTimesR> getRotationGroupTimesR() const { 
	return std::static_pointer_cast<const RotationGroupTimesR> (this->getManifold()); };
      
    private:
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeParameterization( 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoords ) const;		          

      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeInverseParameterization(            
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeJacobiMatrix( 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoords ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeInverseRiemannianMatrix( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoords ) const;  
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination(
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoords ) const;

    };

  } // end namespace RotationGroup
  
} // end namespace LORM

#endif // _EULER_ANGLES_COORDINATES_HPP_
