// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _ROTATION_GROUP_TIMES_R_POINT_ARRAY_
#define _ROTATION_GROUP_TIMES_R_POINT_ARRAY_

#include "../RotationGroup/RotationGroup.hpp"
#include "RotationGroupTimesR.hpp"
#include "RotationGroupTimesRTangentVectorArray.hpp"
#include "../ManifoldPointArray.hpp"

namespace LORM {

  namespace RotationGroup {

    // forward declaration
    template<typename T_scalar>
    class RotationGroupTimesRTangentVectorArray;
    
    /** \class RotationGroupTimesRPointArray
     *
     *  \brief Implements an array of points on the manifold SO(n) x R as
     *  ManifoldPointArray.
     *
     **/
    
    template<typename T_scalar> class RotationGroupTimesRPointArray : 
      public  ManifoldPointArray<T_scalar>, 
    public  std::enable_shared_from_this<RotationGroupTimesRPointArray<T_scalar> > { 
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of RotationGroupTimesR.
      explicit RotationGroupTimesRPointArray (
					      std::shared_ptr<const RotationGroupTimesR> sp_rotationGroupTimesR ) : 
	ManifoldPointArray<T_scalar> ( sp_rotationGroupTimesR ) {};
      
      /// Constructor. One must provide the dimension n of the rotation group
      /// SO(n), so that an instance of SphereTimesR can be constructed.
      explicit RotationGroupTimesRPointArray ( int n) :
	ManifoldPointArray<T_scalar> ( std::make_shared<const RotationGroupTimesR> (n) ) {};
      
      /// Implements the creation of a new ManifoldPointArray returned by a shared
      /// pointer, which is copy-constructed by the actual ManifoldPointArray.
      std::shared_ptr<ManifoldPointArray<T_scalar> > clonePointArray() const {
	return std::make_shared<RotationGroupTimesRPointArray<T_scalar> > ( *this );
      };
      
      /// Implements the creation of a a new ManifoldTangentVectorArray returned by
      /// a shared pointer with ManifoldPointArray as its base point. 
      /// \note To use this function properly, the actual instance of
      /// RotationGroupTimesRPointArray must be created as a shared pointer.
      std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > createTangentVectorArray() const {
	return std::make_shared<RotationGroupTimesRTangentVectorArray<T_scalar> >( this->shared_from_this() );
      };
      
    private:
      void projectPointOnManifold( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coordinates ) {
	int dim = this->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointRotationGroup = p_coordinates->topRows( dim-1 );
	VectorOperations::projectPointOnManifold( &pointRotationGroup );
	p_coordinates->topRows( dim-1 ) = pointRotationGroup;
      };
    };
    
  } // end namespace RotationGroup

} // end namespace LORM

#endif // _ROTATION_GROUP_TIMES_R_POINT_ARRAY_
