// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _ROTATION_GROUP_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_
#define _ROTATION_GROUP_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_

#include "../ManifoldTangentVectorArray.hpp"
#include "RotationGroupTimesRPointArray.hpp"

namespace LORM {

  namespace RotationGroup {

    /** \class RotationGroupTimesRTangentVectorArray
     *
     *  \brief Implements an array of tangent vectors on the manifold SO(n) x R as
     *  ManifoldTangentVectorArray.
     *
     */
    
    // forward declaration
    template<typename T_scalar>
    class RotationGroupTimesRPointArray;
    
    template<typename T_scalar>
    class RotationGroupTimesRTangentVectorArray : public ManifoldTangentVectorArray<T_scalar> {
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of RotationGroupTimesRPointArray.
      explicit RotationGroupTimesRTangentVectorArray ( const std::shared_ptr<const RotationGroupTimesRPointArray<T_scalar> > & sp_basePointArray ) :
	ManifoldTangentVectorArray<T_scalar> ( sp_basePointArray ) {};
    private:
      void projectVectorOnTangentSpace(Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
				       const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupTangentCoords = p_tangentCoords->topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupBasePointCoords = basePointCoords.topRows(dim-1);
	VectorOperations::projectVectorOnTangentSpace( &rotationGroupTangentCoords, rotationGroupBasePointCoords );
	p_tangentCoords->topRows(dim-1) = rotationGroupTangentCoords;
      };
      void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
				Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords,
				T_scalar stepLength ) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupBasePointCoords = 
	  p_basePointCoords->topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupTangentCoords = 
	  p_tangentCoords->topRows(dim-1);
	VectorOperations::computeGeodesicStep( &rotationGroupBasePointCoords, 
							    &rotationGroupTangentCoords, 
							    stepLength);
	p_basePointCoords->topRows(dim-1) = rotationGroupBasePointCoords;
	p_tangentCoords->topRows(dim-1) = rotationGroupTangentCoords;
	(*p_basePointCoords)(dim-1) += stepLength * (*p_tangentCoords)(dim-1);
    };
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
										const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
										const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupBasePointCoords =  basePointCoords.topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> rotationGroupTangentCoords =  tangentCoords.topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> normalCoords;
	normalCoords.setZero(dim);
	normalCoords.topRows(dim-1) = 
	  VectorOperations::computeNormalVectorCoordinates( rotationGroupBasePointCoords, 
								       rotationGroupTangentCoords);
	return normalCoords;
      };
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  computeChristoffelMatrixLinearCombination(
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const {
	return this->computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
											  coefficients, basePointCoords);
      };
  };

  } // end namespace RotationGroup
  
} // end namespace LORM

#endif // _ROTATION_GROUP_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_
