// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "Sphere.hpp"

namespace LORM {
  namespace Sphere {
    namespace VectorOperations {

      template<typename T_scalar>
      void computeGeodesicStep ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords ,
				 Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,
				 T_scalar stepLength) {
	auto & v_old = *p_tangentVectorCoords;
	T_scalar v_norm = v_old.matrix().norm();
	
	if ( v_norm !=0 ) {
	  auto x_old = *p_basePointCoords;
	  *p_basePointCoords = cos( stepLength * v_norm) * x_old + sin( stepLength * v_norm) / v_norm * v_old;
	  *p_tangentVectorCoords  = - v_norm * sin( stepLength * v_norm ) * x_old  + cos( stepLength * v_norm ) * v_old;
	}
      }
    } // end namespace VectorOperations

    template<typename T_scalar>
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> generateRandomSpherePointCoordinates( 
      std::default_random_engine * p_randomNumberGenerator, int dimension ) {
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointCoords;
      pointCoords.resize( dimension + 1 ); 
      std::normal_distribution<T_scalar> distribution(0.0, 1.0);
      for( int i=0; i< dimension + 1; ++i ) 
	pointCoords(i) = distribution( *p_randomNumberGenerator );
      pointCoords.matrix().normalize();
      return pointCoords;
    }
      
    template<typename T_scalar>
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> generateRandomSpherePointCoordinatesArray(
      std::default_random_engine * p_randomNumberGenerator, int dimension, int numberOfPoints ) {
      Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> pointCoordsArray;
      pointCoordsArray.resize( dimension+1, numberOfPoints );
      for ( int i=0; i< numberOfPoints; ++i )
	pointCoordsArray.col(i) = generateRandomSpherePointCoordinates<T_scalar>(
  	  p_randomNumberGenerator, dimension);
      return pointCoordsArray;
    }

  } // end namespace Sphere

} // end namespace LORM

//Macro for explicit instantiations
#define LORM_INSTANTIATE_SPHERE(T_scalar)			\
  template void LORM::Sphere::VectorOperations::computeGeodesicStep<T_scalar> ( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords , \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,T_scalar stepLength); \
  template Eigen::Array<T_scalar, Eigen::Dynamic, 1> \
     LORM::Sphere::generateRandomSpherePointCoordinates<T_scalar>(		\
    std::default_random_engine * p_randomNumberGenerator, int dimension ); \
  template Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> \
     LORM::Sphere::generateRandomSpherePointCoordinatesArray<T_scalar>(	\
    std::default_random_engine * p_randomNumberGenerator, int dimension, int numberOfPoints );
  
#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_SPHERE( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_SPHERE(double)
#endif

#undef LORM_INSTANTIATE_SPHERE
