// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERE_HPP_
#define _SPHERE_HPP_

#include "../Manifold.hpp"
#include <random>

namespace LORM {

  /** \namespace LORM::Sphere
   *
   *  \brief Namespace consisting of classes and templates related to the sphere S^d.
   *
   **/

  namespace Sphere {
    
    /** \class Sphere
     *
     *  \brief Implements the d-dimensional sphere S^d as Manifold.
     *
     **/
    
    class Sphere : public Manifold {
    public:
      explicit Sphere ( int dimension ) :
	Manifold ( "Sphere", dimension, dimension + 1) {};
    };
    
    /** \namespace LORM::Sphere::VectorOperations
     *
     * \brief Namespace consisting of helper functions and templates for vector
     * operations on the d-dimensional sphere S^d, such as projections and
     * geodesic evaluations.
     * 
     **/

    namespace VectorOperations {
    
      /// Projects a point given in Cartesian coordinates on to the sphere S^d of
      /// the corresponding dimension.
      template<typename T_scalar>
      void projectPointOnManifold ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ) {
	p_coords->matrix().normalize(); }
      
      /// Projects a tangent vector given in Cartesian coordinates on to the
      /// tangent space of the sphere S^d determined by the Cartesian coordinates
      /// of the base point.
      template<typename T_scalar>
      void projectVectorOnTangentSpace( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ,
			       const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords )  {
	(*p_coords) -= ((*p_coords) * basePointCoords).sum() * basePointCoords;
      }
      
      /// Computes the geodesic step on the sphere S^d of length \c stepLength
      /// starting at the base point in direction of the tangent vectors. The base
      /// point and the tangent vector are given in Cartesian coordinates and
      /// determine the dimension of the sphere.
      template<typename T_scalar>
      void computeGeodesicStep ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords ,
				 Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,
				 T_scalar stepLength);
      
      /// Returns the Cartesian coordinates of the normal vector on the sphere
      /// S^d, which is determined by the Cartesian coordinates of the base point and tangent vector.
      template<typename T_scalar>
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeNormalVectorCoordinates ( 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) {
	return -tangentCoords.square().sum() * basePointCoords;
      }
      
    } // end namespace SphereVectorOperations
    
    /// Generates a random point on a sphere given in Cartesian coordinates. One
    /// must provide a pointer to a random number generator, and the dimension of
    /// the sphere.
    template<typename T_scalar>
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> generateRandomSpherePointCoordinates( 
      std::default_random_engine * p_randomNumberGenerator, int dimension );
    
    /// Generates a an array of random points on a sphere given in Cartesian
    /// coordinates. One must provide a pointer to a random number generator, the
    /// dimension of the sphere, and the number of points.
    template<typename T_scalar>
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> generateRandomSpherePointCoordinatesArray(
      std::default_random_engine * p_randomNumberGenerator, int dimension, int numberOfPoints );
    
  } // end namespace Sphere
    
} // end namespace LORM

#endif // _SPHERE_HPP_
