// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERE_POINT_ARRAY_HPP_
#define _SPHERE_POINT_ARRAY_HPP_

#include "Sphere.hpp"
#include "SphereTangentVectorArray.hpp"
#include "../ManifoldPointArray.hpp"

namespace LORM {

  namespace Sphere {
    
    // forward declaration
    template<typename T_scalar>
    class SphereTangentVectorArray;

    /** \class SpherePointArray
     *
     *  \brief Implements an array of points on the sphere S^d as
     *  ManifoldPointArray.
     *
     **/
    
    template<typename T_scalar> class SpherePointArray : 
      public  ManifoldPointArray<T_scalar>, 
      public  std::enable_shared_from_this<SpherePointArray<T_scalar> > { 
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of Sphere.
      explicit SpherePointArray ( std::shared_ptr<const Sphere> sp_sphere ) : 
	ManifoldPointArray<T_scalar> ( sp_sphere ) {};
      
      /// Constructor. One must provide the dimension of Sphere, so that an
      /// instance of Sphere can be constructed.
      explicit SpherePointArray ( int dim ) :
	ManifoldPointArray<T_scalar> ( std::make_shared<const Sphere> (dim) ) {};
      
      /// Implements the creation of a new ManifoldPointArray returned by a shared
      /// pointer, which is copy-constructed by the actual ManifoldPointArray.
      std::shared_ptr<ManifoldPointArray<T_scalar> > clonePointArray() const {
	return std::make_shared<SpherePointArray<T_scalar> > ( *this );
      };
      
      /// Implements the creation of a a new ManifoldTangentVectorArray returned by
      /// a shared pointer with ManifoldPointArray as its base point. 
      /// \note To use this function properly, the actual instance of
      /// SpherePointArray must be created as a shared pointer.
      std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > createTangentVectorArray() const {
	return std::make_shared<SphereTangentVectorArray<T_scalar> >( this->shared_from_this() );
      };
      
    private:
      void projectPointOnManifold( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coordinates ) {
	VectorOperations::projectPointOnManifold( p_coordinates );
      };
    };
    
  } // end namespace Sphere

} // end namespace LORM


#endif // _SPHERE_POINT_ARRAY_HPP_
