// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERE_TANGENT_VECTOR_ARRAY_HPP_
#define _SPHERE_TANGENT_VECTOR_ARRAY_HPP_

#include "../ManifoldTangentVectorArray.hpp"
#include "SpherePointArray.hpp"

namespace LORM {

  namespace Sphere {

    /** \class SphereTangentVectorArray
     *
     *  \brief Implements an array of tangent vectors on the sphere S^d as
     *  ManifoldTangentVectorArray.
     *
     */
    
    // forward declaration
    template<typename T_scalar>
    class SpherePointArray;
    
    template<typename T_scalar>
    class SphereTangentVectorArray : public ManifoldTangentVectorArray<T_scalar> {
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of SpherePointArray.
      explicit SphereTangentVectorArray ( const std::shared_ptr<const SpherePointArray<T_scalar> > & sp_basePointArray ) :
	ManifoldTangentVectorArray<T_scalar> ( sp_basePointArray ) {};
    private:
      void projectVectorOnTangentSpace(Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
				       const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const {
	VectorOperations::projectVectorOnTangentSpace( p_tangentCoords, basePointCoords );
      };
      void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
				Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentVectorCoords,
				T_scalar stepLength ) const {
	VectorOperations::computeGeodesicStep( p_basePointCoords, p_tangentVectorCoords, stepLength);
      };
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
           const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	   const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const {
	return VectorOperations::computeNormalVectorCoordinates( basePointCoords, tangentCoords );
      };
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  computeChristoffelMatrixLinearCombination(
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const {
	//return this->computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
	//	 coefficients, basePointCoords);
	// The following implementation differs from the above generic one, that
	// we do not apply the projection matrices from left and right. Hence it
	// is only applicable for tangent vectors, which are already projected
	Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
	christoffelMatrix.setIdentity( basePointCoords.rows(), basePointCoords.rows() );
	christoffelMatrix *= (coefficients*basePointCoords).sum();
	return christoffelMatrix;
      };
    };
    
  } // end namespace Sphere

} // end namespace LORM

#endif // _SPHERE_TANGENT_VECTOR_ARRAY_HPP_
