// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "SphericalCoordinates.hpp"

using namespace LORM::Sphere;

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> SphericalCoordinates<T_scalar>::computeParameterization(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & sphericalCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointCoords;
  pointCoords.resize(3);
  pointCoords(0) = sin( sphericalCoords(0) ) * cos( sphericalCoords(1) );
  pointCoords(1) = sin( sphericalCoords(0) ) * sin( sphericalCoords(1) );
  pointCoords(2) = cos( sphericalCoords(0) );
  return pointCoords;
}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> SphericalCoordinates<T_scalar>::computeInverseParameterization(              
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphericalCoords;
  sphericalCoords.resize(2);
  // theta
  sphericalCoords(0) = acos( pointCoords(2) );
  // phi
  sphericalCoords(1) = pointCoords(1) / 
    ( pointCoords(0) - sqrt(pointCoords(0)*pointCoords(0) + pointCoords(1)*pointCoords(1)) );
  sphericalCoords(1) = 2*atan( sphericalCoords(1) ) + M_PI;
  if ( pointCoords(1) == 0 && 0 <= pointCoords(0) )
    sphericalCoords(1) = 0;
  return sphericalCoords;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinates<T_scalar>::computeJacobiMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> jacobiMatrix;
  jacobiMatrix.resize(3,2);
  // d_theta
  jacobiMatrix(0,0) =  cos(localPointCoords(0)) * cos(localPointCoords(1));
  jacobiMatrix(1,0) =  cos(localPointCoords(0)) * sin(localPointCoords(1));
  jacobiMatrix(2,0) = -sin(localPointCoords(0));
  // d_phi
  jacobiMatrix(0,1) = -sin(localPointCoords(0)) * sin(localPointCoords(1));
  jacobiMatrix(1,1) =  sin(localPointCoords(0)) * cos(localPointCoords(1));      
  jacobiMatrix(2,1) = 0;
  return jacobiMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinates<T_scalar>::computeInverseRiemannianMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> invRiemannMatrix;
  invRiemannMatrix.resize(2,2);
  // theta column
  invRiemannMatrix(0,0) = 1.;
  invRiemannMatrix(1,0) = 0.;
  // phi column
  invRiemannMatrix(0,1) = 0.;
  invRiemannMatrix(1,1) = 1. / ( sin(localPointCoords(0)) * sin(localPointCoords(0)) );
  return invRiemannMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinates<T_scalar>::computeChristoffelMatrixLinearCombination(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
  christoffelMatrix.setZero(2,2);
  // theta_coeff * theta_ChristoffelMatrix
  christoffelMatrix(1,1) += -localCoefficients(0) * sin(localPointCoords(0)) * cos(localPointCoords(0));
  // phi_coeff * phi_ChristoffelMatrix
  christoffelMatrix(0,1) += localCoefficients(1) / tan(localPointCoords(0));
  christoffelMatrix(1,0) += christoffelMatrix(0,1);
  return christoffelMatrix;
}

//Macro for explicit instantiations of SphericalCoordinates<T_scalar>
#define LORM_INSTANTIATE_SPHERICALCOORDINATES(T_scalar)	\
  template class SphericalCoordinates<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_SPHERICALCOORDINATES( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_SPHERICALCOORDINATES(double)
#endif

#undef LORM_INSTANTIATE_SPHERICALCOORDINATES
