// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERICAL_COORDINATES_HPP_
#define _SPHERICAL_COORDINATES_HPP_

#include "Sphere.hpp"
#include "../ManifoldParameterization.hpp"

namespace LORM {

  namespace Sphere {
    
    /** \class SphericalCoordinates 
     *
     *  \brief Implements the parameterization of the sphere S^2 by spherical
     *  coordinates (theta,phi) in [0,pi)x[0,2*pi) as ManifoldParameterization.
     *
     **/
    
    template<typename T_scalar>
    class SphericalCoordinates : 
      public ManifoldParameterization<T_scalar> {
    public:
    /// Default-Constructor.  An instance of a 2-dimensional Sphere is created.
      explicit SphericalCoordinates() :
	ManifoldParameterization<T_scalar> ( std::make_shared<const Sphere> (2) ) {};
      
      /// Constructor. One must provide a valid shared pointer to an instance of
      /// Sphere with dimension = 2. 
      explicit SphericalCoordinates(const std::shared_ptr<const Sphere> & sp_sphere) :
	ManifoldParameterization<T_scalar> ( sp_sphere ) {
	assert( sp_sphere->getDimension() == 2); };
      
      /// Returns the shared pointer to Sphere, which is the underlying manifold.
      const std::shared_ptr<const Sphere> getSphere() const { 
	return std::static_pointer_cast<const Sphere> 
	  (this->getManifold()); };
      
    private:
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeParameterization( 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & sphericalCoords ) const;		                
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeInverseParameterization(              
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const;

      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeJacobiMatrix( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeInverseRiemannianMatrix( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;

      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination(
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;
      
    };

  } // end namespace Sphere

} // end namespace LORM

#endif // _SPHERICAL_COORDINATES_HPP_
