// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERE_TIMES_R_POINT_ARRAY_
#define _SPHERE_TIMES_R_POINT_ARRAY_

#include "../Sphere/Sphere.hpp"
#include "SphereTimesR.hpp"
#include "SphereTimesRTangentVectorArray.hpp"
#include "../ManifoldPointArray.hpp"

namespace LORM {

  namespace Sphere {
    
    // forward declaration
    template<typename T_scalar>
    class SphereTimesRTangentVectorArray;
    
    /** \class SphereTimesRPointArray
     *
     *  \brief Implements an array of points on the manifold S^d x R as
     *  ManifoldPointArray.
     *
     **/
    
    template<typename T_scalar> class SphereTimesRPointArray : 
      public  ManifoldPointArray<T_scalar>, 
      public  std::enable_shared_from_this<SphereTimesRPointArray<T_scalar> > { 
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of SphereTimesR.
      explicit SphereTimesRPointArray ( std::shared_ptr<const SphereTimesR> sp_sphereTimesR ) : 
	ManifoldPointArray<T_scalar> ( sp_sphereTimesR ) {};
      
      /// Constructor. One must provide the dimension d of the sphere S^d, so that an
      /// instance of SphereTimesR can be constructed.
      explicit SphereTimesRPointArray ( int dim ) :
	ManifoldPointArray<T_scalar> ( std::make_shared<const SphereTimesR> (dim) ) {};
      
      /// Implements the creation of a new ManifoldPointArray returned by a shared
      /// pointer, which is copy-constructed by the actual ManifoldPointArray.
      std::shared_ptr<ManifoldPointArray<T_scalar> > clonePointArray() const {
	return std::make_shared<SphereTimesRPointArray<T_scalar> > ( *this );
      };
      
      /// Implements the creation of a a new ManifoldTangentVectorArray returned by
      /// a shared pointer with ManifoldPointArray as its base point. 
      /// \note To use this function properly, the actual instance of
      /// SphereTimesRPointArray must be created as a shared pointer.
      std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > createTangentVectorArray() const {
	return std::make_shared<SphereTimesRTangentVectorArray<T_scalar> >( this->shared_from_this() );
      };
      
    private:
      void projectPointOnManifold( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coordinates ) {
	int dim = this->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointSphere = p_coordinates->topRows( dim-1 );
	VectorOperations::projectPointOnManifold( &pointSphere );
	p_coordinates->topRows( dim-1 ) = pointSphere;
      };
    };
    
  } // end namespace Sphere
    
} // end namespace LORM

#endif // _SPHERE_TIMES_R_POINT_ARRAY_
