// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SPHERE_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_
#define _SPHERE_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_

#include "../ManifoldTangentVectorArray.hpp"
#include "SphereTimesRPointArray.hpp"

namespace LORM {
  
  namespace Sphere {

    /** \class SphereTimesRTangentVectorArray
     *
     *  \brief Implements an array of tangent vectors on the manifold S^d x R as
     *  ManifoldTangentVectorArray.
     *
     */
    
    // forward declaration
    template<typename T_scalar>
    class SphereTimesRPointArray;
    
    template<typename T_scalar>
    class SphereTimesRTangentVectorArray : public ManifoldTangentVectorArray<T_scalar> {
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of SphereTimesRPointArray.
      explicit SphereTimesRTangentVectorArray ( const std::shared_ptr<const SphereTimesRPointArray<T_scalar> > & sp_basePointArray ) :
	ManifoldTangentVectorArray<T_scalar> ( sp_basePointArray ) {};
    private:
      void projectVectorOnTangentSpace(Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
				       const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereTangentCoords = p_tangentCoords->topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereBasePointCoords = basePointCoords.topRows(dim-1);
	VectorOperations::projectVectorOnTangentSpace( &sphereTangentCoords, sphereBasePointCoords );
	p_tangentCoords->topRows(dim-1) = sphereTangentCoords;
      };
      void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
				Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords,
				T_scalar stepLength ) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereBasePointCoords = p_basePointCoords->topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereTangentCoords = p_tangentCoords->topRows(dim-1);
	VectorOperations::computeGeodesicStep( &sphereBasePointCoords, &sphereTangentCoords, stepLength);
	p_basePointCoords->topRows(dim-1) = sphereBasePointCoords;
	p_tangentCoords->topRows(dim-1) = sphereTangentCoords;
	(*p_basePointCoords)(dim-1) += stepLength * (*p_tangentCoords)(dim-1);
      };
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
										const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
										const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const {
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereBasePointCoords =  basePointCoords.topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> sphereTangentCoords =  tangentCoords.topRows(dim-1);
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> normalCoords;
	normalCoords.setZero(dim);
	normalCoords.topRows(dim-1) = 
	  VectorOperations::computeNormalVectorCoordinates( sphereBasePointCoords, 
								    sphereTangentCoords);
	return normalCoords;
      };
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  computeChristoffelMatrixLinearCombination(
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const {
	// return this->computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
	//	 coefficients, basePointCoords);
	// The following implementation differs from the above generic one, that
	// we do not apply the projection matrices from left and right. Hence it
	// is only applicable for tangent vectors, which are already projected
	int dim = this->getBasePointArray()->getManifold()->getAmbientSpaceDimension();
	Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
	christoffelMatrix.setIdentity(dim, dim);
	christoffelMatrix(dim-1, dim-1) = 0;
	christoffelMatrix *= (coefficients.topRows(dim-1)*basePointCoords.topRows(dim-1)).sum();
	return christoffelMatrix;
      };
    };

  } // end namespace Sphere
  
} // end namespace LORM
  
#endif // _SPHERE_TIMES_R_TANGENT_VECTOR_ARRAY_HPP_
