// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "SphericalCoordinatesTimesR.hpp"

using namespace LORM::Sphere;

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> SphericalCoordinatesTimesR<T_scalar>::computeParameterization( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointCoords;
  pointCoords.resize(4);
  pointCoords(0) = sin( localCoords(0) ) * cos( localCoords(1) );
  pointCoords(1) = sin( localCoords(0) ) * sin( localCoords(1) );
  pointCoords(2) = cos( localCoords(0) );
  pointCoords(3) = localCoords(2);
  return pointCoords;
}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> SphericalCoordinatesTimesR<T_scalar>::computeInverseParameterization(              
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> localCoords;
  localCoords.resize(3);
  // theta
  localCoords(0) = acos( pointCoords(2) );
  // phi
  localCoords(1) = pointCoords(1) / 
    ( pointCoords(0) - sqrt(pointCoords(0)*pointCoords(0) + pointCoords(1)*pointCoords(1)) );
  localCoords(1) = 2*atan( localCoords(1) ) + M_PI;
  if ( pointCoords(1) == 0 && 0 <= pointCoords(0) )
    localCoords(1) = 0;
  // w
  localCoords(2) = pointCoords(3);
  return localCoords;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinatesTimesR<T_scalar>::computeJacobiMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> jacobiMatrix;
  jacobiMatrix.resize(4,3);
  // d_theta
  jacobiMatrix(0,0) =  cos(localPointCoords(0)) * cos(localPointCoords(1));
  jacobiMatrix(1,0) =  cos(localPointCoords(0)) * sin(localPointCoords(1));
  jacobiMatrix(2,0) = -sin(localPointCoords(0));
  jacobiMatrix(3,0) = 0;
  // d_phi
  jacobiMatrix(0,1) = -sin(localPointCoords(0)) * sin(localPointCoords(1));
  jacobiMatrix(1,1) =  sin(localPointCoords(0)) * cos(localPointCoords(1));      
  jacobiMatrix(2,1) = 0;
  jacobiMatrix(3,1) = 0;
  // d_w
  jacobiMatrix(0,2) = 0;
  jacobiMatrix(1,2) = 0;      
  jacobiMatrix(2,2) = 0;
  jacobiMatrix(3,2) = 1;
  return jacobiMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinatesTimesR<T_scalar>::computeInverseRiemannianMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> invRiemannMatrix;
  invRiemannMatrix.resize(3,3);
  // theta column
  invRiemannMatrix(0,0) = 1.;
  invRiemannMatrix(1,0) = 0.;
  invRiemannMatrix(2,0) = 0.;
  // phi column
  invRiemannMatrix(0,1) = 0.;
  invRiemannMatrix(1,1) = 1. / ( sin(localPointCoords(0)) * sin(localPointCoords(0)) );
  invRiemannMatrix(2,1) = 0.;
  // w column
  invRiemannMatrix(0,2) = 0;
  invRiemannMatrix(1,2) = 0;
  invRiemannMatrix(2,2) = 1;
  return invRiemannMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> SphericalCoordinatesTimesR<T_scalar>::computeChristoffelMatrixLinearCombination(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
  christoffelMatrix.setZero(3,3);
  // theta_coeff * theta_ChristoffelMatrix
  christoffelMatrix(1,1) += -localCoefficients(0) * sin(localPointCoords(0)) * cos(localPointCoords(0));
  // phi_coeff * phi_ChristoffelMatrix
  christoffelMatrix(0,1) += localCoefficients(1) / tan(localPointCoords(0));
  christoffelMatrix(1,0) += christoffelMatrix(0,1);
  return christoffelMatrix;
}

//Macro for explicit instantiations of SphericalCoordinatesTimesR<T_scalar>
#define LORM_INSTANTIATE_SPHERICALCOORDINATESTIMESR(T_scalar)	\
  template class SphericalCoordinatesTimesR<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_SPHERICALCOORDINATESTIMESR( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_SPHERICALCOORDINATESTIMESR(double)
#endif

#undef LORM_INSTANTIATE_SPHERICALCOORDINATESTIMESR
