// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "Torus.hpp"
#include "../Sphere/Sphere.hpp"

namespace LORM {
  namespace Torus {
    namespace VectorOperations {

      template<typename T_scalar>
      void projectPointOnManifold ( Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ) {
	assert( p_coords->rows() % 2 == 0 );
	for ( int i=0; i<(p_coords->rows() / 2); ++i ) {
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> temp = p_coords->segment(2*i,2);
	  LORM::Sphere::VectorOperations::projectPointOnManifold( &temp );
	  p_coords->segment(2*i,2) = temp;	  
	}
      }

      template<typename T_scalar>
      void projectVectorOnTangentSpace( 
            Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ,
	    const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords ) {
	assert( p_coords->rows() % 2 == 0 );
	for ( int i=0; i<(p_coords->rows() / 2); ++i ) {
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempTangentCoords = 
	    p_coords->segment(2*i,2);
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempBaseCoords = 
	    basePointCoords.segment(2*i,2);
	  LORM::Sphere::VectorOperations::projectVectorOnTangentSpace( 
					      &tempTangentCoords, tempBaseCoords );
	  p_coords->segment(2*i,2) = tempTangentCoords;	  
	}
      }

      template<typename T_scalar>
      void computeGeodesicStep ( 
            Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords ,
	    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,
	    T_scalar stepLength) {
	assert( p_basePointCoords->rows() % 2 == 0 );
	for ( int i=0; i<(p_basePointCoords->rows() / 2); ++i ) {
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempTangentCoords = 
	    p_tangentVectorCoords->segment(2*i,2);
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempBaseCoords = 
	    p_basePointCoords->segment(2*i,2);
	  LORM::Sphere::VectorOperations::computeGeodesicStep( 
				 &tempBaseCoords, &tempTangentCoords, stepLength );
	  p_tangentVectorCoords->segment(2*i,2) = tempTangentCoords;	  
	  p_basePointCoords->segment(2*i,2) = tempBaseCoords;	  
	}
      }

      template<typename T_scalar>
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeNormalVectorCoordinates ( 
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) {
	assert( basePointCoords.rows() % 2 == 0 );
	Eigen::Array<T_scalar, Eigen::Dynamic, 1> normalCoords = 
	  Eigen::Array<T_scalar, Eigen::Dynamic, 1>::Zero(basePointCoords.rows());
	for ( int i=0; i<(basePointCoords.rows() / 2); ++i ) {
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempTangentCoords = 
	    tangentCoords.segment(2*i,2);
	  Eigen::Array<T_scalar, Eigen::Dynamic,1> tempBaseCoords =  
	    basePointCoords.segment(2*i,2);
	  normalCoords.segment(2*i,2) = 
	    LORM::Sphere::VectorOperations::computeNormalVectorCoordinates(
						   tempBaseCoords, tempTangentCoords );
	}
	return normalCoords;
      }
      
    } // end namespace VectorOperations
  } // end namespace Torus
} // end namespace LORM

//Macro for explicit instantiations
#define LORM_INSTANTIATE_TORUS(T_scalar)			\
  template void LORM::Torus::VectorOperations::projectPointOnManifold<T_scalar> ( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords ); \
  template void LORM::Torus::VectorOperations::projectVectorOnTangentSpace<T_scalar> ( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_coords , \
    const Eigen::Array<T_scalar,Eigen::Dynamic,1> & basePointCoords );	\
  template void LORM::Torus::VectorOperations::computeGeodesicStep<T_scalar> ( \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_basePointCoords , \
    Eigen::Array<T_scalar,Eigen::Dynamic,1> * p_tangentVectorCoords ,T_scalar stepLength); \
  template Eigen::Array<T_scalar, Eigen::Dynamic, 1>  \
  LORM::Torus::VectorOperations::computeNormalVectorCoordinates<T_scalar> ( \
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,\
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords);

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_TORUS( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_TORUS(double)
#endif

#undef LORM_INSTANTIATE_TORUS
