// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "TorusParameterization.hpp"

using namespace LORM::Torus;

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> TorusParameterization<T_scalar>::computeParameterization(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> pointCoords;
  pointCoords.resize( 2*localPointCoords.rows() );
  for (int i=0; i<localPointCoords.rows(); ++i ) {
    pointCoords(2*i)   = cos( localPointCoords(i) );
    pointCoords(2*i+1) = sin( localPointCoords(i) );
  }
  return pointCoords;
}

template<typename T_scalar>
Eigen::Array<T_scalar, Eigen::Dynamic, 1> TorusParameterization<T_scalar>::computeInverseParameterization(              
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const
{
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> localPointCoords;
  localPointCoords.resize( pointCoords.rows() / 2 );
  for (int i=0; i<localPointCoords.rows(); ++i) {
    if( pointCoords(2*i) == 1 )
      localPointCoords(i) = 0;
    else
      localPointCoords(i) = 2 * atan( pointCoords(2*i+1) / (pointCoords(2*i)-1) ) + M_PI;  
  }
  return localPointCoords;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> TorusParameterization<T_scalar>::computeJacobiMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> jacobiMatrix;
  jacobiMatrix.setZero( 2*localPointCoords.rows(), 2 );
  for ( int i=0; i<localPointCoords.rows(); ++i) {
    jacobiMatrix(2*i  , i) = - sin( localPointCoords(i) );
    jacobiMatrix(2*i+1, i) = cos( localPointCoords(i) );
  }
  return jacobiMatrix;
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> TorusParameterization<T_scalar>::computeInverseRiemannianMatrix( 
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  return Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>::
    Identity( localPointCoords.rows(), localPointCoords.rows() );
}

template<typename T_scalar>
Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> TorusParameterization<T_scalar>::computeChristoffelMatrixLinearCombination(
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & ,
  const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const
{
  return Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>::
    Zero( localPointCoords.rows(), localPointCoords.rows() );
}

//Macro for explicit instantiations of TorusParameterization<T_scalar>
#define LORM_INSTANTIATE_TORUSPARAMETERIZATION(T_scalar)	\
  template class TorusParameterization<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_TORUSPARAMETERIZATION( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_TORUSPARAMETERIZATION(double)
#endif

#undef LORM_INSTANTIATE_TORUSPARAMETERIZATION
