// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _TORUS_PARAMETERIZATION_HPP_
#define _TORUS_PARAMETERIZATION_HPP_

#include "Torus.hpp"
#include "../ManifoldParameterization.hpp"

namespace LORM {

  namespace Torus {
    
    /** \class TorusParameterization 
     *
     *  \brief Implements the parameterization in local coordinates (alpha_1,
     *  ..., alpha_d) in [0,2*pi)^d of the Torus T^d as
     *  ManifoldParameterization.
     *
     **/
    
    template<typename T_scalar>
    class TorusParameterization : 
      public ManifoldParameterization<T_scalar> {
    public:
    /// Default-Constructor.  An instance of a d-dimensional Torus is created.
      explicit TorusParameterization( int dimension ) :
	ManifoldParameterization<T_scalar> ( std::make_shared<const Torus> (dimension) ) {};
      
      /// Constructor. One must provide a valid shared pointer to an instance of
      /// a torus T^d.
      explicit TorusParameterization(const std::shared_ptr<const Torus> & sp_torus) :
	ManifoldParameterization<T_scalar> ( sp_torus ) {};
      
      /// Returns the shared pointer to Torus, which is the underlying manifold.
      const std::shared_ptr<const Torus> getTorus() const { 
	return std::static_pointer_cast<const Torus> 
	  (this->getManifold()); };
      
    private:
      Eigen::Array<T_scalar, Eigen::Dynamic, 1>  computeParameterization( 
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;		                
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeInverseParameterization(              
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & pointCoords ) const;

      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeJacobiMatrix( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeInverseRiemannianMatrix( 
        const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;

      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> computeChristoffelMatrixLinearCombination(
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localCoefficients,
	const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords ) const;
      
    };

  } // end namespace Torus

} // end namespace LORM

#endif // _TORUS_PARAMETERIZATION_HPP_
