// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _TORUS_TANGENT_VECTOR_ARRAY_HPP_
#define _TORUS_TANGENT_VECTOR_ARRAY_HPP_

#include "../ManifoldTangentVectorArray.hpp"
#include "TorusPointArray.hpp"

namespace LORM {

  namespace Torus {

    /** \class TorusTangentVectorArray
     *
     *  \brief Implements an array of tangent vectors on the torus S^d as
     *  ManifoldTangentVectorArray.
     *
     */
    
    // forward declaration
    template<typename T_scalar>
    class TorusPointArray;
    
    template<typename T_scalar>
    class TorusTangentVectorArray : public ManifoldTangentVectorArray<T_scalar> {
    public:
      /// Constructor. One must provide a valid shared pointer to an instance of TorusPointArray.
      explicit TorusTangentVectorArray ( const std::shared_ptr<const TorusPointArray<T_scalar> > & sp_basePointArray ) :
	ManifoldTangentVectorArray<T_scalar> ( sp_basePointArray ) {};
    private:
      void projectVectorOnTangentSpace(
       Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentCoords, 
       const Eigen::Array<T_scalar, Eigen::Dynamic,1> & basePointCoords) const {
	VectorOperations::projectVectorOnTangentSpace( p_tangentCoords, basePointCoords );
      };
      void computeGeodesicStep( Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_basePointCoords, 
				Eigen::Array<T_scalar, Eigen::Dynamic,1> * p_tangentVectorCoords,
				T_scalar stepLength ) const {	
	VectorOperations::computeGeodesicStep( p_basePointCoords, p_tangentVectorCoords, stepLength);
      };
      Eigen::Array<T_scalar, Eigen::Dynamic, 1> computeNormalVectorCoordinates ( 
           const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords,
	   const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & tangentCoords) const {
	return VectorOperations::computeNormalVectorCoordinates( basePointCoords, tangentCoords );
      };
      
      Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  computeChristoffelMatrixLinearCombination(
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & coefficients, 
													 const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & basePointCoords ) const {
	return this->computeChristoffelMatrixLinearCombinationFromNormalVectorCoordinates( 
		 coefficients, basePointCoords);
	// The following implementation differs from the above generic one, that
	// we do not apply the projection matrices from left and right. Hence it
	// is only applicable for tangent vectors, which are already projected
	// Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> christoffelMatrix;
      // 	christoffelMatrix.setIdentity( basePointCoords.rows(), basePointCoords.rows() );
      // 	for ( int i=0; i<(basePointCoords.rows() / 2); ++i ) {
      // 	  Eigen::Matrix<T_scalar, 2, 2> temp = 
      // 	    ( coefficients.segment<2>(2*i).array() * basePointCoords.segment<2>(2*i).array() ).sum() * 
      // 	    Eigen::Matrix<T_scalar, 2, 2>::Identity();
      // 	  christoffelMatrix.block<2,2>(2*i,2*i) = temp;	   
      // 	}
      // 	return christoffelMatrix;
      };
    };
    
  } // end namespace Torus

} // end namespace LORM

#endif // _TORUS_TANGENT_VECTOR_ARRAY_HPP_
