// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "NFFT2d.hpp"

using namespace LORM::NFFT;

NFFT2d::NFFT2d( int numberOfPoints, Eigen::Array2i polynomialDegrees):
  numberOfPoints_ (numberOfPoints),
  polynomialDegrees_ (polynomialDegrees)
{ 
  initNFFT2dPlan(); 
  setLocalPointCoordinatesArray( Eigen::Array2Xd::Zero(2,numberOfPoints) );
}

NFFT2d::~NFFT2d()
{
  destroyNFFT2dPlan();
}

Eigen::Array2Xd NFFT2d::getLocalPointCoordinatesArray() const 
{
  Eigen::Array2Xd localPointCoords;
  localPointCoords.resize( 2, numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    // note, the nfft2d coordinates are in [-0.5,0.5]^2
    localPointCoords(0,i) = nfft2dPlan_.x[2*i];
    localPointCoords(1,i) = nfft2dPlan_.x[2*i+1];
  }
  return localPointCoords;
}

void NFFT2d::setLocalPointCoordinatesArray( const Eigen::Array2Xd & localPointCoords )
{
  assert( localPointCoords.cols() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    nfft2dPlan_.x[2*i]   = localPointCoords(0,i);
    nfft2dPlan_.x[2*i+1] = localPointCoords(1,i);
  }
  if( nfft2dPlan_.nfft_flags & PRE_ONE_PSI )
    nfft_precompute_one_psi( &nfft2dPlan_ );
}

Eigen::ArrayXcd NFFT2d::getF() const 
{
  Eigen::ArrayXcd f;
  f.resize( numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) 
    f(i) << nfft2dPlan_.f[i];
  return f;
}

NFFT2dFourierCoefficients NFFT2d::getFhat( Eigen::Array2i degrees) const
{
  assert( degrees(0) <= polynomialDegrees_(0) );
  assert( degrees(1) <= polynomialDegrees_(1) );
  NFFT2dFourierCoefficients fhat( degrees );
  for (int k1=-degrees(0); k1<degrees(0); ++k1)
    for (int k2=-degrees(1); k2<degrees(1); ++k2)
      fhat(k1, k2) <<  nfft2dPlan_.f_hat[linearizedIndex(k1,k2)];
  return fhat;
}

void NFFT2d::setF( const Eigen::ArrayXcd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfft2dPlan_.f[i] << f(i);
}

void NFFT2d::setF( const Eigen::ArrayXd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfft2dPlan_.f[i] << std::complex<double> (f(i),0);
}

void NFFT2d::setFhat( const NFFT2dFourierCoefficients & fhat ) 
{   
  Eigen::Array2i degrees = fhat.getPolynomialDegrees();
  assert( degrees(0) <= polynomialDegrees_(0) );
  assert( degrees(1) <= polynomialDegrees_(1) );
  for (int k1=-polynomialDegrees_(0); k1<-polynomialDegrees_(0); ++k1)
    for (int k2=-polynomialDegrees_(1); k2<-polynomialDegrees_(1); ++k2)
      nfft2dPlan_.f_hat[linearizedIndex(k1,k2)] << 
	std::complex<double> (0,0);
  for (int k1=-degrees(0); k1<degrees(0); ++k1)
    for (int k2=-degrees(1); k2<degrees(1); ++k2)
      nfft2dPlan_.f_hat[linearizedIndex(k1,k2)] << fhat(k1, k2);
}

Eigen::ArrayXcd NFFT2d::computeYMatrixMultiplication( 
  const NFFT2dFourierCoefficients & fhat )
{
  setFhat( fhat );
  nfft_trafo( &nfft2dPlan_ );
  return getF();
}

Eigen::Array2Xcd NFFT2d::evalNablaYMatrixMultiplication( 
  const NFFT2dFourierCoefficients & fhat ) 
{
  Eigen::Array2i degrees = fhat.getPolynomialDegrees();
  assert( degrees(0) <= polynomialDegrees_(0) );
  assert( degrees(1) <= polynomialDegrees_(1) );
  // nabla_f = (f_alhpa1, f_alpha2) = NablaY * fhat
  // f_alpha1 = Y * (-2*pi*i*k1) * fhat;
  // f_alpha2 = Y * (-2*pi*i*k2) * fhat;
  Eigen::Array2Xcd nabla_f;
  nabla_f.resize(2, numberOfPoints_);
  NFFT2dFourierCoefficients temp( degrees );
  // f_alpha1
  for ( int k1=-degrees(0); k1<degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2<degrees(1); ++k2 )
      temp(k1,k2) = fhat(k1,k2) * std::complex<double> (0, -2*M_PI*k1);
  nabla_f.row(0) =  computeYMatrixMultiplication( temp ).transpose(); 
  // f_alpha2
  for ( int k1=-degrees(0); k1<degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2<degrees(1); ++k2 )
      temp(k1,k2) = fhat(k1,k2) * std::complex<double> (0, -2*M_PI*k2);
  nabla_f.row(1) = computeYMatrixMultiplication( temp ).transpose(); 
  
  return nabla_f;  
}

NFFT2dFourierCoefficients NFFT2d::evalNablaYMatrixAdjointMultiplication( 
  const Eigen::Array2Xd & f ) 
{
  assert( f.cols() == numberOfPoints_ );
  // fhat = NablaY' * (f_alhpa1; f_alpha2)   =  fhat_alpha1 + fhat_alpha2 
  // fhat_alpha1 = (2*pi*k1) * Y' * f_alpha1 
  // fhat_alpha2 = (2*pi*k2) * Y' * f_alpha2
  
  // fhat_alpha1
  NFFT2dFourierCoefficients fhat_alpha1 = 
    computeYMatrixAdjointMultiplication( (Eigen::ArrayXd) (f.row(0)), polynomialDegrees_ );
  // fhat_alpha2
  NFFT2dFourierCoefficients fhat_alpha2 = 
    computeYMatrixAdjointMultiplication( (Eigen::ArrayXd) (f.row(1)), polynomialDegrees_ );
  for ( int k1=-polynomialDegrees_(0); k1<polynomialDegrees_(0); ++k1 )
    for ( int k2=-polynomialDegrees_(1); k2<polynomialDegrees_(1); ++k2 ) {
      fhat_alpha1(k1,k2) *= std::complex<double> (0,2*M_PI*k1);
      fhat_alpha2(k1,k2) *= std::complex<double> (0,2*M_PI*k2);
    }
  return (fhat_alpha1 += fhat_alpha2); 
}

Eigen::Array3Xcd NFFT2d::evalNabla2YMatrixMultiplication( 
  const NFFT2dFourierCoefficients & fhat )
{
  Eigen::Array2i degrees = fhat.getPolynomialDegrees();
  assert( degrees(0) <= polynomialDegrees_(0) );
  assert( degrees(1) <= polynomialDegrees_(1) );
  // f = (f_alpha1_alpha1; f_alpha1_alpha2; f_alpha1_alpha2) = Nabla^2 Y * fhat
  // f_alpha1_alpha1 = Y * (- 4*pi^2*k1*k1) * fhat)
  // f_alpha1_alpha2 = Y * (- 4*pi^2*k1*k2) * fhat)
  // f_alpha2_alpha2 = Y * (- 4*pi^2*k2*k2) * fhat)
  Eigen::Array3Xcd nabla2_f;
  nabla2_f.resize(3, numberOfPoints_ );
  NFFT2dFourierCoefficients temp( degrees );

  // f_alpha1_alpha1
  for ( int k1=-degrees(0); k1<degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2<degrees(1); ++k2 )
      temp(k1,k2) = fhat(k1,k2) * std::complex<double> (-4*M_PI*M_PI*k1*k1,0);
  nabla2_f.row(0) = computeYMatrixMultiplication( temp ).transpose(); 
  // f_alpha1_alpha2
  for ( int k1=-degrees(0); k1<degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2<degrees(1); ++k2 )
      temp(k1,k2) = fhat(k1,k2) * std::complex<double> (-4*M_PI*M_PI*k1*k2,0);
  nabla2_f.row(1) = computeYMatrixMultiplication( temp ).transpose(); 
  // f_alpha2_alpha2
  for ( int k1=-degrees(0); k1<degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2<degrees(1); ++k2 )
      temp(k1,k2) = fhat(k1,k2) * std::complex<double> (-4*M_PI*M_PI*k2*k2,0);
  nabla2_f.row(2) = computeYMatrixMultiplication( temp ).transpose(); 

  return nabla2_f;
}

int NFFT2d::linearizedIndex( int k1, int k2 ) const
{
  return ((k1+polynomialDegrees_(0)) * 2*polynomialDegrees_(1) +
	  (k2+polynomialDegrees_(1)));
}

void NFFT2d::initNFFT2dPlan() 
{
  Eigen::Array2i dimensions_ = 2*polynomialDegrees_;
  Eigen::Array2i internalDimensions_ = 4*polynomialDegrees_;
  nfft_init_guru( &nfft2dPlan_, 2, dimensions_.data(), numberOfPoints_,
		  internalDimensions_.data(), 7,
		  PRE_PHI_HUT| PRE_FULL_PSI| MALLOC_F_HAT| MALLOC_X| MALLOC_F |
		  FFTW_INIT| FFT_OUT_OF_PLACE,
		  FFTW_ESTIMATE| FFTW_DESTROY_INPUT);
}

void NFFT2d::destroyNFFT2dPlan() 
{ 
  nfft_finalize( &nfft2dPlan_ );
}
