// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFFT2D_HPP_
#define _NFFT2D_HPP_

#include "NFFT2dFourierCoefficients.hpp"

namespace LORM {

  namespace NFFT {

    /** \class NFFT2d
     *
     *  \brief The wrapper class for the 2-dimensional nfft plan of the nfft C
     *  library. It provides, additionally to the evaluation of spherical
     *  harmonic expansions, the fast evaluation of the first and second order
     *  derivatives at nonequispaced points.
     *
     *  \note The class is non-copyable and non-copy-assignable. Hence,
     *  it cannot be passed nor returned by value.
     * 
     **/

    class NFFT2d {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degrees (N1,N2) of the 2-dimensions to set up the nfft plan.
      explicit NFFT2d( int numberOfPoints, Eigen::Array2i polynomialDegrees);
      /// Destructor
      ~NFFT2d();
      
      /// Returns the number of points M.
      int getNumberOfPoints() const { return numberOfPoints_; };
      /// Returns the polynomial degrees (N1,N2).
      Eigen::Array2i getPolynomialDegrees() const { return polynomialDegrees_; };

      /// Returns the evaluation points of the nfft in local coordinates in
      /// [-0.5, 0.5)^2.
      Eigen::Array2Xd getLocalPointCoordinatesArray() const;

      /// Sets the evaluation points of the nfft in local coordinates, which
      /// must be provided in [-0.5,0.5)^2.
      void setLocalPointCoordinatesArray( const Eigen::Array2Xd & localPointCoords );
	
      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfft.
      Eigen::ArrayXXcd evalYMatrixMultiplication( const NFFT2dFourierCoefficients & fhat ) {
	assert( (fhat.getPolynomialDegrees())(0) <= polynomialDegrees_(0) );
	assert( (fhat.getPolynomialDegrees())(1) <= polynomialDegrees_(1) );
	return computeYMatrixMultiplication( fhat ); };

      /// Sets the array f of function values of the nfft and returns the array
      /// fhat of Fourier coefficients after applying an adjoint nfft.
      template<typename ArrayType>
      NFFT2dFourierCoefficients evalYMatrixAdjointMultiplication( const ArrayType & f ) {
	return computeYMatrixAdjointMultiplication( f, polynomialDegrees_ ); }

      /// Returns the array of function values after multiplication with the
      /// matrix of the first order derivatives of the Fourier matrix.
      Eigen::Array2Xcd evalNablaYMatrixMultiplication( const NFFT2dFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint
      /// matrix of the first order derivatives of the Fourier matrix.
      NFFT2dFourierCoefficients evalNablaYMatrixAdjointMultiplication( const Eigen::Array2Xd & f );

      /// Returns the array of function values after multiplication with the
      /// matrix of the second order derivatives of the Fourier matrix.
      Eigen::Array3Xcd evalNabla2YMatrixMultiplication( const NFFT2dFourierCoefficients & fhat );      
      
    private:
      /// \note Non-copyable.
      NFFT2d ( const NFFT2d & ) {};
      /// \note Non-copy-assignable.
      const NFFT2d& operator= ( const NFFT2d & ) { return *this; };

      /// Evaluates the array f of function values of the nfft. 
      Eigen::ArrayXcd getF() const;

      /// Evaluates the array fhat of Fourier coefficients of the nfft up to
      /// the given polynomial degrees.
      NFFT2dFourierCoefficients getFhat( Eigen::Array2i degrees) const;

      /// Sets the array f of function values of the nfft.
      void setF( const Eigen::ArrayXcd & f );

      /// Sets the array f of function values of the nfft.
      void setF( const Eigen::ArrayXd & f );

      /// Sets the array fhat of Fourier coefficient of the nfft.
      void setFhat( const NFFT2dFourierCoefficients & fhat );

      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfft.
      Eigen::ArrayXcd computeYMatrixMultiplication( const NFFT2dFourierCoefficients & fhat );

      template<typename ArrayType>
      NFFT2dFourierCoefficients computeYMatrixAdjointMultiplication( 
	    const ArrayType & f, Eigen::Array2i degrees ) {
	setF( f );
     	nfft_adjoint( &nfft2dPlan_ );
     	return getFhat( degrees ); }

      /// Computes the linearized index for the nfft2d_plan
      int linearizedIndex( int k1, int k2 ) const;

      /// Initialization routine of the nfft2d_plan
      void initNFFT2dPlan();
      
      /// Destruction routine of the nfft2d_plan
      void destroyNFFT2dPlan();

      nfft_plan nfft2dPlan_;
      int numberOfPoints_;
      Eigen::Array2i polynomialDegrees_;
    };

  } // end namespace NFFT

} // end namespace LORM


#endif // _NFFT2D_HPP_
