// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "NFFT2dFourierCoefficients.hpp"

using namespace LORM::NFFT;


std::ostream & LORM::NFFT::operator<< (
  std::ostream& os, const NFFT2dFourierCoefficients & fhat )
{
  return os << fhat.getArray(); 
}

void NFFT2dFourierCoefficients::enlargeFhatArray( 
  const NFFT2dFourierCoefficients & other ) 
{
  int maxD1 = std::max( fhat_array_.rows(), other.fhat_array_.rows() );
  int maxD2 = std::max( fhat_array_.cols(), other.fhat_array_.cols() );
  Eigen::ArrayXXcd temp = fhat_array_;
  fhat_array_.setZero( maxD1, maxD2 );
  fhat_array_.block( fhat_array_.rows()/2 - temp.rows()/2,
		     fhat_array_.cols()/2 - temp.cols()/2,
		     temp.rows(), temp.cols() ) = temp;
}

const NFFT2dFourierCoefficients & NFFT2dFourierCoefficients::operator+=(
  const NFFT2dFourierCoefficients & other )
{
  if ( fhat_array_.rows() != other.fhat_array_.rows() ||
       fhat_array_.rows() != other.fhat_array_.rows() ) {
     
    enlargeFhatArray(other);
    Eigen::Array2i degrees = other.getPolynomialDegrees();
    for(int k1=-degrees(0); k1< degrees(0); ++k1)
      for( int k2=-degrees(1); k2< degrees(1); ++k2) 
	(*this)(k1,k2) +=  other(k1,k2);
  }
  else
    fhat_array_ += other.fhat_array_;

  return *this;
 }

const NFFT2dFourierCoefficients & NFFT2dFourierCoefficients::operator-=(
  const NFFT2dFourierCoefficients & other )
{
  if ( fhat_array_.rows() != other.fhat_array_.rows() ||
       fhat_array_.rows() != other.fhat_array_.rows() ) {
     
    enlargeFhatArray(other);
    Eigen::Array2i degrees = other.getPolynomialDegrees();
    for(int k1=-degrees(0); k1< degrees(0); ++k1)
      for( int k2=-degrees(1); k2< degrees(1); ++k2) 
	(*this)(k1,k2) -= other(k1,k2);
  }
  else
    fhat_array_ -= other.fhat_array_;

  return *this;
}

NFFT2dFourierCoefficients NFFT2dFourierCoefficients::operator* ( 
  const NFFT2dFourierCoefficients & other ) const
{
  Eigen::Array2i maxDims;
  maxDims << std::max(  fhat_array_.rows(), other.fhat_array_.rows() ), 
             std::max(  fhat_array_.cols(), other.fhat_array_.cols() ) ;
  NFFT2dFourierCoefficients mult( getPolynomialDegreesFromDimensions(maxDims) );
    
  if ( fhat_array_.rows() != other.fhat_array_.rows() || 
       fhat_array_.cols() != other.fhat_array_.cols() ) {
    Eigen::Array2i minDegrees;
    minDegrees << std::min(  fhat_array_.rows(), other.fhat_array_.rows() )/2, 
                  std::min(  fhat_array_.cols(), other.fhat_array_.cols() )/2 ;
    for(int k1=-minDegrees(0); k1< minDegrees(0); ++k1)
      for( int k2=-minDegrees(1); k2< minDegrees(1); ++k2) 
	mult(k1,k2) = (*this)(k1,k2) * other(k1,k2);
  }
  else
    mult.fhat_array_ = fhat_array_ * other.fhat_array_;
  
  return mult;
}
