// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFFT2D_FOURIER_COEFFICIENTS_HPP_
#define _NFFT2D_FOURIER_COEFFICIENTS_HPP_

#include "NFFTutils.hpp"

namespace LORM {

  namespace NFFT {

    /** \class NFFT2dFourierCoefficients 
     *
     *  \brief A utility class for the representation of the array of Fourier
     *  coefficients used in the class NFFT2d.
     *
     **/
    
    class NFFT2dFourierCoefficients {
    public:
      /// Default-Constructor.
      explicit NFFT2dFourierCoefficients() :
	fhat_array_ () {};
      /// Constructor. One must provide the polynomial degree and all Fourier
      /// coefficients are initialized to zero.
      explicit NFFT2dFourierCoefficients( Eigen::Array2i degrees ) :
	fhat_array_ () { setZero( degrees ); };
      /// Constructor. One must provide an array of Fourier coefficients. \note
      /// The number of rows and columns of the array must coincide with the dimension of
      /// some polynomial space, i.e. they must be even.
      explicit NFFT2dFourierCoefficients( const Eigen::ArrayXXcd & array ) :
	fhat_array_ () { setArray( array ); };

      /// Returns the array of the Fourier coefficients.
      const Eigen::ArrayXXcd & getArray() const { return fhat_array_; };

      /// Returns the polynomial degree.
      Eigen::Array2i getPolynomialDegrees() const { 
	Eigen::Array2i dims;
	dims << fhat_array_.rows(), fhat_array_.cols();
	return getPolynomialDegreesFromDimensions( dims ); };
      
      /// Returns the dimension of the tensor product polynomial space.
      Eigen::Array2i getDimensions() const {
	Eigen::Array2i dims;
	dims << fhat_array_.rows(), fhat_array_.cols();
	return dims; };

      /// Sets the Array of Fourier coefficients
      void setArray( const Eigen::ArrayXXcd & array ) {
	Eigen::Array2i dims;
	dims << fhat_array_.rows(), fhat_array_.cols();
	getPolynomialDegreesFromDimensions( dims );
	fhat_array_ = array; };

      /// Sets the polyomial degrees and initializes the Fourier coefficients to zero.
      void setZero( Eigen::Array2i degrees ) {
	fhat_array_.setZero( getDimensionsFromPolynomialDegrees( degrees )(0),
			     getDimensionsFromPolynomialDegrees( degrees )(1) ); };

      /// Sets the polyomial degrees and initializes the Fourier coefficients to one.
      void setOne( Eigen::Array2i degrees ) {
	fhat_array_.setOnes( getDimensionsFromPolynomialDegrees( degrees )(0),
			     getDimensionsFromPolynomialDegrees( degrees )(1) ); };
      
      /// Index-Operator for assigning the Fourier coefficient f_(k1,k2).
      std::complex<double> & operator() ( int k1, int k2 ) {
	return fhat_array_( k1 + fhat_array_.rows()/2, k2 + fhat_array_.cols()/2 ); };

      /// Index-Operator for accessing the Fourier coefficient f_(k1,k2).
      const std::complex<double> & operator() ( int k1, int k2 ) const {
	return fhat_array_( k1 + fhat_array_.rows()/2, k2 + fhat_array_.cols()/2 ); };

      /// Addition-Assignment-Operator, which simply adds the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFFT2dFourierCoefficients & operator += ( const NFFT2dFourierCoefficients & other );

      /// Subtraction-Assignment-Operator, which simply subtracts the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFFT2dFourierCoefficients & operator -= ( const NFFT2dFourierCoefficients & other );

      /// Multiplication-Operator. \note The polynomial degrees of the given
      /// Fourier need not to be equal, since they are resized accordingly.
      NFFT2dFourierCoefficients operator* ( const NFFT2dFourierCoefficients & other ) const;
 
    private:
      /// Returns the dimensions of the tensor product polynomial space of given degree.
      static Eigen::Array2i getDimensionsFromPolynomialDegrees( Eigen::Array2i degrees ) {
	Eigen::Array2i dims;
	dims << 2*degrees(0), 2*degrees(1);
	return dims; };

      /// Returns the polynomial degrees from the dimension of the tensor
      /// product polynomial space.
      static Eigen::Array2i getPolynomialDegreesFromDimensions( Eigen::Array2i dims ) {
      	assert( dims(0) % 2 ==  0 );
      	assert( dims(1) % 2 ==  0 );
   	return dims/2; };

      void enlargeFhatArray( const NFFT2dFourierCoefficients & fhat );

      Eigen::ArrayXXcd fhat_array_;
    };

    std::ostream & operator<< (std::ostream& os, const NFFT2dFourierCoefficients & fhat );

  } // end namespac NFFT

} // end namespace LORM

#endif // _NFFT2D_FOURIER_COEFFICIENTS_HPP_
