// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "NFSFT.hpp"

using namespace LORM::NFFT;

NFSFT::NFSFT( int numberOfPoints, int polynomialDegree) :
  numberOfPoints_ (numberOfPoints),
  polynomialDegree_ (polynomialDegree),
  internalPolynomialDegree_ (polynomialDegree + 2),
  oneOverSinThetaArray_ () 
{ 
  initNFSFTPlan(); 
  setLocalPointCoordinatesArray( Eigen::Array2Xd::Zero(2,numberOfPoints) );
}

NFSFT::~NFSFT() 
{
  destroyNFSFTPlan();
}

Eigen::Array2Xd NFSFT::getLocalPointCoordinatesArray() const 
{
  Eigen::Array2Xd localPointCoords;
  localPointCoords.resize( 2, numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    localPointCoords(1,i) = (2*M_PI) * nfsftPlan_.x[2*i]; // phi coordinates of the nfsft are in [0,1]
    localPointCoords(0,i) = (2*M_PI) * nfsftPlan_.x[2*i+1]; // theta coordinates in of the nfsft are in [0,1]
  }
  return localPointCoords;
}

void NFSFT::setLocalPointCoordinatesArray( const Eigen::Array2Xd & localPointCoords ) 
{
  assert( localPointCoords.cols() == numberOfPoints_ );
  oneOverSinThetaArray_.resize( numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    nfsftPlan_.x[2*i] = localPointCoords(1,i) / (2*M_PI); // we need to set phi coordinates in [0,1]
    nfsftPlan_.x[2*i+1] = localPointCoords(0,i) / (2*M_PI); // we need to se theta coordinates in [0,1]
    if (localPointCoords(0,i)!=0 )
      oneOverSinThetaArray_(i) = 1. / sin( localPointCoords(0,i) );
    else
      oneOverSinThetaArray_(i) = 0;
  }
  nfsft_precompute_x( &nfsftPlan_ );
}

Eigen::Array2Xcd NFSFT::evalNablaYMatrixMultiplication( const NFSFTFourierCoefficients & fhat ) 
{
  assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
  // nabla_f = (f_theta; f_phi) = NablaY * fhat
  // f_theta = Y_theta * fhat  = 1/sin(theta) *  Y * Dtheta * fhat
  // f_phi = Y_phi * fhat = Y * Dphi * fhat
  Eigen::Array2Xcd nabla_f;
  nabla_f.resize(2, numberOfPoints_);
  // f_theta
  nabla_f.row(0) = ( computeYMatrixMultiplication( computeDthetaMatrixMultiplication(fhat) ) 
		     * oneOverSinThetaArray_).transpose();
  // f_phi
  nabla_f.row(1) =  computeYMatrixMultiplication( computeDphiMatrixMultiplication(fhat) ).transpose(); 
  return nabla_f;
}

NFSFTFourierCoefficients NFSFT::evalNablaYMatrixAdjointMultiplication( const Eigen::Array2Xd & f ) 
{
  assert( f.cols() == numberOfPoints_ );
  // fhat = NablaY' * (f_theta; f_phi)   =  fhat_theta + fhat_phi 
  // fhat_theta = Y' * Dtheta' * 1 / sin(theta) * f_theta 
  // fhat_phi = Y' * Dphi' * f_phi
  
  // fhat_theta
  NFSFTFourierCoefficients fhat_mult =  computeDthetaMatrixAdjointMultiplication(
         			        computeYMatrixAdjointMultiplication( 
           (Eigen::ArrayXd) (f.row(0) * oneOverSinThetaArray_.transpose()), polynomialDegree_+1 ) );
  // += fhat_phi
  fhat_mult += computeDphiMatrixAdjointMultiplication(
               computeYMatrixAdjointMultiplication( (Eigen::ArrayXd) (f.row(1)), polynomialDegree_ ) );
  return fhat_mult; 
}

Eigen::Array3Xcd NFSFT::evalNabla2YMatrixMultiplication( const NFSFTFourierCoefficients & fhat )
{
  assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
  // f = (f_theta_theta; f_theta_phi; f_phi_phi) = Nabla^2 Y * fhat
  // f_theta_theta = Y_theta_theta * fhat 
  //               = 1 / Sin^2 * Y * Dtheta * Dtheta - Cos * Y *Dtheta * fhat)
  // f_theta_phi = Y_theta_phi * fhat = 1/Sin * Y * Dtheta * Dphi * fhat
  // f_phi_phi = Y_phi_phi * fhat = Y * Dphi * Dphi * fhat
  Eigen::Array3Xcd nabla2_f;
  nabla2_f.resize(3, numberOfPoints_ );
  // f_theta_theta
  nabla2_f.row(0) = ( computeYMatrixMultiplication(
   		      computeDthetaMatrixMultiplication(
		      computeDthetaMatrixMultiplication(fhat) ) ) ).transpose();
  nabla2_f.row(0) -= ( computeYMatrixMultiplication(
		       computeDthetaMatrixMultiplication(fhat) ) ).transpose() 
	             * getLocalPointCoordinatesArray().row(0).cos();
  nabla2_f.row(0) *= oneOverSinThetaArray_.square().transpose();
  // f_theta_phi
  nabla2_f.row(1) = (( computeYMatrixMultiplication(
  		       computeDthetaMatrixMultiplication(
		       computeDphiMatrixMultiplication(fhat) ) ) )
		     * oneOverSinThetaArray_).transpose();
  // f_phi_phi
  nabla2_f.row(2) = ( computeYMatrixMultiplication( 
     	              computeDphiMatrixMultiplication(
		      computeDphiMatrixMultiplication(fhat) ) ) ).transpose();
  return nabla2_f;
}

Eigen::ArrayXcd NFSFT::getF() const 
{
  Eigen::ArrayXcd f;
  f.resize( numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) 
    f(i) << nfsftPlan_.f[i];
  return f;
}

NFSFTFourierCoefficients NFSFT::getFhat( int degree) const
{
  assert( degree <= internalPolynomialDegree_ );
  NFSFTFourierCoefficients fhat( degree );
  for (int k=-degree; k<=degree; ++k)
    for (int n=abs(k); n<=degree; ++n)
      fhat(n, k) <<  nfsftPlan_.f_hat[NFSFT_INDEX(n,k,&nfsftPlan_)];
  return fhat;
}

void NFSFT::setF( const Eigen::ArrayXcd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfsftPlan_.f[i] << f(i);
}

void NFSFT::setF( const Eigen::ArrayXd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfsftPlan_.f[i] << std::complex<double> (f(i),0);
}

void NFSFT::setFhat( const NFSFTFourierCoefficients & fhat ) 
{   
  int degree = fhat.getPolynomialDegree();
  assert( degree <= internalPolynomialDegree_ );
  for (int k=-degree; k<=degree; ++k)
    for (int n=abs(k); n<=degree; ++n)
      nfsftPlan_.f_hat[NFSFT_INDEX(n,k,&nfsftPlan_)] << fhat(n, k);
  for (int n=degree + 1; n<= internalPolynomialDegree_; ++n)
    for( int k=-n; k<=n; ++k )
      nfsftPlan_.f_hat[NFSFT_INDEX(n,k,&nfsftPlan_)] << std::complex<double> (0,0);
}

Eigen::ArrayXcd NFSFT::computeYMatrixMultiplication( const NFSFTFourierCoefficients & fhat )
{
  setFhat( fhat );
  nfsft_trafo( &nfsftPlan_ );
  return getF();
}

NFSFTFourierCoefficients NFSFT::computeDthetaMatrixMultiplication( 
  const NFSFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSFTFourierCoefficients fhat_mult( degree + 1 );
  for (int n = 0; n <= degree; n++) 
    for (int k = -n; k <= n; k++) {
      fhat_mult(n+1,k) =  coef_a(n,k) * fhat(n,k);
      if ( abs(k) <= (n-1))
	fhat_mult(n-1,k) -= coef_b(n,k) * fhat(n,k); 
    }
  return fhat_mult;
}

NFSFTFourierCoefficients NFSFT::computeDthetaMatrixAdjointMultiplication( 
  const NFSFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSFTFourierCoefficients fhat_mult( degree-1 );
  for (int n = 0; n <= degree-1; n++) 
    for (int k = -n; k <= n; k++) {
      fhat_mult(n,k) =  coef_a(n,k) * fhat(n+1,k);
      if ( abs(k) <= (n-1))
	fhat_mult(n,k) -=  coef_b(n,k) * fhat(n-1,k);
    }
  return fhat_mult;
}

NFSFTFourierCoefficients NFSFT::computeDphiMatrixMultiplication( 
  const NFSFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSFTFourierCoefficients fhat_mult( degree );
  for (int n = 0; n <= degree; n++) 
    for (int k = -n; k <= n; k++)
      fhat_mult(n,k) = fhat(n,k) * (std::complex<double> (0.,k)); 
  return fhat_mult;
}

NFSFTFourierCoefficients NFSFT::computeDphiMatrixAdjointMultiplication( 
  const NFSFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSFTFourierCoefficients fhat_mult( degree );
  for (int n = 0; n <= degree; n++) 
    for (int k = -n; k <= n; k++)
      fhat_mult(n,k) = fhat(n,k) * (std::complex<double> (0.,-k)); 
  return fhat_mult;
}

void NFSFT::initNFSFTPlan() 
{
  nfsft_precompute( internalPolynomialDegree_, 1000.0, 0U, 0U);
  nfsft_init_guru( &nfsftPlan_, internalPolynomialDegree_, numberOfPoints_,
		   NFSFT_MALLOC_X | NFSFT_MALLOC_F | NFSFT_MALLOC_F_HAT | 
		   NFSFT_NORMALIZED | NFSFT_PRESERVE_F_HAT,
		   PRE_PHI_HUT | PRE_PSI | FFTW_INIT | FFT_OUT_OF_PLACE, 7);
}

void NFSFT::destroyNFSFTPlan() 
{ 
  nfsft_finalize( &nfsftPlan_ );
  nfsft_forget();	
}
