// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFSFT_HPP_
#define _NFSFT_HPP_

#include "NFSFTFourierCoefficients.hpp"

namespace LORM {

  namespace NFFT {

    /** \class NFSFT 
     *
     *  \brief The wrapper class for the nfsft plan of the nfft C library. It
     *  provides, additionally to the evaluation of spherical harmonic
     *  expansions, the fast evaluation of the first and second order
     *  derivatives at nonequispaced points.
     *
     *  \note The class is non-copyable and non-copy-assignable. Hence,
     *  it cannot be passed nor returned by value.
     * 
     **/

    class NFSFT {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N to set up the nfsft plan.
      explicit NFSFT( int numberOfPoints, int polynomialDegree);
      /// Destructor
      ~NFSFT();
      
      /// Returns the number of points M.
      int getNumberOfPoints() const { return numberOfPoints_; };
      /// Returns the polynomial degree N.
      int getPolynomialDegree() const { return polynomialDegree_; };

      /// Returns the evaluation points of the nfsft in spherical coordinates
      /// (theta,phi) in [0,pi)x[0,2*pi).
      Eigen::Array2Xd getLocalPointCoordinatesArray() const;

      /// Sets the evaluation points of the nfsft in spherical coordinates
      /// (theta,phi) in [0,pi)x[0,2*pi).
      void setLocalPointCoordinatesArray( const Eigen::Array2Xd & localPointCoords );
	
      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfsft.
      Eigen::ArrayXcd evalYMatrixMultiplication( const NFSFTFourierCoefficients & fhat ) {
	assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
	return computeYMatrixMultiplication( fhat ); };

      /// Sets the array f of function values of the nfsft and returns the array
      /// fhat of Fourier coefficients after applying an adjoint nfsft.
      template<typename ArrayType>
      NFSFTFourierCoefficients evalYMatrixAdjointMultiplication( const ArrayType & f ) {
	return computeYMatrixAdjointMultiplication( f, polynomialDegree_ ); }

      /// Returns the array of function values after multiplication with the
      /// matrix of the first order derivatives of the Fourier matrix.
      Eigen::Array2Xcd evalNablaYMatrixMultiplication( const NFSFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint
      /// matrix of the first order derivatives of the Fourier matrix.
      NFSFTFourierCoefficients evalNablaYMatrixAdjointMultiplication( const Eigen::Array2Xd & f );

      /// Returns the array of function values after multiplication with the
      /// matrix of the second order derivatives of the Fourier matrix.
      Eigen::Array3Xcd evalNabla2YMatrixMultiplication( const NFSFTFourierCoefficients & fhat );      
      
    private:
      /// \note Non-copyable.
      NFSFT ( const NFSFT & ) {};
      /// \note Non-copy-assignable.
      const NFSFT& operator= ( const NFSFT & ) { return *this; };

      /// Evaluates the array f of function values of the nfsft. 
      Eigen::ArrayXcd getF() const;

      /// Evaluates the array fhat of Fourier coefficients of the nfsft up to
      /// the given polynomial degree.
      NFSFTFourierCoefficients getFhat( int degree) const;

      /// Sets the array f of function values of the nfsft.
      void setF( const Eigen::ArrayXcd & f );

      /// Sets the array f of function values of the nfsft.
      void setF( const Eigen::ArrayXd & f );

      /// Sets the array fhat of Fourier coefficient of the nfsft.
      void setFhat( const NFSFTFourierCoefficients & fhat );

      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfsft.
      Eigen::ArrayXcd computeYMatrixMultiplication( const NFSFTFourierCoefficients & fhat );

      template<typename ArrayType>
      NFSFTFourierCoefficients computeYMatrixAdjointMultiplication( const ArrayType & f, int degree ) {
	setF( f );
     	nfsft_adjoint( &nfsftPlan_ );
     	return getFhat( degree ); }

      /// Helper functions for the computation of the coefficients of the diagonal matrix Dtheta.
      static double coef_a ( int n, int k) { 
	return n * sqrt( (pow(n+1,2) - pow(k,2)) / ((2.*n+1.) * (2.*n+3.)) );}; 
      static double coef_b (int n, int k) { 
	return (n+1) * sqrt( (pow(n,2) - pow(k,2)) / ((2.*n-1.) * (2.*n+1.)) );};
      
      /// Returns the Fourier coefficients after multiplication with the matrix D_theta.
      /// \note The polynomial degree of the returned Fourier coefficients is increased by 1.
      static NFSFTFourierCoefficients computeDthetaMatrixMultiplication( 
        const NFSFTFourierCoefficients & fhat );     

      /// Returns the Fourier coefficients after multiplication with the adjoint matrix of D_theta.
      /// \note The polynomial degree of the returned Fourier coefficients is decreased by 1.
      static NFSFTFourierCoefficients computeDthetaMatrixAdjointMultiplication( 
        const NFSFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the matrix D_phi.
      static NFSFTFourierCoefficients computeDphiMatrixMultiplication( 
        const NFSFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint matrix of D_phi.
      static NFSFTFourierCoefficients computeDphiMatrixAdjointMultiplication( 
        const NFSFTFourierCoefficients & fhat );
    
      /// Initialization routine of the nfsft_plan
      void initNFSFTPlan();
      
      /// Destruction routine of the nfsft_plan
      void destroyNFSFTPlan();

      nfsft_plan nfsftPlan_;
      int numberOfPoints_;
      int polynomialDegree_;
      int internalPolynomialDegree_;

      Eigen::ArrayXd oneOverSinThetaArray_;
    };


  } // end namespace NFFT

} // end namespace LORM


#endif // _NFSFT_HPP_
