// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFSFT_FOURIER_COEFFICIENTS_HPP_
#define _NFSFT_FOURIER_COEFFICIENTS_HPP_

#include "NFFTutils.hpp"

namespace LORM {

  namespace NFFT {

    /** \class NFSFTFourierCoefficients 
     *
     *  \brief A utility class for the representation of the array of Fourier
     *  coefficients used in the class NFSFT.
     *
     **/
    
    class NFSFTFourierCoefficients {
    public:
      /// Default-Constructor.
      explicit NFSFTFourierCoefficients() :
	fhat_array_ () {};
      /// Constructor. One must provide the polynomial degree and all Fourier
      /// coefficients are initialized to zero.
      explicit NFSFTFourierCoefficients( int degree ) :
	fhat_array_ () { setZero( degree ); };
      /// Constructor. One must provide an array of Fourier coefficients. \note
      /// The number of rows of the array must coincide with the dimension of
      /// some polynomial space.
      explicit NFSFTFourierCoefficients( const Eigen::ArrayXcd & array ) :
	fhat_array_ () { setArray( array ); };

      /// Returns the array of the Fourier coefficients.
      const Eigen::ArrayXcd & getArray() const { return fhat_array_; };

      /// Returns the polynomial degree.
      int getPolynomialDegree() const { 
	return getPolynomialDegreeFromDimension( fhat_array_.rows() ); };
      
      /// Returns the dimension of the polynomial space.
      int getDimension() const {
	return fhat_array_.rows(); };

      /// Sets the Array of Fourier coefficients
      void setArray( const Eigen::ArrayXcd & array ) {
	getPolynomialDegreeFromDimension( array.rows() );
	fhat_array_ = array; };

      /// Sets the polyomial degree and initializes the Fourier coefficients to zero.
      void setZero( int degree ) {
	fhat_array_.setZero( getDimensionFromPolynomialDegree( degree ) ); };

      /// Sets the polyomial degree and initializes the Fourier coefficients to one.
      void setOne( int degree ) {
	fhat_array_.setOnes( getDimensionFromPolynomialDegree( degree ) ); };
      
      /// Index-Operator for assigning the Fourier coefficient f_(n,k).
      std::complex<double> & operator() (int n, int k) {
	return fhat_array_(getIndex(n,k)); };

      /// Index-Operator for accessing the Fourier coefficient f_(n,k).
      const std::complex<double> & operator() (int n, int k) const {
	return fhat_array_(getIndex(n,k)); };

      /// Addition-Assignment-Operator, which simply adds the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFSFTFourierCoefficients & operator += ( const NFSFTFourierCoefficients & other );

      /// Subtraction-Assignment-Operator, which simply subtracts the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFSFTFourierCoefficients & operator -= ( const NFSFTFourierCoefficients & other );

      /// Multiplication-Operator. \note The polynomial degrees of the given
      /// Fourier need not to be equal, since they are resized accordingly.
      NFSFTFourierCoefficients operator* ( const NFSFTFourierCoefficients & other ) const;
 
    private:
      /// Returns the dimension of the polynomial spaces of given degree.
      static int getDimensionFromPolynomialDegree( int degree ) {
	return (degree + 1) * (degree + 1); };

      /// Returns the polynomial degree from the dimension of the polynomial spaces.
      static int getPolynomialDegreeFromDimension( int dim ) {
      	int degree = sqrt( dim - 1 );
	assert( dim == getDimensionFromPolynomialDegree( degree ) );
	return degree; };

      /// Returns the linearized index of the (n,k)th Fourier coefficient.
      static int getIndex( int n, int k ) { return (n+1)*n + k; };

      Eigen::ArrayXcd fhat_array_;
    };

    std::ostream & operator<< (std::ostream& os, const NFSFTFourierCoefficients & fhat );

  } // end namespac NFFT

} // end namespace LORM

#endif // _NFSFT_FOURIER_COEFFICIENTS_HPP_
