// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "NFSOFT.hpp"

using namespace LORM::NFFT;

NFSOFT::NFSOFT( int numberOfPoints, int polynomialDegree) :
  numberOfPoints_ (numberOfPoints),
  polynomialDegree_ (polynomialDegree),
  internalPolynomialDegree_ (polynomialDegree + 2),
  oneOverSinThetaArray_ () 
{ 
  initNFSOFTPlan(); 
  setLocalPointCoordinatesArray( Eigen::Array3Xd::Zero(3,numberOfPoints) );
}

NFSOFT::~NFSOFT()
{
  destroyNFSOFTPlan();
}

Eigen::Array3Xd NFSOFT::getLocalPointCoordinatesArray() const 
{
  Eigen::Array3Xd localPointCoords;
  localPointCoords.resize( 3, numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    localPointCoords(0,i) = nfsoftPlan_.x[3*i]; 
    localPointCoords(1,i) = nfsoftPlan_.x[3*i+1];
    localPointCoords(2,i) = nfsoftPlan_.x[3*i+2];
  }
  return localPointCoords;
}

void NFSOFT::setLocalPointCoordinatesArray( const Eigen::Array3Xd & localPointCoords ) 
{
  assert( localPointCoords.cols() == numberOfPoints_ );
  oneOverSinThetaArray_.resize( numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) {
    nfsoftPlan_.x[3*i] = localPointCoords(0,i); 
    nfsoftPlan_.x[3*i+1] = localPointCoords(1,i);
    nfsoftPlan_.x[3*i+2] = localPointCoords(2,i); 
    if (localPointCoords(1,i)!=0 )
      oneOverSinThetaArray_(i) = 1. / sin( localPointCoords(1,i) );
    else
      oneOverSinThetaArray_(i) = 0;
  }
  nfsoft_precompute( &nfsoftPlan_ );
}

Eigen::Array3Xcd NFSOFT::evalNablaYMatrixMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
  // nabla_f = (f_phi1; f_theta; f_phi2) = NablaY * fhat
  // f_phi1 =  Y_phi1 * fhat = Y * Dphi1 * fhat
  // f_theta = Y_theta * fhat  = 1/sin(theta) *  Y Dtheta * fhat
  // f_phi2 =  Y_phi2 * fhat = Y * Dphi2 * fhat
  Eigen::Array3Xcd nabla_f;
  nabla_f.resize(3, numberOfPoints_);
  // f_phi1
  nabla_f.row(0) = computeYMatrixMultiplication( computeDphi1MatrixMultiplication(fhat) ).transpose(); 
  // f_theta
  nabla_f.row(1) = ( computeYMatrixMultiplication( computeDthetaMatrixMultiplication(fhat) ) 
		   * oneOverSinThetaArray_).transpose();
  // f_phi2
  nabla_f.row(2) = computeYMatrixMultiplication( computeDphi2MatrixMultiplication(fhat) ).transpose(); 
  return nabla_f;
}

NFSOFTFourierCoefficients NFSOFT::evalNablaYMatrixAdjointMultiplication( const Eigen::Array3Xd & f ) 
{
  assert( f.cols() == numberOfPoints_ );
  // fhat = Nabla Y' * (f_phi1; f_theta; f_phi2)   =  fhat_phi1 + fhat_theta + fhat_phi2 
  // fhat_phi1 = Y' * Dphi1 * f_phi1
  // fhat_theta = Y' * Dtheta' * 1 / sin(theta) * f_theta 
  // fhat_phi2 = Y' * Dphi2' * f_phi2
  
  // fhat_phi1
  NFSOFTFourierCoefficients fhat_mult = computeDphi1MatrixAdjointMultiplication(
				        computeYMatrixAdjointMultiplication( 
                       (Eigen::ArrayXd) (f.row(0)), polynomialDegree_ ) );
  // += fhat_theta
  fhat_mult += computeDthetaMatrixAdjointMultiplication(
	       computeYMatrixAdjointMultiplication( (Eigen::ArrayXd) 
	       (f.row(1) * oneOverSinThetaArray_.transpose()), polynomialDegree_+1 ) );
  // += fhat_phi2
  fhat_mult += computeDphi2MatrixAdjointMultiplication(
     	       computeYMatrixAdjointMultiplication( (Eigen::ArrayXd) 
 	       (f.row(2)), polynomialDegree_ ) );
  return fhat_mult; 
}

Eigen::Array6Xcd NFSOFT::evalNabla2YMatrixMultiplication( const NFSOFTFourierCoefficients & fhat ) 
{
  assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
  // f = (f_phi1_phi1; f_phi1_theta; f_phi1_phi2; f_theta_theta; f_theta_phi2; f_phi2_phi2) 
  //   = HY * fhat
  
  // f_phi1_phi1   = Y_phi1_phi1   * fhat  = Y * Dphi1  * Dphi1 * fhat
  // f_phi1_theta  = Y_phi1_theta  * fhat  = 1/Sin * Y * Dtheta * Dphi1 * fhat
  // f_phi1_phi2   = Y_phi1_phi2   * fhat  = Y * Dphi1  * Dphi2 * fhat    
  // f_theta_theta = Y_theta_theta * fhat  
  //               = 1 / Sin^2 * Y * Dtheta * Dtheta - Cos * Y * Dtheta * fhat
  // f_theta_phi2  = Y_theta_phi2 * fhat   = 1/Sin * Y * Dtheta * Dphi2 * fhat
  // f_phi2_phi2   = Y_phi2_phi2  * fhat   = Y * Dphi2 * Dphi2 * fhat

	
  Eigen::Array6Xcd nabla2_f;
  nabla2_f.resize(6, numberOfPoints_ );
  // f_phi1_phi1
  nabla2_f.row(0) = ( computeYMatrixMultiplication( 
         	      computeDphi1MatrixMultiplication(
       		      computeDphi1MatrixMultiplication(fhat) ) ) ).transpose();
  // f_phi1_theta 
  nabla2_f.row(1) = (( computeYMatrixMultiplication(
       		       computeDthetaMatrixMultiplication(
       		       computeDphi1MatrixMultiplication(fhat) ) ) )
       		     * oneOverSinThetaArray_).transpose();
  // f_phi1_phi2
  nabla2_f.row(2) = ( computeYMatrixMultiplication( 
       	   	      computeDphi1MatrixMultiplication(
       		      computeDphi2MatrixMultiplication(fhat) ) ) ).transpose();
  // f_theta_theta
  nabla2_f.row(3) = ( computeYMatrixMultiplication(
         	      computeDthetaMatrixMultiplication(
       		      computeDthetaMatrixMultiplication(fhat) ) ) ).transpose();
  nabla2_f.row(3) -= ( computeYMatrixMultiplication(
       		       computeDthetaMatrixMultiplication(fhat) ) ).transpose() 
       	             * getLocalPointCoordinatesArray().row(1).cos();
  nabla2_f.row(3) *= oneOverSinThetaArray_.square().transpose();
  // f_theta_phi2
  nabla2_f.row(4) = (( computeYMatrixMultiplication(
       		       computeDthetaMatrixMultiplication(
       		       computeDphi2MatrixMultiplication(fhat) ) ) )
       		     * oneOverSinThetaArray_).transpose();
  // f_phi2_phi2
  nabla2_f.row(5) = ( computeYMatrixMultiplication( 
      	              computeDphi2MatrixMultiplication(
      	     	      computeDphi2MatrixMultiplication(fhat) ) ) ).transpose();
  return nabla2_f;
}

Eigen::ArrayXcd NFSOFT::getF() const
{
  Eigen::ArrayXcd f;
  f.resize( numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i) 
    f(i) << nfsoftPlan_.f[i];
  return f;
}

NFSOFTFourierCoefficients NFSOFT::getFhat( int degree) const 
{
  assert( degree <= internalPolynomialDegree_ );
  NFSOFTFourierCoefficients fhat( degree );
  int i=0;
  for (int k2=-internalPolynomialDegree_; k2<=internalPolynomialDegree_; ++k2)
    for (int k1=-internalPolynomialDegree_; k1<=internalPolynomialDegree_; ++k1) {
      int max=(abs(k1)>abs(k2)?abs(k1):abs(k2));
      for (int n=max; n<= internalPolynomialDegree_; ++n) {
	if (n<=degree)
	  fhat(n,k1,k2) << nfsoftPlan_.f_hat[i];
	i++;
      }
    }
  return fhat;
}

void NFSOFT::setF( const Eigen::ArrayXcd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfsoftPlan_.f[i] << f(i);
}

void NFSOFT::setF( const Eigen::ArrayXd & f ) 
{
  assert( f.rows() == numberOfPoints_ );
  for (int i=0; i<numberOfPoints_; ++i)
    nfsoftPlan_.f[i] << std::complex<double> (f(i),0);
}

void NFSOFT::setFhat( const NFSOFTFourierCoefficients & fhat ) 
{   
  int degree = fhat.getPolynomialDegree();
  assert( degree <= internalPolynomialDegree_ );
  int i=0;
  for (int k2=-internalPolynomialDegree_; k2<=internalPolynomialDegree_; ++k2)
    for (int k1=-internalPolynomialDegree_; k1<=internalPolynomialDegree_; ++k1) {
      int max=(abs(k1)>abs(k2)?abs(k1):abs(k2));
      for (int n=max; n<= internalPolynomialDegree_; ++n) {
	if (n<=degree)
	  nfsoftPlan_.f_hat[i] << fhat(n,k1,k2);
	else
	  nfsoftPlan_.f_hat[i] << std::complex<double> (0,0);
	i++;
      }
    }
}

NFSOFTFourierCoefficients NFSOFT::computeDthetaMatrixMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSOFTFourierCoefficients fhat_mult( degree + 1 );
  for (int n = 0; n <= degree; n++) 
    for (int k1 = -n; k1 <= n; k1++) 
      for (int k2 = -n; k2 <= n; k2++) {
	fhat_mult(n+1,k1,k2) += coef_a(n,k1,k2) * fhat(n,k1,k2);
	if (n > 0)
		fhat_mult(n,k1,k2) -= coef_b(n,k1,k2) * fhat(n,k1,k2);
	if ( abs(k1) <= (n-1) && abs(k2) <= (n-1) ) 
	  fhat_mult(n-1,k1,k2) -= coef_c(n,k1,k2) * fhat(n,k1,k2);	       
      }
  return fhat_mult;
}

NFSOFTFourierCoefficients NFSOFT::computeDthetaMatrixAdjointMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSOFTFourierCoefficients fhat_mult( degree-1 );
  for (int n = 0; n <= degree-1; n++) 
    for (int k1 = -n; k1 <= n; k1++) 
      for (int k2 = -n; k2 <= n; k2++) {
	fhat_mult(n,k1,k2) += coef_a(n,k1,k2) * fhat(n+1,k1,k2);
	if (n>0)
	  fhat_mult(n,k1,k2) -= coef_b(n,k1,k2) * fhat(n,k1,k2);
	if ( abs(k1) <= (n-1) && abs(k2) <= (n-1) )
	  fhat_mult(n,k1,k2) -= coef_c(n,k1,k2) * fhat(n-1,k1,k2);
      }
  return fhat_mult;
}

NFSOFTFourierCoefficients NFSOFT::computeDphi1MatrixMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSOFTFourierCoefficients fhat_mult( degree );
  for (int n = 0; n <= degree; ++n) 
    for (int k1 = -n; k1 <= n; ++k1)
      for (int k2 = -n; k2 <= n; ++k2)
	fhat_mult(n,k1,k2) = fhat(n,k1,k2) * (std::complex<double> (0.,-k1));  
  return fhat_mult;
}

NFSOFTFourierCoefficients NFSOFT::computeDphi1MatrixAdjointMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSOFTFourierCoefficients fhat_mult( degree );
  for (int n = 0; n <= degree; ++n) 
    for (int k1 = -n; k1 <= n; ++k1)
      for (int k2 = -n; k2 <= n; ++k2)
	      fhat_mult(n,k1,k2) = fhat(n,k1,k2) * (std::complex<double> (0.,k1));  
  return fhat_mult;
}

NFSOFTFourierCoefficients NFSOFT::computeDphi2MatrixMultiplication( 
  const NFSOFTFourierCoefficients & fhat ) 
{
  int degree = fhat.getPolynomialDegree();
  NFSOFTFourierCoefficients fhat_mult( degree );
  for (int n = 0; n <= degree; ++n) 
    for (int k1 = -n; k1 <= n; ++k1)
      for (int k2 = -n; k2 <= n; ++k2)
	      fhat_mult(n,k1,k2) = fhat(n,k1,k2) * (std::complex<double> (0.,-k2));  
  return fhat_mult;
}

NFSOFTFourierCoefficients NFSOFT::computeDphi2MatrixAdjointMultiplication(  
  const NFSOFTFourierCoefficients & fhat )
{
   int degree = fhat.getPolynomialDegree();
   NFSOFTFourierCoefficients fhat_mult( degree );
   for (int n = 0; n <= degree; ++n) 
     for (int k1 = -n; k1 <= n; ++k1)
       for (int k2 = -n; k2 <= n; ++k2)
	      fhat_mult(n,k1,k2) = fhat(n,k1,k2) * (std::complex<double> (0.,k2));  
   return fhat_mult;
 }

void NFSOFT::initNFSOFTPlan()
{
  nfsoft_init_guru( &nfsoftPlan_, internalPolynomialDegree_, numberOfPoints_,
		    NFSOFT_MALLOC_X | NFSOFT_MALLOC_F | NFSOFT_MALLOC_F_HAT,
		    PRE_PHI_HUT | PRE_PSI | MALLOC_X | MALLOC_F_HAT | MALLOC_F |
		    FFTW_INIT | FFT_OUT_OF_PLACE, 7, 1000);
}

void NFSOFT::destroyNFSOFTPlan() 
{ 
  nfsoft_finalize( &nfsoftPlan_ );
}
