// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFSOFT_HPP_
#define _NFSOFT_HPP_

#include "NFSOFTFourierCoefficients.hpp"

namespace Eigen{
  typedef Array<double, 6, Eigen::Dynamic> Array6Xd;
  typedef Array<std::complex<double>, 6, Eigen::Dynamic> Array6Xcd;
} // end namespace Eigen


namespace LORM {

  namespace NFFT {

    /** \class NFSOFT 
     *
     *  \brief The wrapper class for the nfsoft plan of the nfft C library. It
     *  provides, additionally to the evaluation of Wigner D-function
     *  expansions, the fast evaluation of the first and second order
     *  derivatives at nonequispaced points.
     *
     *  \note The class is non-copyable and non-copy-assignable. Hence,
     *  it cannot be passed nor returned by value.
     * 
     **/
   
    class NFSOFT {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N to set up tha nfsft plan.
      explicit NFSOFT( int numberOfPoints, int polynomialDegree);
      /// Destructor
      ~NFSOFT();
      
      /// Returns the number of points M.
      int getNumberOfPoints() const { return numberOfPoints_; };
      /// Returns the polynomial degree N.
      int getPolynomialDegree() const { return polynomialDegree_; };
      
      /// Returns the evaluation points of the nfsft in Euler angles 
      /// (phi1, theta, phi2) in [0,2*pi)x[0,pi]x[0,2*pi)
      Eigen::Array3Xd getLocalPointCoordinatesArray() const;

      /// Sets the evaluation points of the nfsft in Euler angles
      /// (phi1, theta, phi2) in [0,2*pi)x[0,pi]x[0,2*pi)
      void setLocalPointCoordinatesArray( const Eigen::Array3Xd & localPointCoords );
	
      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfsoft.
      Eigen::ArrayXcd evalYMatrixMultiplication( const NFSOFTFourierCoefficients & fhat ) {
       	assert( fhat.getPolynomialDegree() <= polynomialDegree_ );
       	return computeYMatrixMultiplication( fhat ); };

      /// Sets the array f of function values of the nfsft and returns the array
      /// fhat of Fourier coefficients after applying an adjoint nfsoft.
      template<typename ArrayType>
      NFSOFTFourierCoefficients evalYMatrixAdjointMultiplication( const ArrayType & f ) {
       	return computeYMatrixAdjointMultiplication( f, polynomialDegree_ ); }

      /// Returns the array of function values after multiplication with the
      /// matrix of the first order derivatives of the Fourier matrix.
      Eigen::Array3Xcd evalNablaYMatrixMultiplication( const NFSOFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint
      /// matrix of the first order derivatives of the Fourier matrix.
      NFSOFTFourierCoefficients evalNablaYMatrixAdjointMultiplication( const Eigen::Array3Xd & f );

      /// Returns the array of function values after multiplication with the
      /// matrix of the second order derivatives of the Fourier matrix.
      Eigen::Array6Xcd evalNabla2YMatrixMultiplication( const NFSOFTFourierCoefficients & fhat );      
      
    private:
      /// \note Non-copyable.
      NFSOFT ( const NFSOFT & ) {};
      /// \note Non-copy-assignable.
      const NFSOFT& operator= ( const NFSOFT & ) { return *this; };

      /// Evaluates the array f of function values of the nfsft. 
      Eigen::ArrayXcd getF() const;

      /// Evaluates the array fhat of Fourier coefficients of the nfsoft up to
      /// the given polynomial degree.
      NFSOFTFourierCoefficients getFhat( int degree) const;

      /// Sets the array f of function values of the nfsoft.
      void setF( const Eigen::ArrayXcd & f );

      /// Sets the array f of function values of the nfsoft.
      void setF( const Eigen::ArrayXd & f );

      /// Sets the array fhat of Fourier coefficient of the nfsoft.
      void setFhat( const NFSOFTFourierCoefficients & fhat );

      /// Sets the Fourier coefficients given by fhat and returns the array f of
      /// function values after applying an nfsoft.
      Eigen::ArrayXcd computeYMatrixMultiplication( const NFSOFTFourierCoefficients & fhat ) {
       	setFhat( fhat );
       	nfsoft_trafo( &nfsoftPlan_ );
       	return getF(); };

      template<typename ArrayType>
      NFSOFTFourierCoefficients computeYMatrixAdjointMultiplication( 
        const ArrayType & f, int degree ) {
       	setF( f );
       	nfsoft_adjoint( &nfsoftPlan_ );
       	return getFhat( degree ); }

      /// Helper functions for the computation of the coefficients of the diagonal matrix Dtheta.
      static double coef_a (double n, double k1, double k2) { 
	return n / ((n+1)*(2*n+1)) *  sqrt( (pow(n+1,2) - pow(k1,2)) * (pow(n+1,2) - pow(k2,2)) ); };
      static double coef_b (double n, double k1, double k2)  { 
	return (k1*k2)/(n*(n+1)); };
      static double coef_c (double n, double k1, double k2) {
	return (n+1)/(n*(2*n+1)) * sqrt( (n*n - k1*k1)*(n*n - k2*k2) ); };

      /// Returns the Fourier coefficients after multiplication with the matrix D_theta.
      /// \note The polynomial degree of the returned Fourier coefficients is increased by 1.
      static NFSOFTFourierCoefficients computeDthetaMatrixMultiplication( 
        const NFSOFTFourierCoefficients & fhat );     

      /// Returns the Fourier coefficients after multiplication with the adjoint matrix of D_theta.
      /// \note The polynomial degree of the returned Fourier coefficients is decreased by 1.
      static NFSOFTFourierCoefficients computeDthetaMatrixAdjointMultiplication(  
        const NFSOFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the matrix D_phi1.
      static NFSOFTFourierCoefficients computeDphi1MatrixMultiplication( 
        const NFSOFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint matrix of D_phi1.
      static NFSOFTFourierCoefficients computeDphi1MatrixAdjointMultiplication( 
        const NFSOFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the matrix D_phi2.
      static NFSOFTFourierCoefficients computeDphi2MatrixMultiplication( 
        const NFSOFTFourierCoefficients & fhat );

      /// Returns the Fourier coefficients after multiplication with the adjoint matrix of D_phi2.
      static NFSOFTFourierCoefficients computeDphi2MatrixAdjointMultiplication( 
        const NFSOFTFourierCoefficients & fhat );
    
      /// Initialization routine of the nfsft_plan
      void initNFSOFTPlan();
      
      /// Destruction routine of the nfsft_plan
      void destroyNFSOFTPlan();

      nfsoft_plan nfsoftPlan_;
      int numberOfPoints_;
      int polynomialDegree_;
      int internalPolynomialDegree_;

      Eigen::ArrayXd oneOverSinThetaArray_;
    };


  } // end namespace NFFT

} // end namespace LORM


#endif // _NFSOFT_HPP_
