// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _NFSOFT_FOURIER_COEFFICIENTS_HPP_
#define _NFSOFT_FOURIER_COEFFICIENTS_HPP_

#include "NFFTutils.hpp"

namespace LORM {

  namespace NFFT {

    /** \class NFSOFTFourierCoefficients 
     *
     *  \brief A utility class for the representation of the array of Fourier
     *  coefficients used in the class NFSOFT.
     *
     **/
    
    class NFSOFTFourierCoefficients {
    public:
      /// Default-Constructor.
      explicit NFSOFTFourierCoefficients() :
	fhat_array_ () {};
      /// Constructor. One must provide the polynomial degree and all Fourier
      /// coefficients are initialized to zero.
      explicit NFSOFTFourierCoefficients( int degree ) :
	fhat_array_ () { 
	setZero( degree ); };
      /// Constructor. One must provide an array of Fourier coefficients. \note
      /// The number of rows of the array must coincide with the dimension of
      /// some polynomial space.
      explicit NFSOFTFourierCoefficients( const Eigen::ArrayXcd & array ) :
	fhat_array_ () {
	setArray( array ); };

      /// Returns the array of the Fourier coefficients.
      const Eigen::ArrayXcd & getArray() const {
	return fhat_array_; };

      /// Returns the polynomial degree.
      int getPolynomialDegree() const { 
	return getPolynomialDegreeFromDimension( fhat_array_.rows() ); };
      
      /// Returns the dimension of the polynomial space.
      int getDimension() const {
	return fhat_array_.rows();  };

      /// Sets the Array of Fourier coefficients
      void setArray( const Eigen::ArrayXcd & array ) {
	getPolynomialDegreeFromDimension( array.rows() );
	fhat_array_ = array; };

      /// Sets the polyomial degree and initializes the Fourier coefficients to zero.
      void setZero( int degree ) {
	fhat_array_.setZero( getDimensionFromPolynomialDegree( degree ) );  };

      /// Sets the polyomial degree and initializes the Fourier coefficients to one.
      void setOne( int degree ) {
	fhat_array_.setOnes( getDimensionFromPolynomialDegree( degree ) );  };
      
      /// Index-Operator for assigning the Fourier coefficient f_(n,k1,k2).
      std::complex<double> & operator() (int n, int k1, int k2) {
	return fhat_array_(getIndex(n,k1,k2));  };

      /// Index-Operator for accessing the Fourier coefficient f_(n,k1,k2).
      const std::complex<double> & operator() (int n, int k1, int k2) const {
	return fhat_array_(getIndex(n,k1,k2));  };

      /// Addition-Assignment-Operator, which simply adds the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFSOFTFourierCoefficients & operator += ( const NFSOFTFourierCoefficients & other );

      /// Subtraction-Assignment-Operator, which simply subtracts the given Fourier
      /// coefficients. \note The polynomial degrees of the given Fourier need
      /// not to be equal, since they are resized accordingly.
      const NFSOFTFourierCoefficients & operator -= ( const NFSOFTFourierCoefficients & other );

      /// Multiplication-Operator. \note The polynomial degrees of the given
      /// Fourier need not to be equal, since they are resized accordingly.
      NFSOFTFourierCoefficients operator* ( const NFSOFTFourierCoefficients & other ) const;
 
    private:
      /// Returns the dimension of the polynomial spaces of given degree.
      static int getDimensionFromPolynomialDegree( int degree ) {
	return (2*degree+1)*(2*degree+2)*(2*degree+3)/6; };

      /// Returns the polynomial degree from the dimension of the polynomial spaces.
      static int getPolynomialDegreeFromDimension( int dim ) {
      	int degree = floor(pow(dim*6./8.,1./3.));
	assert( dim == getDimensionFromPolynomialDegree( degree ) );
	return degree; };

      /// Returns the linearized index of the (n,k1,k2)th Fourier coefficient.
      static int getIndex( int n, int k1, int k2 ) { 
	return ((2*n-1)*(2*n)*(2*n+1)/6 + (2*n+1)*(n+k1) + k2 + n); };

      Eigen::ArrayXcd fhat_array_;
    };

    std::ostream & operator<< ( std::ostream& os, const NFSOFTFourierCoefficients & fhat );

  } // end namespace NFFT

} // end namespace LORM

#endif // _NFSOFT_FOURIER_COEFFICIENTS_HPP_
