// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _COULOMB_ENERGY_HPP_
#define _COULOMB_ENERGY_HPP_

#include "PotentialEnergyRadialKernel.hpp"

namespace LORM {

  namespace ObjectiveFunctions {

   /** \class CoulombEnergy
   *
   *  \brief Implements the potential energy for the Coulomb kernel as
   *  PotentialEnergyRadialKernel. \f[ F(\boldsymbol x_1,\dots, \boldsymbol x_M)
   *  := \sum_{i \ne j =1}^M  \Vert \boldsymbol x_i - \boldsymbol x_j \Vert^{-1},
   *  \qquad \boldsymbol x_i \in \mathbb R^n \f]
   *
   */

  template<typename T_scalar>
  class CoulombEnergy : public PotentialEnergyRadialKernel<T_scalar> {
  public:
    /// Default-Constructor.
    explicit CoulombEnergy () :
      PotentialEnergyRadialKernel<T_scalar> () {};
    
    /// Constructor. One must provide a valid shared pointer to an instance of
    /// ManifoldPointArray, which is used to set up a plan for speeding up computations. 
    explicit CoulombEnergy (  const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray) : 
      PotentialEnergyRadialKernel<T_scalar> ( sp_pointArray ) {};
    
  private:
    T_scalar d0kernel( T_scalar r) const { return (T_scalar)  1./r; };
    T_scalar d1kernel( T_scalar r) const { return (T_scalar)  -1./(r*r); };
    T_scalar d2kernel( T_scalar r) const { return (T_scalar)  2./(r*r*r); };
  };

  } // end namespace ObjectiveFunctions

} // end namespace LORM

#endif // _COULOMB_ENERGY_HPP_
