// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ManifoldFunctionSecondOrder.hpp"

using namespace LORM;

template<typename T_scalar, typename T_return>
ManifoldFunctionSecondOrder<T_scalar, T_return>::ManifoldFunctionSecondOrder () :
  sp_pointArray_(),
  sp_gradFx_ (), 
  connectionPointArrayChanged_ (),
  computeFx_ (false),
  computeNablaFxArray_ (false),
  createGradFx_ (false),
  createHessFxMultTangentArray_ (false)  
{}

template<typename T_scalar, typename T_return>
ManifoldFunctionSecondOrder<T_scalar, T_return>::ManifoldFunctionSecondOrder(
  const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray)
{
  setPointArray( sp_pointArray );
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::setPointArray( 
  const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray )
{
  sp_pointArray_  =  sp_pointArray;
  setConnectionToSignalPointArrayChanged();
  setComputationFlags();
  createGradFx_ = true;
  createHessFxMultTangentArray_ = true;
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::evalFunction( 
  T_return * p_fx, const ManifoldPointArray<T_scalar> & pointArray ) const
{
  // check if we can save some computations
  if ( computeFx_ || ( sp_pointArray_.get() != &pointArray ) ) {
    computeFx( p_fx, pointArray.getCoordinates() );
    if ( sp_pointArray_.get() != &pointArray )
      computeFx_ = true;
    else
      computeFx_ = false;
  }
  else 
    *p_fx = fx_;
}

template<typename T_scalar, typename T_return>
T_return ManifoldFunctionSecondOrder<T_scalar, T_return>::evalFunction() const
{ 
  assert (sp_pointArray_.get() != 0);
  evalFunction( &fx_, *sp_pointArray_ );
  return fx_; 
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::evalGradientArray( 
  ManifoldTangentVectorArray<T_scalar> * p_gradientArray,
  const ManifoldPointArray<T_scalar> & pointArray ) const
{
  assert( p_gradientArray->getBasePointArray().get() == &pointArray );
  // check if we can save some computations
  if ( computeNablaFxArray_ || ( sp_pointArray_.get() != &pointArray ) ) {
    nablaFxArray_.resize( pointArray.getCoordinates().rows(), pointArray.getCoordinates().cols());
    computeNablaFxArray( &nablaFxArray_, pointArray.getCoordinates() );
    if ( sp_pointArray_.get() != &pointArray ) 
      computeNablaFxArray_ = true;
    else
	  computeNablaFxArray_ = false;
  }
  p_gradientArray->setCoordinates( nablaFxArray_ );
}

template<typename T_scalar, typename T_return>
const ManifoldTangentVectorArray<T_scalar> & ManifoldFunctionSecondOrder<T_scalar, T_return>::evalGradientArray() const 
{ 
  assert (sp_pointArray_.get() != 0);
  if (createGradFx_) {
    sp_gradFx_ = sp_pointArray_->createTangentVectorArray();
    assert( sp_gradFx_.get() != 0);
    createGradFx_ = false;
  }
  evalGradientArray( sp_gradFx_.get(), *sp_pointArray_ );
  return *sp_gradFx_;
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::evalHessianMatrixMultiplication ( 
  ManifoldTangentVectorArray<T_scalar> * p_tangentArray, 
  const ManifoldPointArray<T_scalar> & pointArray ) const
{
  assert( p_tangentArray->getBasePointArray().get() == &pointArray );
  if ( computeNablaFxArray_ || (sp_pointArray_.get() != p_tangentArray->getBasePointArray().get()) ) {
    computeNablaFxArray( &nablaFxArray_, pointArray.getCoordinates() );
    if ( sp_pointArray_.get() != p_tangentArray->getBasePointArray().get() )
      computeNablaFxArray_ = true;
    else 
      computeNablaFxArray_ = false;
  } 
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  nabla2FxMultTangentArray = 
    p_tangentArray->getCoordinates();
  computeNabla2FxMultiplication( &nabla2FxMultTangentArray, pointArray.getCoordinates() );
  p_tangentArray->applyChristoffelMatrixLinearCombinationMultiplication( nablaFxArray_ );
  nabla2FxMultTangentArray -= p_tangentArray->getCoordinates();
  p_tangentArray->setCoordinates( nabla2FxMultTangentArray );    
}

template<typename T_scalar, typename T_return>
const ManifoldTangentVectorArray<T_scalar> & ManifoldFunctionSecondOrder<T_scalar, T_return>::evalHessianMatrixMultiplication ( 
  const ManifoldTangentVectorArray<T_scalar> & tangentArray ) const
{
  assert (sp_pointArray_.get() != 0);
  assert( sp_pointArray_.get() == tangentArray.getBasePointArray().get() );
  if (createHessFxMultTangentArray_) {
    sp_hessFxMultTangentArray_ = sp_pointArray_->createTangentVectorArray(); 
    assert( sp_hessFxMultTangentArray_.get() != 0);
    createHessFxMultTangentArray_ = false;
  }
  *sp_hessFxMultTangentArray_ = tangentArray;
  evalHessianMatrixMultiplication ( sp_hessFxMultTangentArray_.get(), *sp_pointArray_ );
  return *sp_hessFxMultTangentArray_;      
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::setConnectionToSignalPointArrayChanged()
{
  connectionPointArrayChanged_ = sp_pointArray_->
    connectSignalChanged( boost::bind( & ManifoldFunctionSecondOrder<T_scalar, T_return>::setComputationFlags, this) ); 
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrder<T_scalar, T_return>::setComputationFlags()
{
  computeFx_ = true;
  computeNablaFxArray_ = true;
}

//Macro for explicit instantiations of ManifoldFunctionSecondOrder<T_scalar>
#define LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDER(T_scalar)	\
  template class ManifoldFunctionSecondOrder<T_scalar>; \
  template class ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> >; 

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDER( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDER(double)
#endif

#undef LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDER
