// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _MANIFOLD_FUNCTION_SECOND_ORDER_HPP_
#define _MANIFOLD_FUNCTION_SECOND_ORDER_HPP_

#include "../Manifolds/ManifoldPointArray.hpp"
#include "../Manifolds/ManifoldTangentVectorArray.hpp"
#include "../Manifolds/ManifoldParameterization.hpp"

namespace LORM {
  
  /** \class ManifoldFunctionSecondOrder
   *
   *  \brief The abstract base class template for functions on Riemannian manifolds,
   *         which provide derivatives up to second order.
   *  
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   *
   * \tparam T_scalar Numerical Type, e.g. float, double, long double 
   * \tparam T_return Return Type, default is T_scalar and
   *         Eigen::Array<T_scalar,1,Eigen::Dynamic> might be used for
   *         vectorized function evaluations
   *
   */
  
  template<typename T_scalar, typename T_return = T_scalar>
  class ManifoldFunctionSecondOrder {
  public:
    /// Default-Constructor.
    explicit ManifoldFunctionSecondOrder ();

    /// Constructor. One must provide a valid shared pointer to an instance of
    /// ManifoldPointArray, which is used to set up a plan for speeding up computations. 
    explicit ManifoldFunctionSecondOrder (const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray);
    
    /// Destructor.
    virtual ~ManifoldFunctionSecondOrder () {};
    
    /// Return the shared pointer to the registered ManifoldPointArray.
    const std::shared_ptr<const ManifoldPointArray<T_scalar> > & getPointArray() const { 
      return sp_pointArray_;};

    /// Registers the shared pointer of the ManifoldPointArray.
    void setPointArray( const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray );

    /// Evaluates the function at \c pointArray and returns the value in \c fx.
    void evalFunction( T_return * p_fx, const ManifoldPointArray<T_scalar> & pointArray ) const;
    
    /// Returns the function value at the registered ManifoldPointArray.
    T_return evalFunction() const;

    /// Evaluates the gradient of the function at \c pointArray and returns it
    /// as ManifoldTangentVectorArray in \c gradientArray.
    void evalGradientArray( ManifoldTangentVectorArray<T_scalar> * p_gradientArray,
			    const ManifoldPointArray<T_scalar> & pointArray ) const;

    /// Returns the gradient of the function as ManifoldTangentVectorArray at
    /// the registered ManifoldPointArray.
    const ManifoldTangentVectorArray<T_scalar> & evalGradientArray() const;

    /// Evaluates the matrix vector multiplicaton of \c tangentArray with the
    /// Hessian of f evaluated at \c pointArray. The result is returned in \c
    /// tangentArray.
    void evalHessianMatrixMultiplication ( ManifoldTangentVectorArray<T_scalar> * p_tangentArray, 
					   const ManifoldPointArray<T_scalar> & pointArray ) const;

    /// Returns the result of the matrix vector multiplication of \c
    /// tangentArray with the Hessian of f evaluated at the registered
    /// ManifoldPointArray.
    const ManifoldTangentVectorArray<T_scalar> & evalHessianMatrixMultiplication ( 
             const ManifoldTangentVectorArray<T_scalar> & tangentArray ) const;

  protected:
    /// Must implement the computation of f at the point given in Cartesian coordinates.
    virtual void computeFx( T_return * p_fx, 
			    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const =0;

    /// Must implement the computation of the array of first order derivatives
    /// \c nablaFxArray of f at the point given in Cartesian coordinates.
    virtual void computeNablaFxArray( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablaFxArray,
				      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const =0;

    /// Must implement the computation of the multiplication with the usual Hessian
    /// matrix consisting of the second order derivatives at the point given in
    /// Cartesian coordinates.
    virtual void computeNabla2FxMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nabla2fxMultTangentArray,
					const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const =0;
    
    /// Returns the array of the first order derivatives.
    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & getNablaFxArray() const 
    { return nablaFxArray_; };
    
  private:
    /// \note Non-copyable.
    ManifoldFunctionSecondOrder ( const ManifoldFunctionSecondOrder<T_scalar, T_return> & ) {};
    /// \note Non-copy-assignable.
    const ManifoldFunctionSecondOrder<T_scalar, T_return> & operator= ( const ManifoldFunctionSecondOrder<T_scalar, T_return> & ) { return *this; };

    void setConnectionToSignalPointArrayChanged();
    void setComputationFlags();
        
    std::shared_ptr<const ManifoldPointArray<T_scalar> > sp_pointArray_; 
    mutable std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > sp_gradFx_; 
    mutable std::shared_ptr<ManifoldTangentVectorArray<T_scalar> > sp_hessFxMultTangentArray_;   

    mutable T_return fx_;
    mutable Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> nablaFxArray_;

    boost::signals2::scoped_connection connectionPointArrayChanged_;
    mutable bool computeFx_;
    mutable bool computeNablaFxArray_;
    mutable bool createGradFx_;
    mutable bool createHessFxMultTangentArray_;
  };

  /** \namespace LORM::ObjectiveFunctions
   *
   *  \brief Namespace consisting of several objective functions on Riemannian
   *  manifolds, such as Coulomb energies and worst case quadrature errors.
   *
   **/

} // end namespace LORM

#endif // _MANIFOLD_FUNCTION_SECOND_ORDER_HPP_
