// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _MANIFOLD_FUNCTION_SECOND_ORDER_PARAMETERIZED_HPP_
#define _MANIFOLD_FUNCTION_SECOND_ORDER_PARAMETERIZED_HPP_

#include "ManifoldFunctionSecondOrder.hpp"

namespace LORM {

  /** \class ManifoldFunctionSecondOrderParameterized
   *
   *  \brief The abstract base class template for functions on Riemannian
   *         manifolds, which are evaluated from a parameterization and provide
   *         derivatives up to second order.
   *  
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   *
   * \tparam T_scalar Numerical Type, e.g. float, double, long double
   * \tparam T_return Return Type, default is T_scalar and
   *         Eigen::Array<T_scalar,1,Eigen::Dynamic> might be used for
   *         vectorized function evaluations
   *
   */
 
  template<typename T_scalar, typename T_return = T_scalar>
  class ManifoldFunctionSecondOrderParameterized : 
    public ManifoldFunctionSecondOrder<T_scalar, T_return> {
  public:
    /// Constructor. One must provide a valid shared pointer to an instance of
    /// ManifoldParameterization. 
    ManifoldFunctionSecondOrderParameterized ( const std::shared_ptr<const ManifoldParameterization<T_scalar> > & sp_parameterization );

    /// Destructor.
    virtual ~ManifoldFunctionSecondOrderParameterized () {};

    /// Returns the used ManifoldParameterization.
    const ManifoldParameterization<T_scalar> & getParameterization() const 
    { return *sp_parameterization_; }; 
    
  protected:    
    /// Must implement the computation of f at the given point in local coordinates.
    virtual void computeFxParameterized( T_return * p_fx, const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const =0;

    /// Must implement the computation of the array of first order derivatives
    /// \c nablaFxParameterizedArray of f at the point given in local
    /// coordinates.
    virtual void computeNablaFxParameterizedArray( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
						   const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const =0;
    
    /// Must implement the computation of the multiplication with the usual
    /// Hessian matrix consisting of the second order derivatives at the point
    /// given in local coordinates.
    virtual void computeNabla2FxParameterizedMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
							     const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const =0;
    
  private:
    void computeFx( T_return * p_fx, const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const;

    void computeNablaFxArray( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablaFxArray,
			      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const;

    void computeNabla2FxMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nabla2FxMultTangentArray,
					const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray ) const;

    // Helper function to convert Cartesian point coordinates to local point coordinates.
    void computeLocalPointCoordsArray ( const Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> & pointCoordsArray ) const;


    // temporary arrays
    mutable Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> localPointCoordsArray_;   
    mutable Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> nablaFxParameterizedArray_; 
    mutable Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> nabla2FxParameterizedMultTangentArray_;
   
    const std::shared_ptr<const ManifoldParameterization<T_scalar> > sp_parameterization_;
  };


} // end namespace LORM

#endif // _MANIFOLD_FUNCTION_SECOND_ORDER_PARAMETERIZED_HPP_
