// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POLYNOMIAL_S2_SUM_HPP_
#define _POLYNOMIAL_S2_SUM_HPP_

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Sphere/SphericalCoordinates.hpp"
#include "../NFFTwrapper/NFSFT.hpp"

namespace LORM {

  namespace ObjectiveFunctions {
    
    /** \class PolynomialS2Sum
     *
     *  \brief A test function implemented as
     *  ManifoldFunctionSecondOrderParameterized, which computes the sum of a
     *  random spherical harmonic expansion over given evaluation points, by
     *  means of an NFFT::NFSFT. \f[ F(\boldsymbol x_1,\dots, \boldsymbol x_M) :=
     *  \sum_{i=1}^M f (\boldsymbol x_i), \qquad \boldsymbol x_i \in \mathbb S^2
     *  \f] where \f[ f(\boldsymbol x) := \sum_{n=0}^N \sum_{k=-n}^n \hat
     *  f_{n,k} Y_{n,k}(\boldsymbol x), \qquad \boldsymbol x \in \mathbb S^2 \f]
     * 
     **/
    
    class PolynomialS2Sum :  
      public ManifoldFunctionSecondOrderParameterized<double> {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N. 
      PolynomialS2Sum( int numberOfPoints, int degree) :
	ManifoldFunctionSecondOrderParameterized<double> ( 
	  std::make_shared<const Sphere::SphericalCoordinates<double> > () ),
	nfsftPlan_ ( numberOfPoints, degree ),
	fhat_ ( Eigen::ArrayXcd::Random((degree+1)*(degree+1)) ) {};
      
    /// Returns the parameterization as SphericalCoordinates.
      const Sphere::SphericalCoordinates<double> & getSphericalCoordinates() const { 
	return *(static_cast<const Sphere::SphericalCoordinates<double> * > (&(this->getParameterization()))); };
      
    private:
      void computeFxParameterized( double * p_fx, const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
	nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
	*p_fx = ((Eigen::ArrayXXcd) nfsftPlan_.evalYMatrixMultiplication( fhat_ )).real().sum();     
      };
      
      void computeNablaFxParameterizedArray( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
					     const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
	nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
	*p_nablaFxParameterizedArray = ((Eigen::ArrayXXcd) nfsftPlan_.evalNablaYMatrixMultiplication( fhat_ )).real();
      };
      
      void computeNabla2FxParameterizedMultiplication( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
						       const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const {
      assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
      nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
      Eigen::Array3Xd nabla2FxParameterizedArray = ((Eigen::ArrayXXcd) nfsftPlan_.evalNabla2YMatrixMultiplication( fhat_ )).real();  
      
      for (int i=0; i<nfsftPlan_.getNumberOfPoints(); ++i) {
	Eigen::Array2d temp = p_nabla2FxParameterizedMultTangentArray->col(i);
	(*p_nabla2FxParameterizedMultTangentArray)(0,i) =  
	  nabla2FxParameterizedArray(0,i) * temp(0) +  nabla2FxParameterizedArray(1,i) * temp(1);
	(*p_nabla2FxParameterizedMultTangentArray)(1,i) =  
	  nabla2FxParameterizedArray(1,i) * temp(0) +  nabla2FxParameterizedArray(2,i) * temp(1);
      }
    };
      
      
      mutable NFFT::NFSFT nfsftPlan_;
      NFFT::NFSFTFourierCoefficients fhat_;
    };    

  } // end namespace ObjectiveFunctions
    
} // end namespace LORM

#endif // _POLYNOMIAL_S2_SUM_HPP_
