// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POLYNOMIAL_SO3_SUM_HPP_
#define _POLYNOMIAL_SO3_SUM_HPP_

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/RotationGroup/EulerAngles.hpp"
#include "../NFFTwrapper/NFSOFT.hpp"

namespace LORM {

  namespace ObjectiveFunctions {

    /** \class PolynomialSO3Sum
     *
     *  \brief A test function implemented as
     *  ManifoldFunctionSecondOrderParameterized, which computes the sum of a
     *  random Wigner D-function expansion over given evaluation points, by
     *  means of an NFFT::NFSOFT. \f[ F(\boldsymbol x_1,\dots, \boldsymbol x_M) :=
     *  \sum_{i=1}^M f (\boldsymbol x_i), \qquad \boldsymbol x_i \in \mathrm{SO(3)}
     *  \f] where \f[ f(\boldsymbol x) := \sum_{n=0}^N \sum_{k_1,k_2=-n}^n \hat
     *  f_{n,k_1,k_2} D_{n,k_1,k_2}(\boldsymbol x), \qquad \boldsymbol x \in \mathrm{SO(3)} \f]
     * 
     **/
    
    class PolynomialSO3Sum :  
      public ManifoldFunctionSecondOrderParameterized<double> {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N. 
      PolynomialSO3Sum( int numberOfPoints, int degree) :
	ManifoldFunctionSecondOrderParameterized<double> ( 
	  std::make_shared<const RotationGroup::EulerAngles<double> > () ),
	nfsoftPlan_ ( numberOfPoints, degree ),
	fhat_ ( Eigen::ArrayXcd::Random((2*degree+1)*(2*degree+2)*(2*degree+3)/6) ) {};
      
    /// Returns the parameterization as EulerAngles.
      const RotationGroup::EulerAngles<double> & getEulerAngles() const { 
	return *(static_cast<const RotationGroup::EulerAngles<double> * > (&(this->getParameterization()))); };
      
    private:
      void computeFxParameterized( double * p_fx, const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
	nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
	*p_fx = ((Eigen::ArrayXXcd) nfsoftPlan_.evalYMatrixMultiplication( fhat_ )).real().sum();     
      };
      
      void computeNablaFxParameterizedArray( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
					     const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
	nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
	*p_nablaFxParameterizedArray = ((Eigen::ArrayXXcd) nfsoftPlan_.evalNablaYMatrixMultiplication( fhat_ )).real();
      };
      
      void computeNabla2FxParameterizedMultiplication( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
						       const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const {
      assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
      nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
      Eigen::Array6Xd nabla2FxParameterizedArray = ((Eigen::ArrayXXcd) nfsoftPlan_.evalNabla2YMatrixMultiplication( fhat_ )).real();  
      
      for (int i=0; i<nfsoftPlan_.getNumberOfPoints(); ++i) {
	Eigen::Array3d temp = p_nabla2FxParameterizedMultTangentArray->col(i);
	(*p_nabla2FxParameterizedMultTangentArray)(0,i) =  
	  nabla2FxParameterizedArray(0,i) * temp(0) +  
	  nabla2FxParameterizedArray(1,i) * temp(1) +
	  nabla2FxParameterizedArray(2,i) * temp(2);
	(*p_nabla2FxParameterizedMultTangentArray)(1,i) =  
	  nabla2FxParameterizedArray(1,i) * temp(0) +  
	  nabla2FxParameterizedArray(3,i) * temp(1) +
	  nabla2FxParameterizedArray(4,i) * temp(2);
	(*p_nabla2FxParameterizedMultTangentArray)(2,i) =  
	  nabla2FxParameterizedArray(2,i) * temp(0) +  
	  nabla2FxParameterizedArray(4,i) * temp(1) +
	  nabla2FxParameterizedArray(5,i) * temp(2);
      }
    };
      
      
      mutable NFFT::NFSOFT nfsoftPlan_;
      NFFT::NFSOFTFourierCoefficients fhat_;
    };    

  } // end namespace ObjectiveFunctions
    
} // end namespace LORM

#endif // _POLYNOMIAL_SO3_SUM_HPP_
