// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "PolynomialSO3Vectorized.hpp"

using namespace LORM::ObjectiveFunctions;

PolynomialSO3Vectorized::PolynomialSO3Vectorized( int numberOfPoints, int degree) :
  ManifoldFunctionSecondOrderParameterized<double, Eigen::Array<double, 1, Eigen::Dynamic> > ( 
    std::make_shared<const RotationGroup::EulerAngles<double> > () ),
  nfsoftPlan_ ( numberOfPoints, degree ),
  fhat_ ( Eigen::ArrayXcd::Random((2*degree+1)*(2*degree+2)*(2*degree+3)/6) ) 
{}

void PolynomialSO3Vectorized::computeFxParameterized( 
  Eigen::Array<double, 1, Eigen::Dynamic> * p_fx, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  *p_fx = ((Eigen::ArrayXXcd) nfsoftPlan_.evalYMatrixMultiplication( fhat_ )).real().transpose();     
}

void PolynomialSO3Vectorized::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  *p_nablaFxParameterizedArray = 
    ((Eigen::ArrayXXcd) nfsoftPlan_.evalNablaYMatrixMultiplication( fhat_ )).real();
}

void PolynomialSO3Vectorized::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  Eigen::Array6Xd nabla2FxParameterizedArray = 
    ((Eigen::ArrayXXcd) nfsoftPlan_.evalNabla2YMatrixMultiplication( fhat_ )).real();  
      
  for (int i=0; i<nfsoftPlan_.getNumberOfPoints(); ++i) {
    Eigen::Array3d temp = p_nabla2FxParameterizedMultTangentArray->col(i);
    (*p_nabla2FxParameterizedMultTangentArray)(0,i) =  
      nabla2FxParameterizedArray(0,i) * temp(0) +  
      nabla2FxParameterizedArray(1,i) * temp(1) +
      nabla2FxParameterizedArray(2,i) * temp(2);
    (*p_nabla2FxParameterizedMultTangentArray)(1,i) =  
      nabla2FxParameterizedArray(1,i) * temp(0) +  
      nabla2FxParameterizedArray(3,i) * temp(1) +
      nabla2FxParameterizedArray(4,i) * temp(2);
    (*p_nabla2FxParameterizedMultTangentArray)(2,i) =  
      nabla2FxParameterizedArray(2,i) * temp(0) +  
      nabla2FxParameterizedArray(4,i) * temp(1) +
      nabla2FxParameterizedArray(5,i) * temp(2);
  }
}
