// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POLYNOMIAL_SO3_VECTORIZED_HPP_
#define _POLYNOMIAL_SO3_VECTORIZED_HPP_

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/RotationGroup/EulerAngles.hpp"
#include "../NFFTwrapper/NFSOFT.hpp"

namespace LORM {

  namespace ObjectiveFunctions {

    /** \class PolynomialSO3Vectorized
     *
     *  \brief A polynomial on the rotation group SO(3) with vectorized
     *  evaluations implemented as ManifoldFunctionSecondOrderParameterized,
     *  which computes Wigner D-function expansion over given evaluation points,
     *  by means of an NFFT::NFSOFT. \f[ F(\boldsymbol x_1,\dots, \boldsymbol
     *  x_M) := \big( f (\boldsymbol x_i) \big)_{i=1}^M, \qquad \boldsymbol x_i
     *  \in \mathrm{SO(3)} \f] where \f[ f(\boldsymbol x) := \sum_{n=0}^N
     *  \sum_{k_1,k_2=-n}^n \hat f_{n,k_1,k_2} D_{n,k_1,k_2}(\boldsymbol x), \qquad
     *  \boldsymbol x \in \mathrm{SO(3)} \f]
     * 
     **/
    
    class PolynomialSO3Vectorized :  
      public ManifoldFunctionSecondOrderParameterized<double, Eigen::Array<double, 1, Eigen::Dynamic> > {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N. 
      PolynomialSO3Vectorized( int numberOfPoints, int degree);
      
    /// Returns the parameterization as EulerAngles.
      const RotationGroup::EulerAngles<double> & getEulerAngles() const { 
	return *(static_cast<const RotationGroup::EulerAngles<double> * > 
		 (&(this->getParameterization()))); };
      
    private:
      void computeFxParameterized( 
        Eigen::Array<double, 1, Eigen::Dynamic> * p_fx, 
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
      void computeNablaFxParameterizedArray( 
        Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
      void computeNabla2FxParameterizedMultiplication( 
        Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const;   
      
      mutable NFFT::NFSOFT nfsoftPlan_;
      NFFT::NFSOFTFourierCoefficients fhat_;
    };    

  } // end namespace ObjectiveFunctions
    
} // end namespace LORM

#endif // _POLYNOMIAL_SO3_VECTORIZED_HPP_
