// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POLYNOMIAL_T2_SUM_HPP_
#define _POLYNOMIAL_T2_SUM_HPP_

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Torus/TorusParameterization.hpp"
#include "../NFFTwrapper/NFFT2d.hpp"

namespace LORM {

  namespace ObjectiveFunctions {
    
    /** \class PolynomialT2Sum
     *
     *  \brief A test function implemented as
     *  ManifoldFunctionSecondOrderParameterized, which computes the sum of a
     *  random trigonometric polynomial expansion over given evaluation points,
     *  by means of an NFFT::NFFT2d. \f[ F(\boldsymbol x_1,\dots, \boldsymbol
     *  x_M) := \sum_{i=1}^M f (\boldsymbol x_i), \qquad \boldsymbol x_i \in
     *  \mathbb T^2 \f] where \f[ f(\boldsymbol x(\alpha_1, \alpha_2)) :=
     *  \sum_{k_1=-N_1/2}^{N_1/2-1} \sum_{k_2=-N_2/2}^{N_2/2-1} \hat f_{k_1,k_2}
     *  \mathrm{e}^{-2\pi i (k_1,k_2)^{\top} (\alpha_1, \alpha_2)}, \qquad
     *  \boldsymbol x \in \mathbb T^2 \f]
     * 
     **/
    
    class PolynomialT2Sum :  
      public ManifoldFunctionSecondOrderParameterized<double> {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degrees (N1,N2). 
      PolynomialT2Sum( int numberOfPoints, Eigen::Array2i degrees) :
	ManifoldFunctionSecondOrderParameterized<double> ( 
	  std::make_shared<const Torus::TorusParameterization<double> > (2) ),
	nfft2dPlan_ ( numberOfPoints, degrees ),
	fhat_ ( (Eigen::ArrayXXcd) Eigen::ArrayXXcd::Random( 2*degrees(0), 2*degrees(1))) {};
      
    /// Returns the parameterization as TorusParameterization.
      const Torus::TorusParameterization<double> & getTorusParameterization() const { 
	return *(static_cast<const Torus::TorusParameterization<double> * > (&(this->getParameterization()))); };
      
    private:
      void computeFxParameterized( double * p_fx, const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
	nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
	*p_fx = ((Eigen::ArrayXXcd) nfft2dPlan_.evalYMatrixMultiplication( fhat_ )).real().sum();     
      };
      
      void computeNablaFxParameterizedArray( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
					     const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
	assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
	nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
	*p_nablaFxParameterizedArray = ((Eigen::ArrayXXcd) nfft2dPlan_.evalNablaYMatrixMultiplication( fhat_ )).real() / (2*M_PI);
      };
      
      void computeNabla2FxParameterizedMultiplication( Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
						       const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const {
      assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
      nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
      Eigen::Array3Xd nabla2FxParameterizedArray = ((Eigen::ArrayXXcd) nfft2dPlan_.evalNabla2YMatrixMultiplication( fhat_ )).real() / (4*M_PI*M_PI);  
      
      for (int i=0; i<nfft2dPlan_.getNumberOfPoints(); ++i) {
	Eigen::Array2d temp = p_nabla2FxParameterizedMultTangentArray->col(i);
	(*p_nabla2FxParameterizedMultTangentArray)(0,i) =  
	  nabla2FxParameterizedArray(0,i) * temp(0) +  nabla2FxParameterizedArray(1,i) * temp(1);
	(*p_nabla2FxParameterizedMultTangentArray)(1,i) =  
	  nabla2FxParameterizedArray(1,i) * temp(0) +  nabla2FxParameterizedArray(2,i) * temp(1);
      }
    };
      
      
      mutable NFFT::NFFT2d nfft2dPlan_;
      NFFT::NFFT2dFourierCoefficients fhat_;
    };    

  } // end namespace ObjectiveFunctions
    
} // end namespace LORM

#endif // _POLYNOMIAL_T2_SUM_HPP_
