// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "PolynomialT2Vectorized.hpp"

using namespace LORM::ObjectiveFunctions;

PolynomialT2Vectorized::PolynomialT2Vectorized( int numberOfPoints, Eigen::Array2i degrees ) :
  ManifoldFunctionSecondOrderParameterized<double, Eigen::Array<double, 1, Eigen::Dynamic> > ( 
    std::make_shared<const Torus::TorusParameterization<double> > (2) ),
  nfft2dPlan_ ( numberOfPoints, degrees ),
  fhat_ ( (Eigen::ArrayXXcd) Eigen::ArrayXXcd::Random( 2*degrees(0), 2*degrees(1) ) ) 
{}

void PolynomialT2Vectorized::setFourierCoefficients( const NFFT::NFFT2dFourierCoefficients & fHat ) {
  Eigen::Array2i degrees = nfft2dPlan_.getPolynomialDegrees();
  assert( (fHat.getPolynomialDegrees())(0) <= degrees(0) );
  assert( (fHat.getPolynomialDegrees())(1) <= degrees(1) );
  fhat_ = fHat;
}


void PolynomialT2Vectorized::computeFxParameterized( 
  Eigen::Array<double, 1, Eigen::Dynamic> * p_fx, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
  *p_fx = ((Eigen::ArrayXXcd) nfft2dPlan_.evalYMatrixMultiplication( fhat_ )).real().transpose();     
}

void PolynomialT2Vectorized::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
  *p_nablaFxParameterizedArray = 
    ((Eigen::ArrayXXcd) nfft2dPlan_.evalNablaYMatrixMultiplication( fhat_ )).real() / (2*M_PI);
}

void PolynomialT2Vectorized::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
  Eigen::Array3Xd nabla2FxParameterizedArray = 
    ((Eigen::ArrayXXcd) nfft2dPlan_.evalNabla2YMatrixMultiplication( fhat_ )).real() / (4*M_PI*M_PI);  
      
  for (int i=0; i<nfft2dPlan_.getNumberOfPoints(); ++i) {
    Eigen::Array2d temp = p_nabla2FxParameterizedMultTangentArray->col(i);
    (*p_nabla2FxParameterizedMultTangentArray)(0,i) =  
      nabla2FxParameterizedArray(0,i) * temp(0) +  nabla2FxParameterizedArray(1,i) * temp(1);
    (*p_nabla2FxParameterizedMultTangentArray)(1,i) =  
      nabla2FxParameterizedArray(1,i) * temp(0) +  nabla2FxParameterizedArray(2,i) * temp(1);
  }
}
