// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POLYNOMIAL_T2_VECTORIZED_HPP_
#define _POLYNOMIAL_T2_VECTORIZED_HPP_

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Torus/TorusParameterization.hpp"
#include "../NFFTwrapper/NFFT2d.hpp"

namespace LORM {

  namespace ObjectiveFunctions {

    /** \class PolynomialT2Vectorized
     *
     *  \brief A trigonometric polynomial on the torus T^2 with vectorized
     *  evaluations implemented as ManifoldFunctionSecondOrderParameterized,
     *  which computes the trigonometric expansion over given evaluation points,
     *  by means of an NFFT::NFFT2d. \f[ F(\boldsymbol x_1,\dots, \boldsymbol
     *  x_M) := \big( f (\boldsymbol x_i) \big)_{i=1}^M, \qquad \boldsymbol x_i
     *  \in \mathbb T^2 \f] where \f[ f(\boldsymbol x(\alpha_1, \alpha_2)) :=
     *  \sum_{k_1=-N_1/2}^{N_1/2-1} \sum_{k_2=-N_2/2}^{N_2/2-1} \hat f_{k_1,k_2}
     *  \mathrm{e}^{-2\pi i (k_1,k_2)^{\top} (\alpha_1, \alpha_2)}, \qquad
     *  \boldsymbol x \in \mathbb T^2 \f]
     * 
     **/
    
    class PolynomialT2Vectorized :  
      public ManifoldFunctionSecondOrderParameterized<double, Eigen::Array<double, 1, Eigen::Dynamic> > {
    public:
      /// Constructor. One must provide the number of points M and the
      /// polynomial degree N. 
      PolynomialT2Vectorized( int numberOfPoints, Eigen::Array2i degrees);
      
      /// Returns the parameterization as TorusParameterization.
      const Torus::TorusParameterization<double> & getTorusParameterization() const { 
	return *(static_cast<const Torus::TorusParameterization<double> * > 
		 (&(this->getParameterization()))); };

      /// Sets the Fourier coefficients.
      void setFourierCoefficients( const NFFT::NFFT2dFourierCoefficients & fHat );
      
    private:
      void computeFxParameterized( 
        Eigen::Array<double, 1, Eigen::Dynamic> * p_fx, 
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
      void computeNablaFxParameterizedArray( 
        Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
      void computeNabla2FxParameterizedMultiplication( 
        Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
	const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const;    
      
      mutable NFFT::NFFT2d nfft2dPlan_;
      NFFT::NFFT2dFourierCoefficients fhat_;
    };    
    
    
  } // end namespace ObjectiveFunctions
  
} // end namespace LORM

#endif // _POLYNOMIAL_T2_VECTORIZED_HPP_
