// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "PotentialEnergyRadialKernel.hpp"

using namespace LORM::ObjectiveFunctions;

template<typename T_scalar>
void PotentialEnergyRadialKernel<T_scalar>::computeFx( 
  T_scalar * p_fx, 
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const
{
  *p_fx = 0;
  // simple summation
  for( int i=0; i<pointCoordsArray.cols(); ++i) {
    // *p_fx += d0kernel((T_scalar) 0.); 
    for( int j=i+1; j<pointCoordsArray.cols(); ++j ) {
      T_scalar r = (pointCoordsArray.col(i) - pointCoordsArray.col(j)).matrix().norm();
	  *p_fx += 2 * d0kernel(r);
    }
  }
}

template<typename T_scalar>
void PotentialEnergyRadialKernel<T_scalar>::computeNablaFxArray( 
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablafxArray, 
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const
{
  (*p_nablafxArray).setZero();
  for( int i=0; i<pointCoordsArray.cols(); ++i) {
    for( int j=i+1; j<pointCoordsArray.cols(); ++j) { // derivative d1kernel(0)=0
      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> d_ij = 
	pointCoordsArray.col(i) - pointCoordsArray.col(j);
      T_scalar r = d_ij.matrix().norm();
      (*p_nablafxArray).col(i) += 2 * (d1kernel(r)/r * d_ij);
      (*p_nablafxArray).col(j) -= 2 * (d1kernel(r)/r * d_ij);
    }
  }
}

template<typename T_scalar>
void PotentialEnergyRadialKernel<T_scalar>::computeNabla2FxMultiplication( 
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nabla2fxMultTangentArray, 
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const 
{
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> tangentArray = 
    *p_nabla2fxMultTangentArray;
  (*p_nabla2fxMultTangentArray).setZero();
  for( int i=0; i<pointCoordsArray.cols(); ++i) {
    for( int j=0; j<pointCoordsArray.cols(); ++j) {
      const Eigen::Array<T_scalar, Eigen::Dynamic, 1> d_ij = 
	pointCoordsArray.col(i) - pointCoordsArray.col(j);
      T_scalar r = d_ij.matrix().norm();
      T_scalar s1 = (d2kernel(r)*r-d1kernel(r))/(r*r*r);
      T_scalar s2 =  d1kernel(r)/r;
      if ( i!=j ) { // derivatives  d1kernel(0)=d2kernel(0)=0
	(*p_nabla2fxMultTangentArray).col(i) -= 2 * (s1*d_ij * (d_ij * tangentArray.col(j)).sum()
						     + s2 * tangentArray.col(j)); // off-diagonal blocks
	(*p_nabla2fxMultTangentArray).col(i) += 2 * (s1*d_ij * (d_ij * tangentArray.col(i)).sum()
						     + s2 * tangentArray.col(i)); // diagonal summation
      }
    }
  }
}

//Macro for explicit instantiations of PotentialEnergyRadialKernel<T_scalar>
#define LORM_INSTANTIATE_POTENTIALENERGYRADIALKERNEL(T_scalar)	\
  template class PotentialEnergyRadialKernel<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_POTENTIALENERGYRADIALKERNEL( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_POTENTIALENERGYRADIALKERNEL(double)
#endif

#undef LORM_INSTANTIATE_POTENTIALENERGYRADIALKERNEL
