// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _POTENTIAL_ENERGY_RADIAL_KERNEL_HPP_
#define _POTENTIAL_ENERGY_RADIAL_KERNEL_HPP_

#include "ManifoldFunctionSecondOrder.hpp"

namespace LORM {

  namespace ObjectiveFunctions {

  /** \class PotentialEnergyRadialKernel
   *
   *  \brief Abstract base class template for the potential energy for a radial
   *  kernel implemented as ManifoldFunctionSecondOrder. \f[ F(\boldsymbol
   *  x_1,\dots, \boldsymbol x_M) := \sum_{i \ne j =1}^M K( \Vert \boldsymbol
   *  x_i - \boldsymbol x_j \Vert ), \qquad \boldsymbol x_i \in \mathbb R^n \f]
   *
   */

  template<typename T_scalar>
  class PotentialEnergyRadialKernel : public ManifoldFunctionSecondOrder<T_scalar> {
  public:
    /// Default-Constructor.
    explicit PotentialEnergyRadialKernel () :
      ManifoldFunctionSecondOrder<T_scalar>()  {};
    
    /// Constructor. One must provide a valid shared pointer to an instance of
    /// ManifoldPointArray, which is used to set up a plan for speeding up computations. 
    explicit PotentialEnergyRadialKernel ( 
      const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray) : 
      ManifoldFunctionSecondOrder<T_scalar>( sp_pointArray )  {};
    
  protected:
    /// Must implement the function evaluation of the radial kernel.
    virtual T_scalar d0kernel( T_scalar r) const =0 ;
    /// Must implement the first order derivative evaluation of the radial kernel.
    virtual T_scalar d1kernel( T_scalar r) const =0;
    /// Must implement the second order derivative evaluation of the radial kernel.
    virtual T_scalar d2kernel( T_scalar r) const =0;
    
  private:
    void computeFx( 
      T_scalar * p_fx, 
      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const;
    
    void computeNablaFxArray( 
      Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablafxArray, 
      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const;

    void computeNabla2FxMultiplication( 
      Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nabla2fxMultTangentArray, 
      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const;
    
  };

  } // end namespace ObjectiveFunctions

} // end namespace LORM


#endif // _POTENTIAL_ENERGY_RADIAL_KERNEL_HPP_
