// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "QuadratureErrorS2.hpp"

using namespace LORM::ObjectiveFunctions;

QuadratureErrorS2::QuadratureErrorS2 ( int numberOfPoints, int degree ) :
  ManifoldFunctionSecondOrderParameterized<double> ( 
	std::make_shared<const Sphere::SphericalCoordinates<double> > () ),
  nfsftPlan_ ( numberOfPoints, degree ),
  lambdaHat_ ( degree ),
  nuHat_ ( degree ), 
  w_ ( Eigen::ArrayXd::Zero(numberOfPoints) ) 
{
  initLambdaHat();
  initNuHat();      
  initWeights();
}

void QuadratureErrorS2::setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda ) 
{
  assert( lambda.rows() == nfsftPlan_.getPolynomialDegree()+1 );
  for ( int n=0; n<= nfsftPlan_.getPolynomialDegree(); ++n )
    for ( int k=-n; k<=n; ++k)
      lambdaHat_(n,k) = lambda(n);
}

void QuadratureErrorS2::computeFxParameterized( 
  double * p_fx, 
 const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  
  auto & e = evalErrorVector();
  *p_fx = (e.getArray().conjugate() * lambdaHat_.getArray() * e.getArray()).sum().real();	
}

void QuadratureErrorS2::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  
  auto le = lambdaHat_ * evalErrorVector();
  *p_nablaFxParameterizedArray = 
         2 * ( ( (Eigen::ArrayXXcd) nfsftPlan_.evalNablaYMatrixMultiplication( le )
		 ).real().rowwise() * w_.transpose() ).array();
}

void QuadratureErrorS2::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const 
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );

  Eigen::Array2Xd tangentArray = *p_nabla2FxParameterizedMultTangentArray;
  *p_nabla2FxParameterizedMultTangentArray =  
      2 * ( ( (Eigen::ArrayXXcd) nfsftPlan_.evalNablaYMatrixMultiplication( 
				 lambdaHat_ * nfsftPlan_.evalNablaYMatrixAdjointMultiplication( 
				 tangentArray.rowwise() * w_.transpose())) ).real().rowwise() 
	    * w_.transpose() ).array();

  auto le = lambdaHat_ * evalErrorVector();
  Eigen::Array3Xd nabla2Array = 
     ( (Eigen::ArrayXXcd) nfsftPlan_.evalNabla2YMatrixMultiplication( le ) ).real();

  (*p_nabla2FxParameterizedMultTangentArray).row(0) += 
    2 * w_.transpose() * (   nabla2Array.row(0) * tangentArray.row(0)
			   + nabla2Array.row(1) * tangentArray.row(1) );
  (*p_nabla2FxParameterizedMultTangentArray).row(1) += 
    2 * w_.transpose() * (   nabla2Array.row(1) * tangentArray.row(0)
			   + nabla2Array.row(2) * tangentArray.row(1) );
}
