// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _QUADRATURE_ERROR_S2_HPP
#define _QUADRATURE_ERROR_S2_HPP

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Sphere/SphericalCoordinates.hpp"
#include "../NFFTwrapper/NFSFT.hpp"

namespace LORM {
  
  namespace ObjectiveFunctions {

  
    /** \class QuadratureErrorS2
     *
     *  \brief The worst case quadrature error of equal weighted points on the
     *  sphere S^2 with respect to a given kernel and a measure implemented as
     *  ManifoldFunctionSecondOrderParameterized and computed by means of the
     *  NFFT::NFSFT. \f[ \mathrm{err}(\boldsymbol x_1,\dots, \boldsymbol x_M) :=
     *  \sum_{n=0}^N \lambda_n \sum_{k=-n}^n \big| \hat\nu_{n,k} - \frac{1}{M}
     *  \sum_{i=1}^M Y_{n,k} (\boldsymbol x_i) \big|^2, \qquad \boldsymbol x_i
     *  \in \mathbb S^2 \f]
     * 
     **/  
 

  class QuadratureErrorS2 :
    public ManifoldFunctionSecondOrderParameterized<double> {
  public:
    /// Constructor. One must provide the number of points M and the
    /// polynomial degree N.   
    explicit QuadratureErrorS2 ( int numberOfPoints, int degree );

    /// Returns the parameterization as SphericalCoordinates.
    const Sphere::SphericalCoordinates<double> & getSphericalCoordinates() const { 
      return *(static_cast<const Sphere::SphericalCoordinates<double> * > 
	       (&(this->getParameterization())) ); };

    /// Sets the Fourier coefficients lambda of the radial kernel. Initially
    /// these are the Fourier coefficients of the Dirichlet kernel.
    void setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda );

    /// Sets the Fourier coefficients nuHat of the measure. Initially these are
    /// Fourier coefficients of the cannonical surface measure.
    void setMeasureFourierCoefficients( const NFFT::NFSFTFourierCoefficients & nuHat ) {
      assert( nuHat.getPolynomialDegree() <= nfsftPlan_.getPolynomialDegree() );
      nuHat_ = nuHat; };
    
  private:
    void computeFxParameterized( 
      double * p_fx,  
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
    void computeNablaFxParameterizedArray( 
      Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
    
    void computeNabla2FxParameterizedMultiplication( 
       Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
       const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const;

    void initLambdaHat() {
      lambdaHat_.setOne( nfsftPlan_.getPolynomialDegree() ); };

    void initNuHat() {
      nuHat_(0,0) = std::complex<double> (1. / sqrt( 4. * M_PI ), 0); };

    void initWeights() {
      w_.setOnes( nfsftPlan_.getNumberOfPoints() );
      w_ /= (double) nfsftPlan_.getNumberOfPoints(); };

    const NFFT::NFSFTFourierCoefficients & evalErrorVector() const {
      e_ = nfsftPlan_.evalYMatrixAdjointMultiplication( w_ );
      e_ -= nuHat_;
      return e_; };
    
    mutable NFFT::NFSFT nfsftPlan_;
    mutable NFFT::NFSFTFourierCoefficients e_;
    NFFT::NFSFTFourierCoefficients lambdaHat_;
    NFFT::NFSFTFourierCoefficients nuHat_;
    mutable Eigen::ArrayXd w_;
  };

  } // end namespace ObjectiveFunctions

} // end namespace LORM

#endif // _QUADRATURE_ERROR_S2_HPP
