// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "QuadratureErrorS2R.hpp"

using namespace LORM::ObjectiveFunctions;

QuadratureErrorS2R::QuadratureErrorS2R ( int numberOfPoints, int degree ) :
  ManifoldFunctionSecondOrderParameterized<double> ( 
    std::make_shared<const Sphere::SphericalCoordinatesTimesR<double> > () ),
  nfsftPlan_ ( numberOfPoints, degree ),
  nuHat_ ( degree ), 
  w_ ( Eigen::ArrayXd::Zero(numberOfPoints) ) 
{
  initLambdaHat();
  initNuHat();      
}

void QuadratureErrorS2R::setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda ) 
{
  assert( lambda.rows() == nfsftPlan_.getPolynomialDegree()+1 );
  for ( int n=0; n<= nfsftPlan_.getPolynomialDegree(); ++n )
    for ( int k=-n; k<=n; ++k)
      lambdaHat_(n,k) = lambda(n);
}

void  QuadratureErrorS2R::computeFxParameterized( 
   double * p_fx, 
   const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(2) );
  w_ = localPointCoordsArray.row(2);
  
  auto & e = evalErrorVector();
  *p_fx = (e.getArray().conjugate() * lambdaHat_.getArray() * e.getArray()).sum().real();	
}

void  QuadratureErrorS2R::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(2) );
  w_ = localPointCoordsArray.row(2);
  
  *p_nablaFxParameterizedArray = 2 * computeReDEmultiplication( lambdaHat_ * evalErrorVector() );
}

void  QuadratureErrorS2R::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const 
{
  assert( localPointCoordsArray.cols() == nfsftPlan_.getNumberOfPoints() );
  nfsftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(2) );
  w_ = localPointCoordsArray.row(2);
      
  Eigen::Array3Xd tangentArray = *p_nabla2FxParameterizedMultTangentArray; 
  *p_nabla2FxParameterizedMultTangentArray = 
    2 * computeReDEmultiplication( lambdaHat_ * computeDEadjointMultiplication( tangentArray ) );

  auto le = lambdaHat_ * evalErrorVector();
  Eigen::Array3Xd nabla2SphereArray = 
    ( (Eigen::ArrayXXcd) nfsftPlan_.evalNabla2YMatrixMultiplication( le )).real();
  (*p_nabla2FxParameterizedMultTangentArray).row(0) += 
    2 * w_.transpose() * ( nabla2SphereArray.row(0) * tangentArray.row(0) +
			   nabla2SphereArray.row(1) * tangentArray.row(1));
  (*p_nabla2FxParameterizedMultTangentArray).row(1) += 
    2 * w_.transpose() * ( nabla2SphereArray.row(1) * tangentArray.row(0) +
			   nabla2SphereArray.row(2) * tangentArray.row(1));
  Eigen::Array2Xd nablaSphereArray = 
    ((Eigen::ArrayXXcd) nfsftPlan_.evalNablaYMatrixMultiplication( le )).real();
  (*p_nabla2FxParameterizedMultTangentArray).topRows(2) += 
    2 *	(nablaSphereArray.rowwise() *  tangentArray.row(2)).array();
  (*p_nabla2FxParameterizedMultTangentArray).row(2) += 
    2 * (nablaSphereArray.row(0) * tangentArray.row(0) + 
	 nablaSphereArray.row(1) * tangentArray.row(1));
}

Eigen::Array3Xd QuadratureErrorS2R::computeReDEmultiplication( 
  const NFFT::NFSFTFourierCoefficients & fhat ) const 
{
  Eigen::Array3Xd reDEfhat;
  reDEfhat.resize( 3, nfsftPlan_.getNumberOfPoints() );
  reDEfhat.topRows(2) = ( ( (Eigen::ArrayXXcd) 
			    nfsftPlan_.evalNablaYMatrixMultiplication( fhat )
			    ).real().rowwise() * w_.transpose() ).array();
  reDEfhat.row(2) = ( ( (Eigen::ArrayXXcd) 
			nfsftPlan_.evalYMatrixMultiplication( fhat )
			).real() ).transpose();
  return reDEfhat;    
}

LORM::NFFT::NFSFTFourierCoefficients QuadratureErrorS2R::computeDEadjointMultiplication( 
  const Eigen::Array3Xd & f ) const
{
  LORM::NFFT::NFSFTFourierCoefficients DEadjointf = 
    nfsftPlan_.evalNablaYMatrixAdjointMultiplication( (f.topRows(2)).rowwise() * w_.transpose() );
  DEadjointf += nfsftPlan_.evalYMatrixAdjointMultiplication( (Eigen::ArrayXd) f.row(2) );     
  return DEadjointf; 
}
