// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "QuadratureErrorSO3.hpp"

using namespace LORM::ObjectiveFunctions;

QuadratureErrorSO3::QuadratureErrorSO3 ( int numberOfPoints, int degree ) :
  ManifoldFunctionSecondOrderParameterized<double> ( 
    std::make_shared<const RotationGroup::EulerAngles<double> > () ),
  nfsoftPlan_ ( numberOfPoints, degree ),
  lambdaHat_ ( degree ),
  nuHat_ ( degree ), 
  w_ ( Eigen::ArrayXd::Zero(numberOfPoints) ) 
{
  initLambdaHat();
  initNuHat();      
  initWeights();
}

void QuadratureErrorSO3::setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda )
{
  assert( lambda.rows() == nfsoftPlan_.getPolynomialDegree()+1 );
  for ( int n=0; n<= nfsoftPlan_.getPolynomialDegree(); ++n )
    for ( int k1=-n; k1<=n; ++k1 )
      for ( int k2=-n; k2<=n; ++k2)
	lambdaHat_(n,k1,k2) = lambda(n);
}

void QuadratureErrorSO3::computeFxParameterized( 
  double * p_fx, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  
  auto & e = evalErrorVector();
  *p_fx = (e.getArray().conjugate() * lambdaHat_.getArray() * e.getArray()).sum().real();	
}

void QuadratureErrorSO3::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );
  
  auto le = lambdaHat_ * evalErrorVector();
  *p_nablaFxParameterizedArray = 
    2 * ( ( (Eigen::ArrayXXcd) nfsoftPlan_.evalNablaYMatrixMultiplication( le )
	    ).real().rowwise() * w_.transpose() ).array();
}

void QuadratureErrorSO3::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const 
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray );

  Eigen::Array3Xd tangentArray = *p_nabla2FxParameterizedMultTangentArray;
  *p_nabla2FxParameterizedMultTangentArray =  
    2 * ( ( (Eigen::ArrayXXcd) nfsoftPlan_.evalNablaYMatrixMultiplication( 
        lambdaHat_ * nfsoftPlan_.evalNablaYMatrixAdjointMultiplication( tangentArray.rowwise() * 
									w_.transpose() ) )
	    ).real().rowwise() * w_.transpose() ).array();

  auto le = lambdaHat_ * evalErrorVector();
  Eigen::Array6Xd nabla2Array = 
    ( (Eigen::ArrayXXcd) nfsoftPlan_.evalNabla2YMatrixMultiplication( le )).real();
  (*p_nabla2FxParameterizedMultTangentArray).row(0) += 
    2 * w_.transpose() * ( nabla2Array.row(0) * tangentArray.row(0) +
			   nabla2Array.row(1) * tangentArray.row(1) +
			   nabla2Array.row(2) * tangentArray.row(2) );
  (*p_nabla2FxParameterizedMultTangentArray).row(1) += 
    2 * w_.transpose() * ( nabla2Array.row(1) * tangentArray.row(0) +
			   nabla2Array.row(3) * tangentArray.row(1) +
			   nabla2Array.row(4) * tangentArray.row(2) );
  (*p_nabla2FxParameterizedMultTangentArray).row(2) += 
    2 * w_.transpose() * ( nabla2Array.row(2) * tangentArray.row(0) +
			   nabla2Array.row(4) * tangentArray.row(1) +
			   nabla2Array.row(5) * tangentArray.row(2) );
}


void QuadratureErrorSO3::initLambdaHat() 
{
  Eigen::ArrayXd lambda;
  lambda.resize( nfsoftPlan_.getPolynomialDegree() + 1);
  for ( int n=0; n<= nfsoftPlan_.getPolynomialDegree(); ++n )
    lambda(n) = (2*n + 1) / (8 * M_PI * M_PI);
  setRadialKernelFourierCoefficients( lambda );
}
