// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "QuadratureErrorSO3R.hpp"

using namespace LORM::ObjectiveFunctions;

QuadratureErrorSO3R::QuadratureErrorSO3R( int numberOfPoints, int degree ) :
  ManifoldFunctionSecondOrderParameterized<double> ( 
    std::make_shared<const RotationGroup::EulerAnglesTimesR<double> > () ),
  nfsoftPlan_ ( numberOfPoints, degree ),
  lambdaHat_ ( degree ),
  nuHat_ ( degree ), 
  w_ ( Eigen::ArrayXd::Zero(numberOfPoints) ) 
{
  initLambdaHat();
  initNuHat();      
}

void QuadratureErrorSO3R::setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda ) 
{
  assert( lambda.rows() == nfsoftPlan_.getPolynomialDegree()+1 );
  for ( int n=0; n<= nfsoftPlan_.getPolynomialDegree(); ++n )
    for ( int k1=-n; k1<=n; ++k1 )
      for ( int k2=-n; k2<=n; ++k2)
	lambdaHat_(n,k1,k2) = lambda(n);
}

void QuadratureErrorSO3R::computeFxParameterized( 
  double * p_fx, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(3) );
  w_ = localPointCoordsArray.row(3);

  auto & e = evalErrorVector();
  *p_fx = (e.getArray().conjugate() * lambdaHat_.getArray() * e.getArray()).sum().real();	
}

void QuadratureErrorSO3R::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(3) );
  w_ = localPointCoordsArray.row(3);
  
  *p_nablaFxParameterizedArray = 2 * computeReDEmultiplication( lambdaHat_ * evalErrorVector() );
}

void QuadratureErrorSO3R::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const
{
  assert( localPointCoordsArray.cols() == nfsoftPlan_.getNumberOfPoints() );
  nfsoftPlan_.setLocalPointCoordinatesArray( localPointCoordsArray.topRows(3) );
  w_ = localPointCoordsArray.row(3);
      
  Eigen::Array4Xd tangentArray = *p_nabla2FxParameterizedMultTangentArray; 
  *p_nabla2FxParameterizedMultTangentArray =
    2 * computeReDEmultiplication( lambdaHat_ * computeDEadjointMultiplication( tangentArray ) );

  auto le = lambdaHat_ * evalErrorVector();
  Eigen::Array6Xd nabla2RotationGroupArray = 
    ( (Eigen::ArrayXXcd) nfsoftPlan_.evalNabla2YMatrixMultiplication( le )).real();
  (*p_nabla2FxParameterizedMultTangentArray).row(0) += 
    2 * w_.transpose() * ( nabla2RotationGroupArray.row(0) * tangentArray.row(0) +
			   nabla2RotationGroupArray.row(1) * tangentArray.row(1) +
			   nabla2RotationGroupArray.row(2) * tangentArray.row(2));
  (*p_nabla2FxParameterizedMultTangentArray).row(1) += 
    2 * w_.transpose() * ( nabla2RotationGroupArray.row(1) * tangentArray.row(0) +
			   nabla2RotationGroupArray.row(3) * tangentArray.row(1) +
			   nabla2RotationGroupArray.row(4) * tangentArray.row(2));
  (*p_nabla2FxParameterizedMultTangentArray).row(2) += 
    2 * w_.transpose() * ( nabla2RotationGroupArray.row(2) * tangentArray.row(0) +
			   nabla2RotationGroupArray.row(4) * tangentArray.row(1) +
			   nabla2RotationGroupArray.row(5) * tangentArray.row(2));
  Eigen::Array3Xd nablaRotationGroupArray = 
    ((Eigen::ArrayXXcd) nfsoftPlan_.evalNablaYMatrixMultiplication( le )).real();
  (*p_nabla2FxParameterizedMultTangentArray).topRows(3) += 
    2 *	(nablaRotationGroupArray.rowwise() *  tangentArray.row(3)).array();
  (*p_nabla2FxParameterizedMultTangentArray).row(3) += 
    2 * ( nablaRotationGroupArray.row(0) * tangentArray.row(0) + 
	  nablaRotationGroupArray.row(1) * tangentArray.row(1) +
	  nablaRotationGroupArray.row(2) * tangentArray.row(2));
}

Eigen::Array4Xd QuadratureErrorSO3R::computeReDEmultiplication( 
  const NFFT::NFSOFTFourierCoefficients & fhat ) const
{
  Eigen::Array4Xd reDEfhat;
  reDEfhat.resize( 4, nfsoftPlan_.getNumberOfPoints() );
  reDEfhat.topRows(3) = ( ( (Eigen::ArrayXXcd) 
			    nfsoftPlan_.evalNablaYMatrixMultiplication( fhat )
			    ).real().rowwise() * w_.transpose() ).array();
  reDEfhat.row(3) = ( ( (Eigen::ArrayXXcd) 
			nfsoftPlan_.evalYMatrixMultiplication( fhat )
			).real() ).transpose();
  return reDEfhat;    
}

LORM::NFFT::NFSOFTFourierCoefficients QuadratureErrorSO3R::computeDEadjointMultiplication( 
  const Eigen::Array4Xd & f ) const
{
  LORM::NFFT::NFSOFTFourierCoefficients DEadjointf = 
    nfsoftPlan_.evalNablaYMatrixAdjointMultiplication( (f.topRows(3)).rowwise() * w_.transpose() );
  DEadjointf += nfsoftPlan_.evalYMatrixAdjointMultiplication( (Eigen::ArrayXd) f.row(3) );     
  return DEadjointf;
}

void QuadratureErrorSO3R::initLambdaHat() {
  Eigen::ArrayXd lambda;
  lambda.resize( nfsoftPlan_.getPolynomialDegree() + 1);
  for ( int n=0; n<= nfsoftPlan_.getPolynomialDegree(); ++n )
    lambda(n) = (2*n + 1) / (8 * M_PI * M_PI);
  setRadialKernelFourierCoefficients( lambda );
}
