// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _QUADRATURE_ERROR_SO3R_HPP
#define _QUADRATURE_ERROR_SO3R_HPP

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/RotationGroupTimesR/EulerAnglesTimesR.hpp"
#include "../NFFTwrapper/NFSOFT.hpp"

namespace LORM {

  namespace ObjectiveFunctions {
  
    /** \class QuadratureErrorSO3R
     *
     *  \brief The worst case quadrature error of weighted points on the
     *  rotation group SO(3) with respect to a given kernel and a measure
     *  implemented as ManifoldFunctionSecondOrderParameterized and computed by
     *  means of the NFFT::NFSOFT. \f[ \mathrm{err}(\boldsymbol x_1, w_1, \dots,
     *  \boldsymbol x_M, w_M) := \sum_{n=0}^N \lambda_n \sum_{k1,k2=-n}^n \big|
     *  \hat\nu_{n,k1,k2} - w_i \sum_{i=1}^M D_{n,k1,k2} (\boldsymbol x_i) \big|^2,
     *  \qquad (\boldsymbol x_i, w_i) \in \mathrm{SO(3)} \times \mathbb R \f]
     * 
     **/  
 

  class QuadratureErrorSO3R :
    public ManifoldFunctionSecondOrderParameterized<double> {
  public:
    /// Constructor. One must provide the number of points M and the
    /// polynomial degree N.   
    explicit QuadratureErrorSO3R ( int numberOfPoints, int degree );

    /// Returns the parameterization as SphericalCoordinates.
    const RotationGroup::EulerAnglesTimesR<double> & getEulerAnglesTimesR() const { 
      return *(static_cast<const RotationGroup::EulerAnglesTimesR<double> * > 
	       (&(this->getParameterization()))); };

    /// Sets the Fourier coefficients lambda of the radial kernel. Initially
    /// these are the Fourier coefficients of the Dirichlet kernel.
    void setRadialKernelFourierCoefficients( const Eigen::ArrayXd & lambda );
 
    /// Sets the Fourier coefficients nuHat of the measure. Initially these are
    /// Fourier coefficients of the cannonical surface measure.
    void setMeasureFourierCoefficients( const NFFT::NFSOFTFourierCoefficients & nuHat ) {
      assert( nuHat.getPolynomialDegree() <= nfsoftPlan_.getPolynomialDegree() );
      nuHat_ = nuHat; };
    
  private:
    void computeFxParameterized( double * p_fx, const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
    void computeNablaFxParameterizedArray( 
      Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
    
    void computeNabla2FxParameterizedMultiplication( 
      Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const;

    Eigen::Array4Xd computeReDEmultiplication( const NFFT::NFSOFTFourierCoefficients & fhat ) const;

    NFFT::NFSOFTFourierCoefficients computeDEadjointMultiplication( const Eigen::Array4Xd & f ) const;

    void initLambdaHat();
    void initNuHat() {
      nuHat_(0,0,0) = std::complex<double> (1., 0); };

    const NFFT::NFSOFTFourierCoefficients & evalErrorVector() const {
      e_ = nfsoftPlan_.evalYMatrixAdjointMultiplication( w_ );
      e_ -= nuHat_;
      return e_;  };
    
    mutable NFFT::NFSOFT nfsoftPlan_;
    mutable NFFT::NFSOFTFourierCoefficients e_;
    NFFT::NFSOFTFourierCoefficients lambdaHat_;
    NFFT::NFSOFTFourierCoefficients nuHat_;
    mutable Eigen::ArrayXd w_;
  };

  } // end namespace ObjectiveFunctions

} // end namespace LORM

#endif // _QUADRATURE_ERROR_SO3R_HPP
