// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "QuadratureErrorT2.hpp"

using namespace LORM::ObjectiveFunctions;

QuadratureErrorT2::QuadratureErrorT2 ( int numberOfPoints, Eigen::Array2i degrees ) :
  ManifoldFunctionSecondOrderParameterized<double> ( 
	std::make_shared<const Torus::TorusParameterization<double> > (2) ),
  nfft2dPlan_ ( numberOfPoints, degrees ),
  lambdaHat_ ( degrees ),
  nuHat_ ( degrees ), 
  w_ ( Eigen::ArrayXd::Zero(numberOfPoints) ) 
{
  initLambdaHat();
  initNuHat();      
  initWeights();
}

void QuadratureErrorT2::setKernelFourierCoefficients( const Eigen::ArrayXXd & lambda ) 
{
  Eigen::Array2i degrees = nfft2dPlan_.getPolynomialDegrees();
  assert( lambda.rows() == 2*degrees(0) );
  assert( lambda.cols() == 2*degrees(1) );
  for ( int k1=-degrees(0); k1< degrees(0); ++k1 )
    for ( int k2=-degrees(1); k2< degrees(1); ++k2 )
      lambdaHat_(k1,k2) = lambda(k1+degrees(0),k2+degrees(1));
}

void QuadratureErrorT2::computeFxParameterized( 
  double * p_fx, 
 const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
  
  auto & e = evalErrorVector();
  *p_fx = (e.getArray().conjugate() * lambdaHat_.getArray() * e.getArray()).sum().real();	
}

void QuadratureErrorT2::computeNablaFxParameterizedArray( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const 
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );
  
  auto le = lambdaHat_ * evalErrorVector();

  *p_nablaFxParameterizedArray = 
         2 * ( ( (Eigen::ArrayXXcd) nfft2dPlan_.evalNablaYMatrixMultiplication( le )
		 ).real().rowwise() * w_.transpose() ).array() / (2*M_PI);
}

void QuadratureErrorT2::computeNabla2FxParameterizedMultiplication( 
  Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
  const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const 
{
  assert( localPointCoordsArray.cols() == nfft2dPlan_.getNumberOfPoints() );
  nfft2dPlan_.setLocalPointCoordinatesArray( localPointCoordsArray / (2*M_PI) - 0.5 );

  Eigen::Array2Xd tangentArray = *p_nabla2FxParameterizedMultTangentArray;
  *p_nabla2FxParameterizedMultTangentArray =  
      2 * ( ( (Eigen::ArrayXXcd) nfft2dPlan_.evalNablaYMatrixMultiplication( 
				 lambdaHat_ * nfft2dPlan_.evalNablaYMatrixAdjointMultiplication( 
				 tangentArray.rowwise() * w_.transpose())) ).real().rowwise() 
	    * w_.transpose() ).array();

  auto le = lambdaHat_ * evalErrorVector();
  Eigen::Array3Xd nabla2Array = 
    ( (Eigen::ArrayXXcd) nfft2dPlan_.evalNabla2YMatrixMultiplication( le ) ).real();

  (*p_nabla2FxParameterizedMultTangentArray).row(0) += 
    2 * w_.transpose() * (   nabla2Array.row(0) * tangentArray.row(0)
			   + nabla2Array.row(1) * tangentArray.row(1) );
  (*p_nabla2FxParameterizedMultTangentArray).row(1) += 
    2 * w_.transpose() * (   nabla2Array.row(1) * tangentArray.row(0)
			   + nabla2Array.row(2) * tangentArray.row(1) );

  (*p_nabla2FxParameterizedMultTangentArray) /= (4*M_PI*M_PI);
}
