// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _QUADRATURE_ERROR_T2_HPP
#define _QUADRATURE_ERROR_T2_HPP

#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Torus/TorusParameterization.hpp"
#include "../NFFTwrapper/NFFT2d.hpp"

namespace LORM {
  
  namespace ObjectiveFunctions {

  
    /** \class QuadratureErrorT2
     *
     *  \brief The worst case quadrature error of equal weighted points on the
     *  torus T^2 with respect to a given kernel and a measure implemented as
     *  ManifoldFunctionSecondOrderParameterized and computed by means of the
     *  NFFT::NFFT2d. \f[ \mathrm{err}(\boldsymbol x_1,\dots, \boldsymbol x_M)
     *  := \sum_{k_1=-N_1/2}^{N_1/2-1} \sum_{k_2=-N_2/2}^{N_2-1}
     *  \lambda_{k_1,k_2} \big| \hat\nu_{k_1,k_2} - \frac{1}{M} \sum_{i=1}^M
     *  \exp_{k_1,k_2} (\boldsymbol x_i) \big|^2, \qquad \boldsymbol x_i \in
     *  \mathbb T^2 = \mathbb S^1 \times \mathbb S^1 \f]
     * 
     **/  
 

  class QuadratureErrorT2 :
    public ManifoldFunctionSecondOrderParameterized<double> {
  public:
    /// Constructor. One must provide the number of points M and the
    /// polynomial degrees (N_1,N_2).   
    explicit QuadratureErrorT2 ( int numberOfPoints, Eigen::Array2i degrees );

    /// Returns the parameterization as TorusParameterization.
    const Torus::TorusParameterization<double> & getTorusParameterization() const { 
      return *(static_cast<const Torus::TorusParameterization<double> * > 
	       (&(this->getParameterization())) ); };

    /// Sets the Fourier coefficients lambda of the periodic kernel. Initially
    /// these are the Fourier coefficients of the Dirichlet kernel.
    void setKernelFourierCoefficients( const Eigen::ArrayXXd & lambda );

    /// Sets the Fourier coefficients nuHat of the measure. Initially these are
    /// Fourier coefficients of the cannonical surface measure.
    void setMeasureFourierCoefficients( const NFFT::NFFT2dFourierCoefficients & nuHat ) {
      assert( (nuHat.getPolynomialDegrees())(0) <= (nfft2dPlan_.getPolynomialDegrees())(0) );
      assert( (nuHat.getPolynomialDegrees())(1) <= (nfft2dPlan_.getPolynomialDegrees())(1) );
      nuHat_ = nuHat; };
    
  private:
    void computeFxParameterized( 
      double * p_fx,  
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
      
    void computeNablaFxParameterizedArray( 
      Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
      const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const;
    
    void computeNabla2FxParameterizedMultiplication( 
       Eigen::Array<double, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
       const Eigen::Array<double, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const;

    void initLambdaHat() {
      lambdaHat_.setOne(nfft2dPlan_.getPolynomialDegrees()); };

    void initNuHat() {
      nuHat_(0,0) = std::complex<double> (1, 0); };

    void initWeights() {
      w_.setOnes( nfft2dPlan_.getNumberOfPoints() );
      w_ *= 1 / (double) nfft2dPlan_.getNumberOfPoints(); };

    const NFFT::NFFT2dFourierCoefficients & evalErrorVector() const {
      e_ = nfft2dPlan_.evalYMatrixAdjointMultiplication( w_ );
      e_ -= nuHat_;
      return e_; };
    
    mutable NFFT::NFFT2d nfft2dPlan_;
    mutable NFFT::NFFT2dFourierCoefficients e_;
    NFFT::NFFT2dFourierCoefficients lambdaHat_;
    NFFT::NFFT2dFourierCoefficients nuHat_;
    mutable Eigen::ArrayXd w_;
  };

  } // end namespace ObjectiveFunctions

} // end namespace LORM

#endif // _QUADRATURE_ERROR_T2_HPP
