// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _TEST_FUNCTION_HPP_
#define _TEST_FUNCTION_HPP_

#include  <boost/utility/base_from_member.hpp>
#include "ManifoldFunctionSecondOrder.hpp"
#include "ManifoldFunctionSecondOrderParameterized.hpp"
#include "../Manifolds/Sphere/SphericalCoordinates.hpp"
#include "../Manifolds/RotationGroup/EulerAngles.hpp"

namespace LORM {

namespace ObjectiveFunctions {

  /** \class QuadraticFunction
   *
   *  \brief Simple quadratic function implemented as
   *  ManifoldFunctionSecondOrder. \f[ F(\boldsymbol x_1,\dots, \boldsymbol x_M)
   *  := \sum_{i=1}^M \Vert\boldsymbol x_i - \boldsymbol x_i^*\Vert^2, \qquad
   *  \boldsymbol x_i \in \mathbb R^n \f]
   *
   */

  template<typename T_scalar>
  class QuadraticFunction : public ManifoldFunctionSecondOrder<T_scalar> {
  public:
    /// Constructor. One must provide an array of Cartesian coordinates for the
    /// location of the minimum \f$(\boldsymbol x_1^*, \dots, \boldsymbol x_M^*)\f$. 
    explicit QuadraticFunction ( const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & originArray ) :
      ManifoldFunctionSecondOrder<T_scalar>(),
      originArray_ (originArray)
    {};

    /// Constructor. One must provide an array of Cartesian coordinates for the
    /// location of the minimum \f$(\boldsymbol x_1^*, \dots, \boldsymbol
    /// x_M^*)\f$ and a valid shared pointer to an instance of
    /// ManifoldPointArray, which is used to set up a plat for speeding up
    /// computations.
    explicit QuadraticFunction ( const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & originArray,
				 const std::shared_ptr<const ManifoldPointArray<T_scalar> > & sp_pointArray) : 
      ManifoldFunctionSecondOrder<T_scalar>( sp_pointArray ),
      originArray_ (originArray) 
    {};
    
  private:
    void computeFx( T_scalar * p_fx, 
		    const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const {
      assert( pointCoordsArray.rows() == originArray_.rows() );
      assert( pointCoordsArray.cols() == originArray_.cols() );

      *p_fx = (pointCoordsArray - originArray_).square().sum();
    };
    void computeNablaFxArray( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablafxArray, 
			      const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const {
      assert( pointCoordsArray.rows() == originArray_.rows() );
      assert( pointCoordsArray.cols() == originArray_.cols() );

      *p_nablafxArray = 2* (pointCoordsArray - originArray_);
    };
    void computeNabla2FxMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * 
					p_nabla2fxMultTangentArray, 
					const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & 
					pointCoordsArray) const {
      assert( pointCoordsArray.rows() == originArray_.rows() );
      assert( pointCoordsArray.cols() == originArray_.cols() );

      *p_nabla2fxMultTangentArray *= 2;
    };
    
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> originArray_;
  };
 
  /** \class TestFunctionSphereS2
   *
   *  \brief Simple function for testing the implementation of a function
   *  parameterized in spherical coordinates as ManifoldFunctionSecondOrderParameterized.
   *
   **/

  template<typename T_scalar>
  class TestFunctionSphereS2 :  
    public ManifoldFunctionSecondOrderParameterized<T_scalar> {
  public:
    /// Default-Constructor.
    TestFunctionSphereS2() :
      ManifoldFunctionSecondOrderParameterized<T_scalar> ( 
	std::make_shared<const Sphere::SphericalCoordinates<T_scalar> > () ) {};

    /// Returns the parameterization as SphericalCoordinates.
    const Sphere::SphericalCoordinates<T_scalar> & getSphericalCoordinates() const { 
      return *(static_cast<const Sphere::SphericalCoordinates<T_scalar> * > (&(this->getParameterization()))); };
    
  private:  
    void computeFxParameterized( T_scalar * p_fx, const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
      // *p_fx = localPointCoordsArray.row(0).cos().sum();
      *p_fx = (localPointCoordsArray.row(0) * localPointCoordsArray.row(1)).sum();
    };

    void computeNablaFxParameterizedArray( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
					   const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
      // (*p_nablaFxParameterizedArray).row(0) = - localPointCoordsArray.row(0).sin();
      // (*p_nablaFxParameterizedArray).row(1).setZero();
      (*p_nablaFxParameterizedArray).row(0) = localPointCoordsArray.row(1);
      (*p_nablaFxParameterizedArray).row(1) = localPointCoordsArray.row(0);

    };

    void computeNabla2FxParameterizedMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
						     const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const {
      // (*p_nabla2FxParameterizedMultTangentArray).row(0) *= - localPointCoordsArray.row(0).cos();
      // (*p_nabla2FxParameterizedMultTangentArray).row(1).setZero();
      localPointCoordsArray.row(0);
      Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  temp = (*p_nabla2FxParameterizedMultTangentArray);
      (*p_nabla2FxParameterizedMultTangentArray).row(0) = temp.row(1);
      (*p_nabla2FxParameterizedMultTangentArray).row(1) = temp.row(0);
    };
  };

    /** \class TestFunctionRotationGroupSO3
   *
   *  \brief Simple function for testing the implementation of a function
   *  parameterized in Euler angles as ManifoldFunctionSecondOrderParameterized.
   *
   **/

  template<typename T_scalar>
  class TestFunctionRotationGroupSO3 :  
    public ManifoldFunctionSecondOrderParameterized<T_scalar> {
  public:
    /// Default-Constructor.
    TestFunctionRotationGroupSO3() :
      ManifoldFunctionSecondOrderParameterized<T_scalar> ( 
	std::make_shared<const RotationGroup::EulerAngles<T_scalar> > () ) {};

    /// Returns the parameterization as EulerAngles.
    const RotationGroup::EulerAngles<T_scalar> & getEulerAngles() const { 
      return *(static_cast<const RotationGroup::EulerAngles<T_scalar> * > (&(this->getParameterization()))); };
    
  private:  
    void computeFxParameterized( T_scalar * p_fx, const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
      *p_fx = (localPointCoordsArray.row(0) * 
	       localPointCoordsArray.row(1) * 
	       localPointCoordsArray.row(2)).sum();
    };

    void computeNablaFxParameterizedArray( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nablaFxParameterizedArray,
					   const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray) const {
      (*p_nablaFxParameterizedArray).row(0) = localPointCoordsArray.row(1)*localPointCoordsArray.row(2);
      (*p_nablaFxParameterizedArray).row(1) = localPointCoordsArray.row(0)*localPointCoordsArray.row(2);
      (*p_nablaFxParameterizedArray).row(2) = localPointCoordsArray.row(0)*localPointCoordsArray.row(1);
    };

    void computeNabla2FxParameterizedMultiplication( Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> * p_nabla2FxParameterizedMultTangentArray, 
						     const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & localPointCoordsArray ) const {
      Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic>  temp = 
	(*p_nabla2FxParameterizedMultTangentArray);
      (*p_nabla2FxParameterizedMultTangentArray).row(0) = localPointCoordsArray.row(2)*temp.row(1)
	+ localPointCoordsArray.row(1) * temp.row(2);
      (*p_nabla2FxParameterizedMultTangentArray).row(1) = localPointCoordsArray.row(2)*temp.row(0)
	+ localPointCoordsArray.row(0) * temp.row(2);
      (*p_nabla2FxParameterizedMultTangentArray).row(2) = localPointCoordsArray.row(1)*temp.row(0)
	+ localPointCoordsArray.row(0) * temp.row(1);
    };
  };

} // end namespace ObjectiveFunctions

} // end namespace LORM

#endif // _TEST_FUNCTION_HPP_
