// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ArmijoMethod.hpp"

using namespace LORM;

template<typename T_scalar>
int ArmijoMethod<T_scalar>::performLineSearch (
  ManifoldPointArray<T_scalar> * p_pointArray,
  ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
  const ManifoldFunctionSecondOrder<T_scalar> & function,
  T_scalar initialStepLength) const
{
  auto p0 = p_pointArray->getCoordinates();
  auto v0 = p_tangentArray->getCoordinates();
  
  T_scalar f0 = function.evalFunction();
  const T_scalar mu = 0.4;
  const T_scalar armijoCondition = mu * computeDotProduct( function.evalGradientArray(), 
							   *p_tangentArray ); 
  T_scalar alpha = initialStepLength;
      
  p_tangentArray->applyGeodesicStep( p_pointArray, alpha );
  
  int i=0;
  while( function.evalFunction() - f0 > alpha * armijoCondition && i < this->getIterationBound() ) {
    alpha *= 0.8;
    p_pointArray->setCoordinates( p0 );
    p_tangentArray->setCoordinates( v0 );
    p_tangentArray->applyGeodesicStep( p_pointArray, alpha );
    ++i;
  }
  
  if (i == this->getIterationBound() ) {
    p_pointArray->setCoordinates( p0 );
    p_tangentArray->setCoordinates( v0 );
    return 1;
  }
  else
    return 0;
}

//Macro for explicit instantiations of ArmijoMethod<T_scalar>
#define LORM_INSTANTIATE_ARMIJOMETHOD(T_scalar)	\
  template class ArmijoMethod<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_ARMIJOMETHOD( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_ARMIJOMETHOD(double)
#endif

#undef LORM_INSTANTIATE_ARMIJOMETHOD
