// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ArmijoMethodVectorized.hpp"

using namespace LORM;

template<typename T_scalar>
int ArmijoMethodVectorized<T_scalar>::performLineSearch(
   ManifoldPointArray<T_scalar> * p_pointArray,
   ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
   const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function,
   Eigen::Array<T_scalar, 1, Eigen::Dynamic> initialStepLengthArray) const
{
  auto p0 = p_pointArray->getCoordinates();
  auto v0 = p_tangentArray->getCoordinates();
      
  auto f0 = function.evalFunction();
  const T_scalar mu = 0.4;
  const Eigen::Array<T_scalar, 1, Eigen::Dynamic>  armijoConditionArray = 
    mu * computeDotProductVectorized( function.evalGradientArray(), *p_tangentArray ); 
  Eigen::Array<T_scalar, 1, Eigen::Dynamic> alphaArray = initialStepLengthArray;
  int M = initialStepLengthArray.cols();
      
  p_tangentArray->applyGeodesicStep( p_pointArray, alphaArray );
  
  int i=0;	
  int overallCondition = 1;
  while( overallCondition && i < this->getIterationBound() ) {	
    overallCondition = 0;
    for (int k=0; k<M; ++k) 
      if( (function.evalFunction())(k) - f0(k) > alphaArray(k) * armijoConditionArray(k)) {
	alphaArray(k) *= 0.8;
	overallCondition = 1;
      }
    p_pointArray->setCoordinates( p0 );
    p_tangentArray->setCoordinates( v0 );
    p_tangentArray->applyGeodesicStep( p_pointArray, alphaArray );
    ++i;
  }

  int errorCount = 0;
  for( int k=0; k<M; ++k)
    if( (function.evalFunction())(k) - f0(k) > alphaArray(k) * armijoConditionArray(k)) {
      alphaArray(k) = 0;
      errorCount += 1;
    }
  
  p_pointArray->setCoordinates( p0 );
  p_tangentArray->setCoordinates( v0 );
  p_tangentArray->applyGeodesicStep( p_pointArray, alphaArray );
  
  if( errorCount == M)
    return 1;
  else
    return 0;  
}

//Macro for explicit instantiations of ArmijoMethodVectorized<T_scalar>
#define LORM_INSTANTIATE_ARMIJOMETHODVECTORIZED(T_scalar)	\
  template class ArmijoMethodVectorized<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_ARMIJOMETHODVECTORIZED( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_ARMIJOMETHODVECTORIZED(double)
#endif

#undef LORM_INSTANTIATE_ARMIJOMETHODVECTORIZED
