// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ConjugateGradientMethod.hpp"
#include "../Utils/Math.hpp"

using namespace LORM;

template<typename T_scalar>
int ConjugateGradientMethod<T_scalar>::performOptimization(
  ManifoldPointArray<T_scalar> * p_pointArray,
  const ManifoldFunctionSecondOrder<T_scalar> & function ) const 
{
  assert( p_pointArray == function.getPointArray().get() );
  
  auto & p = *p_pointArray;
  auto sp_v = p.createTangentVectorArray();
  auto & v = *sp_v;
  v.setNegative( function.evalGradientArray() );

  T_scalar gradientNorm = computeNorm( function.evalGradientArray() );
  int errorCode = 0;
      
  int i=0;
  int r=0;
  int dof= v.getCoordinates().cols() * p.getManifold()->getDimension();
  std::cout << "CG dof: " << dof << std::endl;

  while( gradientNorm > this->getToleranceBound() && 
	 i < this->getIterationBound() ) {
    std::cout << "CG iteration: " << i << std::endl 
	      << "  f(x_" << i << ") = " << function.evalFunction() << std::endl
	      << "  GradientNorm: " <<  gradientNorm << std::endl;
    T_scalar vHv_alpha = computeDotProduct(v, function.evalHessianMatrixMultiplication(v));
    T_scalar alpha = 1;
    if (vHv_alpha != 0)
      alpha = Math::realAbs( computeDotProduct(function.evalGradientArray(), v) / vHv_alpha);
    
    errorCode = this->runLineSearch( &p, &v, function, alpha );
    if (errorCode != 0)
      break;
    
    T_scalar vHv_beta = computeDotProduct(v, function.evalHessianMatrixMultiplication(v));
    if ( vHv_beta != 0) {
      T_scalar gHv_beta = computeDotProduct(function.evalGradientArray(),
					    function.evalHessianMatrixMultiplication(v));
      T_scalar beta = gHv_beta / vHv_beta;
      v.setAlphaTimesXPlusBetaTimesY( -1, function.evalGradientArray(), beta, v);
      if ( computeDotProduct(function.evalGradientArray(), v) > 0 || 
	   ((i+1) % dof) == (r % dof)) {
	v.setNegative( function.evalGradientArray() );
	r = i+1;
      }
    }
    else {
      v.setNegative( function.evalGradientArray() );
      r = i+1;
    }
    
    ++i;
    gradientNorm = computeNorm( function.evalGradientArray() );
  }
  
  return errorCode;
}

//Macro for explicit instantiations of ConjugateGradientMethod<T_scalar>
#define LORM_INSTANTIATE_CONJUGATEGRADIENTMETHOD(T_scalar)	\
  template class ConjugateGradientMethod<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_CONJUGATEGRADIENTMETHOD( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_CONJUGATEGRADIENTMETHOD(double)
#endif

#undef LORM_INSTANTIATE_CONJUGATEGRADIENTMETHOD
