// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ConjugateGradientMethodVectorized.hpp"
#include "../Utils/Math.hpp"

using namespace LORM;

template <typename T_scalar>
int ConjugateGradientMethodVectorized<T_scalar>::performOptimization ( 
  ManifoldPointArray<T_scalar> * p_pointArray,
  const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function ) const
{
  assert( p_pointArray == function.getPointArray().get() );

  auto & p = *p_pointArray;
  auto sp_v = p.createTangentVectorArray();
  auto & v = *sp_v;
  v.setNegative( function.evalGradientArray() );

  auto gvArray = computeDotProductVectorized( v, function.evalGradientArray() );
  auto gradientNormArray =  computeNormVectorized( function.evalGradientArray() );
  int M = gradientNormArray. cols();
  int errorCode = 0;
  
  int i=0;
  int dof= p.getManifold()->getDimension();
  std::cout << "CG dof: " << dof << std::endl;
  
  while( gradientNormArray.sum() / M > this->getToleranceBound() && 
	 i < this->getIterationBound() ) {
    std::cout << "CG iteration: " << i << std::endl 
	      << "  Sum of f(x_" << i << ") = " << function.evalFunction().sum() << std::endl
	      << "  Sum of GradientNorm / number of points: " <<  gradientNormArray.sum() / M 
	      << std::endl;
    
    Eigen::Array<T_scalar, 1, Eigen::Dynamic> alphaArray;
    alphaArray.setOnes(M);
    auto & hess_v_alpha = function.evalHessianMatrixMultiplication(v);
    for ( int k=0; k<M; ++k) {
      T_scalar vHv_alpha = (v.getCoordinates().col(k) * hess_v_alpha.getCoordinates().col(k)).sum();
      if (vHv_alpha != 0)
	alphaArray(k) = Math::realAbs( (function.evalGradientArray().getCoordinates().col(k) *
					v.getCoordinates().col(k)).sum() / vHv_alpha );
    }
    
    errorCode = this->runLineSearch( &p, &v, function, alphaArray );
    if (errorCode != 0)
      break;
    
    auto & hess_v_beta = function.evalHessianMatrixMultiplication(v);
    Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> vNewCoords = v.getCoordinates();
    for ( int k=0; k<M; ++k) {
      auto vHv_beta = ( v.getCoordinates().col(k) * hess_v_beta.getCoordinates().col(k) ).sum(); 
      if ( ((i+1) % dof != 0) && (vHv_beta != 0) ) {
	auto gHv_beta = ( function.evalGradientArray().getCoordinates().col(k) * 
			  hess_v_beta.getCoordinates().col(k) ).sum(); 
	T_scalar beta = gHv_beta / vHv_beta;
	vNewCoords.col(k) =  - function.evalGradientArray().getCoordinates().col(k)
	  + beta * v.getCoordinates().col(k);
	
	if ( (function.evalGradientArray().getCoordinates().col(k) * vNewCoords.col(k)).sum() > 0 )
	  vNewCoords.col(k) = - vNewCoords.col(k);
      }
      else
	vNewCoords.col(k) = - function.evalGradientArray().getCoordinates().col(k);
    }
    v.setCoordinates( vNewCoords );
    
    ++i;
    gradientNormArray = computeNormVectorized( function.evalGradientArray() );
  }
  
  return errorCode;
}

//Macro for explicit instantiations of ConjugateGradientMethodVectorized<T_scalar>
#define LORM_INSTANTIATE_CONJUGATEGRADIENTMETHODVECTORIZED(T_scalar)	\
  template class ConjugateGradientMethodVectorized<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_CONJUGATEGRADIENTMETHODVECTORIZED( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_CONJUGATEGRADIENTMETHODVECTORIZED(double)
#endif

#undef LORM_INSTANTIATE_CONJUGATEGRADIENTMETHODVECTORIZED
