// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _DESCENT_METHOD_HPP_
#define _DESCENT_METHOD_HPP_

#include "LineSearchMethod.hpp"

namespace LORM {

  /** \class DescentMethod
   *
   * \brief The abstract base class template for the implementation of descent
   * methods on Riemannian manifolds, which are intended to compute iteratively
   * a local minimizer of a given function.
   *
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   * 
   */

  template<typename T_scalar>
  class DescentMethod {
  public:     
    /// Constructor, which sets the tolerance bound to 1e-8 and the iteration
    /// bound to 100. One must provide a valid shared pointer to an instance of
    /// LineSearchMethod.
    explicit DescentMethod(const std::shared_ptr<const LineSearchMethod<T_scalar> > & sp_lineSearch) :
      sp_lineSearch_(sp_lineSearch),
      toleranceBound_(1e-8),
      iterationBound_(100) {};
    /// Default-Destructor.
    virtual ~DescentMethod() {};
    
    /// This function performs the computation of a local minimizer of a given
    /// function of type ManifoldFunctionSecondorder with the initial point
    /// given as a ManifoldPointArray. The computed minimizer is returned by
    /// overriding the initial point.
    int run ( ManifoldPointArray<T_scalar> * p_pointArray,
	      const ManifoldFunctionSecondOrder<T_scalar> & function ) const;

    /// Returns the tolerance bound.
    T_scalar getToleranceBound() const { return toleranceBound_; };
    /// Returns the iteration bound.
    int getIterationBound() const { return iterationBound_; };

  protected:
    /// Must implement the descent method computations of run().
    virtual int performOptimization ( 
      ManifoldPointArray<T_scalar> * p_pointArray,
      const ManifoldFunctionSecondOrder<T_scalar> & function ) const =0; 
    
    /// Calls run() of LineSearchMethod.
    int runLineSearch( ManifoldPointArray<T_scalar> * p_pointArray,
		       ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
		       const ManifoldFunctionSecondOrder<T_scalar> & function,
		       T_scalar initialStepLength) const { 
      return sp_lineSearch_->run( p_pointArray, p_tangentArray, function, initialStepLength ); };
			
  private:   
    /// \note Non-copyable.
    DescentMethod ( const DescentMethod<T_scalar> & ) {};
    /// \note Non-copy-assignable.
    const DescentMethod<T_scalar> & operator= ( const DescentMethod<T_scalar> & ) { return *this; };

    const std::shared_ptr<const LineSearchMethod<T_scalar> > sp_lineSearch_;
    T_scalar toleranceBound_;
    int iterationBound_;    
  };
  
} // end namespace LORM

#endif // _DESCENT_METHOD_HPP_
