// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _DESCENT_METHOD_VECTORIZED_HPP_
#define _DESCENT_METHOD_VECTORIZED_HPP_

#include "LineSearchMethodVectorized.hpp"

namespace LORM {

  /** \class DescentMethod
   *
   * \brief The abstract base class template for the implementation of descent
   * methods on Riemannian manifolds with vectorized function evaluations, which
   * are intended to compute iteratively a local minimizer of a given function.
   *
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   * 
   */

  template<typename T_scalar>
  class DescentMethodVectorized {
  public:     
    /// Constructor, which sets the tolerance bound to 1e-8 and the iteration
    /// bound to 10. One must provide a valid shared pointer to an instance of
    /// LineSearchMethodVectorized.
    explicit DescentMethodVectorized(const std::shared_ptr<const LineSearchMethodVectorized<T_scalar> > & sp_lineSearch) :
      sp_lineSearch_(sp_lineSearch),
      toleranceBound_(1e-8),
      iterationBound_(10) {};
    /// Default-Destructor.
    virtual ~DescentMethodVectorized() {};
    
    /// This function performs the computation of a local minimizer of a given
    /// function of type ManifoldFunctionSecondorder with the initial point
    /// given as a ManifoldPointArray. The computed minimizer is returned by
    /// overriding the initial point.
    int run ( 
      ManifoldPointArray<T_scalar> * p_pointArray,
      const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function ) const;

    /// Returns the tolerance bound.
    T_scalar getToleranceBound() const { return toleranceBound_; };
    /// Returns the iteration bound.
    int getIterationBound() const { return iterationBound_; };

  protected:
    /// Must implement the descent method computations of run().
    virtual int performOptimization ( ManifoldPointArray<T_scalar> * p_pointArray,
				      const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function ) const =0; 
    
    /// Calls run() of LineSearchMethod.
    int runLineSearch(
      ManifoldPointArray<T_scalar> * p_pointArray,
      ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
      const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function,
      Eigen::Array<T_scalar, 1, Eigen::Dynamic> initialStepLengthArray) const { 
      return sp_lineSearch_->run( p_pointArray, p_tangentArray, function, initialStepLengthArray ); };
			
  private:   
    /// \note Non-copyable.
    DescentMethodVectorized ( const DescentMethodVectorized<T_scalar> & ) {};
    /// \note Non-copy-assignable.
    const DescentMethodVectorized<T_scalar> & operator= ( const DescentMethodVectorized<T_scalar> & ) { return *this; };

    const std::shared_ptr<const LineSearchMethodVectorized<T_scalar> > sp_lineSearch_;
    T_scalar toleranceBound_;
    int iterationBound_;    
  };
  
} // end namespace LORM

#endif // _DESCENT_METHOD_VECTORIZED_HPP_
