// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _LINE_SEARCH_METHOD_HPP_
#define _LINE_SEARCH_METHOD_HPP_

#include "../ObjectiveFunctions/ManifoldFunctionSecondOrder.hpp"
// #include "../Utils/TestUtils.hpp"

namespace LORM {

  /** \class LineSearchMethod
   *
   *  \brief The abstract base class template for the implementation of the line
   *  search on Riemannian manifolds, which is intended to compute iteratively a local
   *  minimizer of a given function along a geodesic curve.
   *
   *  \note The class is non-copyable and non-copy-assignable. Hence, it cannot
   *  be passed nor returned by value.
   *
   */

  template<typename T_scalar>
  class LineSearchMethod {
  public:
    /// Default-Constructor, which sets the tolerance bound to 1e-8 and the
    /// iteration bound to 10.
    explicit LineSearchMethod() :
      toleranceBound_(1e-8),
      iterationBound_(50) {};
    /// Default-Destructor.
    virtual ~LineSearchMethod() {};
    
    /// This function performs the line search along a geodesic starting at the
    /// point given as a ManifoldPointArray with direction given as
    /// ManifoldTangentVectorArray for a function of type
    /// ManifoldFunctionSecondOrder, where the length for the initial search
    /// step is provide by \c initialStepLength. The computed minimizer and
    /// corresponding direction is returned by overriding the initial
    /// conditions.
    int run ( ManifoldPointArray<T_scalar> * p_pointArray,
	      ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
	      const ManifoldFunctionSecondOrder<T_scalar> & function,
	      T_scalar initialStepLength) const;
  
    /// Returns the tolerance bound.
    T_scalar getToleranceBound() const { return toleranceBound_; };
    /// Returns the iteration bound.
    int getIterationBound() const { return iterationBound_; };

  protected:
    /// Must implement the line search computations for run() and returns an
    /// error code (sucess=0, fail=1).
    virtual int performLineSearch( ManifoldPointArray<T_scalar> * p_pointArray,
				   ManifoldTangentVectorArray<T_scalar> * p_tangentArray,
				   const ManifoldFunctionSecondOrder<T_scalar> & function,
				   T_scalar initialStepLength) const =0;

  private:
    /// \note Non-copyable.
    LineSearchMethod ( const LineSearchMethod<T_scalar> & ) {};
    /// \note Non-copy-assignable.
    const LineSearchMethod<T_scalar> & operator= ( const LineSearchMethod<T_scalar> & ) { 
      return *this; };

    T_scalar toleranceBound_;
    int iterationBound_;
  };
    
} // End namespace LORM

#endif // _LINE_SEARCH_METHOD_HPP_
