// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "SteepestDescentMethod.hpp"
#include "../Utils/Math.hpp"

using namespace LORM;

template<typename T_scalar>
int SteepestDescentMethod<T_scalar>::performOptimization(
  ManifoldPointArray<T_scalar> * p_pointArray,
  const ManifoldFunctionSecondOrder<T_scalar> & function ) const
{
  assert( p_pointArray == function.getPointArray().get() );
  
  auto & p = *p_pointArray;
  auto sp_v = p.createTangentVectorArray();
  auto & v = *sp_v;
  v.setNegative( function.evalGradientArray() );
  
  T_scalar gradientNorm = computeNorm( function.evalGradientArray() );
  int errorCode = 0;
  
  int i=0;
  while( gradientNorm > this->getToleranceBound() && 
	 i < this->getIterationBound() ) {
    std::cout << "SD iteration: " << i << std::endl 
	      << "  f(x_" << i << ") = " << function.evalFunction() << std::endl
	      << "  GradientNorm: " <<  gradientNorm << std::endl;
    
    T_scalar vHv = computeDotProduct(v, function.evalHessianMatrixMultiplication(v) );
    T_scalar alpha = 1;
    if ( vHv != 0 )
      alpha = Math::realAbs( computeDotProduct(function.evalGradientArray(), v) / vHv);
    
    errorCode = this->runLineSearch( &p, &v, function, alpha );
    if (errorCode != 0)
      break;
    v.setNegative( function.evalGradientArray() );
    
    ++i;
    gradientNorm = computeNorm( function.evalGradientArray() );
  }
  
  return errorCode;
}

//Macro for explicit instantiations of SteepestDescentMethod<T_scalar>
#define LORM_INSTANTIATE_STEEPESTDESCENTMETHOD(T_scalar)	\
  template class SteepestDescentMethod<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_STEEPESTDESCENTMETHOD( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_STEEPESTDESCENTMETHOD(double)
#endif

#undef LORM_INSTANTIATE_STEEPESTDESCENTMETHOD
