// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "SteepestDescentMethodVectorized.hpp"
#include "../Utils/Math.hpp"

using namespace LORM;

template <typename T_scalar>
int SteepestDescentMethodVectorized<T_scalar>::performOptimization ( 
  ManifoldPointArray<T_scalar> * p_pointArray,
  const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> > & function ) const
{
  assert( p_pointArray == function.getPointArray().get() );
      
  auto & p = *p_pointArray;
  auto sp_v = p.createTangentVectorArray();
  auto & v = *sp_v;
  v.setNegative( function.evalGradientArray() );
  
  Eigen::Array<T_scalar, 1, Eigen::Dynamic>  gradientNormArray = 
    computeNormVectorized( function.evalGradientArray() );
  int M = gradientNormArray.cols();
  int errorCode = 0;
  
  int i=0;
  while( gradientNormArray.sum() / M > this->getToleranceBound() && 
	 i < this->getIterationBound() ) {
    std::cout << "SD iteration: " << i << std::endl 
	      << " Sum of f(x_" << i << "): " << function.evalFunction().sum() << std::endl
	      << " Sum of GradientNorm / number of points: " <<  gradientNormArray.sum() / M 
	      << std::endl;
    
    Eigen::Array<T_scalar, 1, Eigen::Dynamic> alphaArray;
    alphaArray.setOnes(M);
    auto & hess_v = function.evalHessianMatrixMultiplication(v);
    for( int k=0; k<M; ++k) {
      auto vHv = (v.getCoordinates().col(k) * hess_v.getCoordinates().col(k)).sum();
      if( vHv != 0 )
	alphaArray(k) = Math::realAbs( (function.evalGradientArray().getCoordinates().col(k) *
					v.getCoordinates().col(k)).sum() / vHv );
    }
    
    errorCode = this->runLineSearch( &p, &v, function, alphaArray );
    if (errorCode != 0)
      break;
    v.setNegative( function.evalGradientArray() );
    
    ++i;
    gradientNormArray = computeNormVectorized( function.evalGradientArray() );
  }
  
  return errorCode;
}

//Macro for explicit instantiations of SteepestDescentMethodVectorized<T_scalar>
#define LORM_INSTANTIATE_STEEPESTDESCENTMETHODVECTORIZED(T_scalar)	\
  template class SteepestDescentMethodVectorized<T_scalar>;

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_STEEPESTDESCENTMETHODVECTORIZED( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_STEEPESTDESCENTMETHODVECTORIZED(double)
#endif

#undef LORM_INSTANTIATE_STEEPESTDESCENTMETHODVECTORIZED
