// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#ifndef _SIGNAL_ON_CHANGE_HPP_
#define _SIGNAL_ON_CHANGE_HPP_

namespace LORM {

  /** \class SignalOnChange
   *
   *  \brief An abstract utility class for classes which need to send a signal if states have changed.
   *
   **/

  class SignalOnChange {
  public:
    /// Connects the function provided by \c actionOnSignal to the signal, which
    /// will be invoked by SignalOnChange::sendSignal().
    boost::signals2::connection connectSignalChanged( const std::function<void ()> & actionOnSignal ) const { 
      return signal_.connect( actionOnSignal); };
    
  protected:
    explicit SignalOnChange () : signal_ () {};
    /// \note The copy constructor constructs a new signal, so that existing connections are not duplicated.
    SignalOnChange ( const SignalOnChange &  ) : signal_  () {}; // construct new signal
    /// \note The assignment oprator does nothing but returning a const reference to the object.
    const SignalOnChange & operator= ( const SignalOnChange & ) { return *this; }; // do not copy the signal

    /// Sends the signal to all connections. This function should be called if
    /// the state of the inherited object has changed. 
    void sendSignal() const { signal_(); };
       
  private:
    mutable boost::signals2::signal<void ()> signal_;
  };

} // end namespace LORM

#endif // _SIGNAL_ON_CHANGE_HPP_
