// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "TestUtils.hpp"

namespace LORM {

  template <typename T_scalar>
  void evalLineFunctionWithQuadraticApproximation(
     const ManifoldFunctionSecondOrder<T_scalar> & function,
     const ManifoldPointArray<T_scalar> & pointArray,
     const ManifoldTangentVectorArray<T_scalar> & tangentArray,
     T_scalar stepLength, 
     int sampleSize,
     const char * filename ) 
  {
    assert( tangentArray.getBasePointArray().get() == &pointArray); 
    
    auto sp_tangent = pointArray.createTangentVectorArray();
    
    T_scalar f0; 
    function.evalFunction(&f0, pointArray);
    function.evalGradientArray( &(*sp_tangent), pointArray );
    T_scalar d1f0 = computeDotProduct( *sp_tangent, tangentArray );
    sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    function.evalHessianMatrixMultiplication( &(*sp_tangent), pointArray );
    T_scalar d2f0 = computeDotProduct( *sp_tangent, tangentArray );
    
    std::cout << "Quadratic Approximation: " << std::endl
     	      << f0 << " + "  << d1f0 << " * t + " << d2f0 << " / 2 * t^2 "
	      << "  (Output File: " << filename << ")" << std::endl << std::endl;
    
    auto sp_point = pointArray.clonePointArray();
    sp_tangent = sp_point->createTangentVectorArray();
    
    sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    
    std::ofstream file (filename);
    assert( file.is_open() );
    file.precision(12);
    
    for( int i=0; i<sampleSize; ++i) {
      T_scalar t = stepLength * (2*i-sampleSize) / (T_scalar) (sampleSize); 
      sp_tangent->applyGeodesicStep( &(*sp_point), t);
      T_scalar fx = 0;
      function.evalFunction(&fx, *sp_point);
      file << t << " " << fx << " " << f0 + d1f0*t + 0.5*d2f0*t*t << std::endl;
      sp_point->setCoordinates( pointArray.getCoordinates() );
      sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    }
    
    file.close();
  }
  
  template <typename T_scalar>
  void evalLineFunctionVectorizedWithQuadraticApproximation(
    const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> >& function,
    const ManifoldPointArray<T_scalar> & pointArray,
    const ManifoldTangentVectorArray<T_scalar> & tangentArray,
    T_scalar stepLength, 
    int sampleSize,
    const char * filename ) 
  {
    assert( tangentArray.getBasePointArray().get() == &pointArray); 
    
    auto sp_tangent = pointArray.createTangentVectorArray();
    
    Eigen::Array<T_scalar, 1, Eigen::Dynamic> f0; 
    function.evalFunction(&f0, pointArray);
    function.evalGradientArray( &(*sp_tangent), pointArray );
    Eigen::Array<T_scalar, 1, Eigen::Dynamic> d1f0 = 
      computeDotProductVectorized( *sp_tangent, tangentArray );
    sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    function.evalHessianMatrixMultiplication( &(*sp_tangent), pointArray );
    Eigen::Array<T_scalar, 1, Eigen::Dynamic> d2f0 = 
    computeDotProductVectorized( *sp_tangent, tangentArray );
    
    std::cout << "Quadratic Approximation: " << std::endl
	      << f0 << " + "  << std::endl 
	      << d1f0 << " * t + " << std::endl 
	      << d2f0 << " / 2 * t^2 " << std::endl
	      << "  (Output File: " << filename << ")" << std::endl << std::endl;
    
    auto sp_point = pointArray.clonePointArray();
    sp_tangent = sp_point->createTangentVectorArray();
    
    sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    
    std::ofstream file (filename);
    assert( file.is_open() );
    file.precision(12);
    
    for( int i=0; i<sampleSize; ++i) {
      T_scalar t = stepLength * (2*i-sampleSize) / (T_scalar) (sampleSize); 
      sp_tangent->applyGeodesicStep( &(*sp_point), t);
      Eigen::Array<T_scalar, 1, Eigen::Dynamic> fx;
      function.evalFunction(&fx, *sp_point);
      file << t  << " " 
	   << fx << " " 
	   << f0 + d1f0*t + 0.5*d2f0*t*t  << std::endl;
      sp_point->setCoordinates( pointArray.getCoordinates() );
      sp_tangent->setCoordinates( tangentArray.getCoordinates() );
    }
    
    file.close();
  }

} // end namespace LORM

//Macro for explicit instantiations of TestUtils<T_scalar>
#define LORM_INSTANTIATE_TESTUTILS(T_scalar)				\
  template void LORM::evalLineFunctionWithQuadraticApproximation<T_scalar>( \
    const ManifoldFunctionSecondOrder<T_scalar> & function,\
    const ManifoldPointArray<T_scalar> & pointArray,		\
    const ManifoldTangentVectorArray<T_scalar> & tangentArray,	\
    T_scalar stepLength,					   \
    int sampleSize,				\
    const char * filename );						\
  template void LORM::evalLineFunctionVectorizedWithQuadraticApproximation<T_scalar>( \
    const ManifoldFunctionSecondOrder<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> >& function, \
    const ManifoldPointArray<T_scalar> & pointArray,			\
    const ManifoldTangentVectorArray<T_scalar> & tangentArray,		\
    T_scalar stepLength,							\
    int sampleSize,							\
    const char * filename );
  
#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_TESTUTILS( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_TESTUTILS(double)
#endif

#undef LORM_INSTANTIATE_TESTUTILS
