// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroup/EulerAngles.hpp"
#include "../src/ObjectiveFunctions/TestFunctions.hpp"
#include "../src/Manifolds/RotationGroup/RotationGroupPointArray.hpp"
#include "../src/Utils/TestUtils.hpp"

/** \file test_EulerAngles.cpp
   *
   *  \brief A simple example demonstrating the parameterization in Euler angles
   *  by LORM::ObjectiveFunctions::TestFunctionRotationGroupSO3.
   *
   */

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
typedef LORM_INSTANTIATE_SCALAR_TYPE T_scalar;
#else
typedef double T_scalar;
#endif

int main() {
  
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  
  // creates an instance of an test function which makes use of Euler angles
  // parameterization
  TestFunctionRotationGroupSO3<T_scalar> fun;
    
  // gets the associated Euler Angeles parameterization
  auto & paramSO3 = fun.getEulerAngles();
  // gets the associated SO(3) manifold
  auto & sp_rotationGroup = paramSO3.getRotationGroup();
  
  // prints general informations of the manifold object
  cout << "Manifold name: " << sp_rotationGroup->getName() << endl;
  cout << "Manifold dimension: " << sp_rotationGroup->getDimension() << endl;
  cout << "Manifold ambient space dimension: " 
       << sp_rotationGroup->getAmbientSpaceDimension() << endl << endl;
  
  // generates random Euler angles coordinates of a single point
  Eigen::Array<T_scalar, 3, 1> localCoords;
  localCoords.setRandom();
  // computes the cooresponding rotation matrix
  auto matrixCoords = paramSO3.evalParameterization( localCoords );
  Eigen::Map<const Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > map 
    ( matrixCoords.data(), 3, 3 );

  // prints the results of the conversion between local and Cartesian coordinates
  cout << "Euler angles: " << endl << localCoords << endl;
  cout << "Rotation matrix: " << endl << map << endl;
  cout << "Computed Euler angles: " << endl 
       << paramSO3.evalInverseParameterization( matrixCoords ) << endl;
  matrixCoords = paramSO3.evalParameterization( 
	           paramSO3.evalInverseParameterization( matrixCoords ) );
  cout << "Recomputed rotation matrix: " << endl <<  map << endl << endl;

  // prints several matrices needed for Riemannian geometry computations
  cout << "Jacobi Matrix: " << endl 
       << paramSO3.evalJacobiMatrix( localCoords ) << endl;
  cout << "inverse Riemann Matrix: " << endl 
       << paramSO3.evalInverseRiemannianMatrix( localCoords ) << endl;
  // generates random local coefficients for the Christoffel matrix
  Eigen::Array<T_scalar, 3, 1> localCoeffs;
  localCoeffs.setRandom();
  cout << "Christoffel Matrix linear combination: " << endl
       << paramSO3.evalChristoffelMatrixLinearCombination( localCoeffs, localCoords ) 
       << endl << endl;

  // creates a shared pointer to new instance of an array of points on SO(3) of size M
  const int M = 10;
  auto sp_points = make_shared<RotationGroup::RotationGroupPointArray<T_scalar> >
    ( sp_rotationGroup );
  // creates an associated array of tangent vectors 
  RotationGroup::RotationGroupTangentVectorArray<T_scalar> tangents( sp_points);
  // sets random points and random tangent vectors
  sp_points->setCoordinates( Eigen::Array<T_scalar, 9, M>::Random() );
  tangents.setCoordinates( Eigen::Array<T_scalar, 9, M>::Random() );
    
  cout << "Test function in Euler angles parameterization:" << endl
       << " number of Points: " << M << endl;  

  // outputs a data file with 1000 equidistant sampling values of the the test
  // function and its quadratic approximation along a geodesic starting at the
  // points in direction of the tangent vectors
  evalLineFunctionWithQuadraticApproximation ( 
     fun, *sp_points, tangents, (T_scalar) 1., 1000, 
     "test_EulerAngles_TestFunctionQuadraticApproximation.dat");

  return 0;
}
