// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Torus/TorusPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/ObjectiveFunctions/QuadratureErrorT2.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"
#include <fstream>


/** \file test_ditherT2.cpp
   *
   *  \brief A brief demonstrating of dithering/halftoning on the torus T^2 by
   *  means of LORM::ObjectiveFunctions::QuadratureErrorT2 and the nonlinear CG
   *  method. 
   *
   */


using namespace std;
using namespace LORM;
using namespace ObjectiveFunctions;
using namespace NFFT;

NFFT2dFourierCoefficients loadMeasureFourierCoefficients( const char * p_fileName ) {
  ifstream file (p_fileName);
  assert( file.is_open() );
  int dim1, dim2;
  file >> dim1; 
  file >> dim2;
  assert( dim1 % 2 == 0 && dim2 % 2 == 0);
  Eigen::Array2i degrees;
  degrees << dim1/2, dim2/2;
  NFFT2dFourierCoefficients fHat( degrees );
  for ( int k2=-degrees(1); k2<degrees(1); ++k2)
    for ( int k1=-degrees(0); k1<degrees(0); ++k1) {
      double real, im;
      char unit;
      file >> real;
      file >> im;
      file >> unit;
      fHat(k1,k2) = std::complex<double>(real,im);
    }
  file.close();
  return fHat;
}

void normalizeNFFT2dFourierCoefficients( NFFT2dFourierCoefficients * p_fHat ) {
  Eigen::Array2i degrees = p_fHat->getPolynomialDegrees();
  double fHat00 = ((*p_fHat)(0,0)).real(); 
  for ( int k1=-degrees(0); k1<degrees(0); ++k1)
    for ( int k2=-degrees(1); k2<degrees(1); ++k2)
      (*p_fHat)(k1,k2) /= fHat00;    
}

Eigen::ArrayXXd computeDiscrepancyKernelFourierCoefficients( const Eigen::Array2i & degrees ) {
  Eigen::ArrayXXd fHatArray;
  fHatArray.setZero( 2*degrees(0), 2*degrees(1) );
  for ( int k1=-degrees(0)+1; k1<degrees(0); ++k1)
    for ( int k2=-degrees(1)+1; k2<degrees(1); ++k2) {
      fHatArray(k1+degrees(0),k2+degrees(1)) = 
	1. / pow( (double) (k1*k1 + k2*k2), 1.5 );
    }
  fHatArray(degrees(0),degrees(1)) = 0;
  return fHatArray;
}

void saveCoordinatesArray( const Eigen::ArrayXXd & coordsArray, 
			   const char * p_fileName ) 
{
  std::ofstream file (p_fileName);
  assert( file.is_open() );
  file << coordsArray.rows() << std::endl;
  file << coordsArray.cols() << std::endl;
  file << coordsArray.transpose() << std::endl;
  file.close();
}

int main() {

  // load and normalize the Fourier coefficients of the measure nu
  auto nuHat = loadMeasureFourierCoefficients("../data/nu_fhatT2.dat");
  normalizeNFFT2dFourierCoefficients( &nuHat );
  cout << "nuHat(0,0) = " << nuHat(0,0) << endl;

  // creates a new instance of the quadrature error function on the torus T^2
  // for polynomial degrees (N1,N2) and M points
  int M=2000;
  Eigen::Array2i degrees=nuHat.getPolynomialDegrees();
  QuadratureErrorT2 error(M, degrees);
  error.setMeasureFourierCoefficients(nuHat);
  error.setKernelFourierCoefficients( computeDiscrepancyKernelFourierCoefficients( degrees ) );

  // creates a shared pointer to a new instance of an array of M points on T^2
  auto sp_points = std::make_shared<Torus::TorusPointArray<double> > 
    ( error.getTorusParameterization().getTorus() );
  // registers the point array to the error function
  error.setPointArray( sp_points );

  // generates uniformly distributed random points on T^2 by projecting points
  // on S^3 onto T^2 = S^1 x S^1
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( 
    Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 3, M) );

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<double> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<double> cgMethod( sp_lineSearch );

  // runs the optimization method with initial guess given by the points on T^2
  cgMethod.run( &(*sp_points), error );

  saveCoordinatesArray( computeLocalCoordinatesArrayFromPointCoordinatesArray(
	     sp_points->getCoordinates(), error.getTorusParameterization() ),
			"test_ditherT2_PointCoordinates.dat");
  return 0;
}
