// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/ObjectiveFunctions/PolynomialT2Vectorized.hpp"
#include "../src/Manifolds/Torus/TorusPointArray.hpp"
#include "../src/Manifolds/Torus/TorusTangentVectorArray.hpp"
#include "../src/Utils/TestUtils.hpp"
#include "../src/Utils/Math.hpp"
#include "../src/OptimizationMethods/ArmijoMethodVectorized.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethodVectorized.hpp"


/** \file test_globalMinT2.cpp
   *
   *  \brief A simple example computing a global minimum of a given function on
   *  the torus by means of vectorized optimization using the 2-dimensional nfft
   *  and the nonlinear CG method.
   *
   */


using namespace std;
using namespace LORM;
using namespace ObjectiveFunctions;
using namespace NFFT;

/// test function on [-1,1]^2
double testFun( double x, double y ) 
{
  double xm=x-0.5, ym=y+0.2;
  return xm*xm+ym*ym+ abs(sin(2*xm)*cos(3*ym));
  //  return ( 1./4.*(x*x+y*y) + exp(sin(50*x)) + sin(60*exp(y)) + sin(70*sin(x)) + sin(sin(80*y)) + sin(10*(x+y)) );
}

/// rescaled and mirrored test function on [-0.5,0.5]^2
double computeMirroredRescaledTestFun( double x, double y) 
{
  double xs = 4*abs(x)-1;
  double ys = 4*abs(y)-1;
  return testFun(xs,ys);
}


NFFT2dFourierCoefficients computefHat( const Eigen::Array2i & degrees ) 
{
  int M1=2*degrees(0)+1, M2=2*degrees(1)+1;
  int M=M1*M2;
  NFFT2d nfft2dPlan( M, degrees );
  Eigen::Array2Xd localCoords;
  localCoords.resize(2,M);
  for( int i=0; i<M1; ++i)
    for( int j=0; j<M2; ++j) {
      localCoords(0,i*M2+j) = -0.5 + i / (double) M1;
      localCoords(1,i*M2+j) = -0.5 + j / (double) M2;
    }
  nfft2dPlan.setLocalPointCoordinatesArray( localCoords );
  Eigen::ArrayXd weightedFun;
  weightedFun.resize(M);
  for( int i=0; i<M; ++i)
    weightedFun(i) = computeMirroredRescaledTestFun( localCoords(0,i), localCoords(1,i) ) / (double) M;

  NFFT2dFourierCoefficients fHat = 
    nfft2dPlan.evalYMatrixAdjointMultiplication(weightedFun);  
  // enforce real-valued function
  for ( int k=-degrees(0); k<degrees(0); ++k )
    fHat(k,-degrees(1))=0;
  for ( int k=-degrees(1); k<degrees(1); ++k )
    fHat(-degrees(0),k)=0;
  
  // compute error on finer grid
  M1*=2; M2*=2; M=M1*M2;
  localCoords.resize(2,M);
  for( int i=0; i<M1; ++i)
    for( int j=0; j<M2; ++j) {
      localCoords(0,i*M2+j) = -0.5 + i / (double) M1;
      localCoords(1,i*M2+j) = -0.5 + j / (double) M2;
    }
  NFFT2d nfft2dPlanError( M, degrees );
  nfft2dPlanError.setLocalPointCoordinatesArray( localCoords );
  auto fun = nfft2dPlanError.evalYMatrixMultiplication( fHat );
  double errorInf = 0;
  for( int i=0; i<M; ++i) {
    double err = abs(fun(i).real()- computeMirroredRescaledTestFun( 
					   localCoords(0,i), localCoords(1,i) ));
    if( err > errorInf ) 
	errorInf = err;
  }
 
  cout << "approximation Error: " << errorInf << endl;
  return fHat;
}


int main() {
  
  // creates new instance of a trigonometric test functions which is computed by
  // means of the nfft
  int M1=10, M2=10, M=M1*M2;
  Eigen::Array2i degrees;
  degrees << 16,16;
  
  auto fHat = computefHat( degrees );
  PolynomialT2Vectorized polyVectorized(M, degrees);
  polyVectorized.setFourierCoefficients( fHat );

  // creates a shared pointer to a new instance of an array of M points on T^2
  auto sp_points = std::make_shared<Torus::TorusPointArray<double> > 
    ( polyVectorized.getTorusParameterization().getTorus() );
  // registers the point array to the test function
  polyVectorized.setPointArray( sp_points );

  // generate grid points
  Eigen::ArrayXXd localCoords;
  localCoords.resize(2,M);
  for( int i=0; i<M1; ++i)
    for( int j=0; j<M2; ++j) {
      localCoords(0,i*M2+j) = 2*M_PI* i / (double) M1;
      localCoords(1,i*M2+j) = 2*M_PI* j / (double) M2;
    } 
  computeManifoldPointArrayFromLocalCoordinatesArray( 
    &(*sp_points), localCoords, polyVectorized.getTorusParameterization() );
  
  // creates an shared pointer to a new instance of ArmijoMethodVectorized for
  // the vectorized line search in the optimization method
  auto sp_lineSearchVectorized = std::make_shared<ArmijoMethodVectorized<double> > ();
  // creates an instance of the vectorized conjugate gradient method with the
  // ArmijoMethodVectorized
  ConjugateGradientMethodVectorized<double> cgMethod( sp_lineSearchVectorized );
  
  // runs the optimization method with initial guess given by the points on T^2
  cgMethod.run( &(*sp_points), polyVectorized );
  
  double min=1e300;
  int indexMin=M;
  auto funValues = polyVectorized.evalFunction();
  for (int i=0; i<M; ++i)
    if ( funValues(i) < min ) {
      min = funValues(i);
      indexMin=i;
    }

  Eigen::Array2d x_min = polyVectorized.getTorusParameterization().evalInverseParameterization( 
       sp_points->getCoordinates().col(indexMin) )  /  (2*M_PI) - 0.5; 
  // x_min in [-0.5,0.5]^2 for mirrord function, so we need to rescale
  x_min(0) = 4*abs(x_min(0))-1;
  x_min(1) = 4*abs(x_min(1))-1;
 
  cout << endl << "Test Result" << endl;
  cout << "  x_min:" << endl << 0.5 << endl << -0.2 << endl;
  cout << "  x_min_approx:" << endl <<  x_min  << endl;
  cout << "  f(x_min): " << testFun(0.5,-0.2) << endl;
  cout << "  f(x_min_approx): " << testFun(x_min(0), x_min(1)) << endl;
  cout << "  f_approx(x_min): " << min << endl;
  
  return 0;
}
