// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Sphere/SphericalCoordinates.hpp"
#include "../src/Manifolds/Sphere/SpherePointArray.hpp"
#include "../src/NFFTwrapper/NFSFT.hpp"


/** \file test_nfsft.cpp
   *
   *  \brief A simple example demonstrating the functionality of the nfsft
   *  wrapper class LORM::NFFT::NFSFT.
   *
   */


int main () {

  using namespace std;
  using namespace LORM;
  using namespace NFFT;

  int M = 3; // number of points
  int N = 2; // polynomial degree

  // generates uniformly distributed random points on S^2
  std::default_random_engine randomGenerator;
  auto points = Sphere::generateRandomSpherePointCoordinatesArray<double>
    ( &randomGenerator, 2, M);
  // compute the local coordinates needed for the nfsft
  Sphere::SphericalCoordinates<double> paramS2;
  auto pointsLocal = computeLocalCoordinatesArrayFromPointCoordinatesArray( 
                      points, paramS2 ); 

  // creates new instance of an NFSFT plan with polynomial degree N and number of points M
  NFSFT nfsftPlan(M,N);
  // sets the point coordinates
  nfsftPlan.setLocalPointCoordinatesArray( pointsLocal );
 
  cout << "Initialized NFSFT plan with polynomial degree N=" << N 
       << " and number of Points M=" << M << endl;
  cout << "nfsft local point coordinats: " << endl 
       << nfsftPlan.getLocalPointCoordinatesArray() << endl << endl;

  // creates a new instance of NFSFT Fourier coefficients
  NFSFTFourierCoefficients fhat( Eigen::ArrayXcd::Random((N+1)*(N+1)) );
  // prints the transformations provided by the NFSFTplan
  cout << "Random Fourier coefficients, fhat:" << endl << fhat << endl;
  cout << "nfsft Trafo (Y * fhat):" << endl 
       << nfsftPlan.evalYMatrixMultiplication( fhat ) << endl;
  cout << "nfsft Nabla Trafo ( Nabla Y * fhat ):" << endl
       << nfsftPlan.evalNablaYMatrixMultiplication( fhat ) << endl;
  cout << "nfsft Nabla^2 Trafo( Nabla^2 Y * fhat ):" << endl
       << nfsftPlan.evalNabla2YMatrixMultiplication( fhat ) << endl << endl;

  // creats an array of function values
  Eigen::ArrayXcd f = Eigen::ArrayXcd::Random(M);
  // prints the adjoint transformations provided by the NFSFTplan
  cout << "Random (complex) Function values, f:" << endl << f << endl;
  cout << "nfsft Adjoint Trafo (Y' * f): " << endl 
       << nfsftPlan.evalYMatrixAdjointMultiplication( f ) << endl << endl;
  // creats an array of nabla function values
  Eigen::Array2Xd f2 = Eigen::Array2Xd::Random(2,M);
  cout << "Random (real) Nabla values, f2:" << endl << f2 << endl;
  cout << "nfsft Adjoint Nabla Trafo ( (Nabla Y)' * f2 ):" << endl
       << nfsftPlan.evalNablaYMatrixAdjointMultiplication( f2 ) << endl;

  return 0;
}
