// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroup/EulerAngles.hpp"
#include "../src/Manifolds/RotationGroup/RotationGroupPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/NFFTwrapper/NFSOFT.hpp"

/** \file test_nfsoft.cpp
   *
   *  \brief A simple example demonstrating the functionality of the nfsoft
   *  wrapper class LORM::NFFT::NFSOFT.
   *
   */

int main () {

  using namespace std;
  using namespace LORM;
  using namespace NFFT;

  int M = 3; // number of points
  int N = 2; // polynomial degree

  // generates uniformly distributed random points on SO(3) in R^(3x3) by
  // projecting uniformly distributed random points on S^8 onto SO(3)
  std::default_random_engine randomGenerator;
  auto points = Sphere::generateRandomSpherePointCoordinatesArray<double>
    ( &randomGenerator, 8, M);
  // compute the local coordinates needed for the nfsft
  RotationGroup::EulerAngles<double> paramSO3;
  auto pointsLocal = computeLocalCoordinatesArrayFromPointCoordinatesArray( 
                      points, paramSO3 ); 

  // creates new instance of an NFSOFT plan with polynomial degree N and number
  // of points M
  NFSOFT nfsoftPlan(M,N);
  // sets the point coordinates
  nfsoftPlan.setLocalPointCoordinatesArray( pointsLocal );

  cout << "Initialized NFSOFTplan with polynomial degree N=" << N 
       << " and number of Points M=" << M << endl;
  cout << "nfsoft local point coordinats: " << endl 
       << nfsoftPlan.getLocalPointCoordinatesArray() << endl << endl;

  // creates a new instance of NFSOFT Fourier coefficients
  NFSOFTFourierCoefficients fhat( Eigen::ArrayXcd::Random((2*N+1)*(2*N+2)*(2*N+3)/6) );
  // prints the transformations provided by the NFSOFTplan
  cout << "Random Fourier coefficients, fhat:" << endl << fhat << endl;
  cout << "nfsoft Trafo (Y * fhat):" << endl 
       << nfsoftPlan.evalYMatrixMultiplication( fhat ) << endl;
  cout << "nfsoft Nabla Trafo ( Nabla Y * fhat ):" << endl
       << nfsoftPlan.evalNablaYMatrixMultiplication( fhat ) << endl;
  cout << "nfsoft Nabla^2 Trafo( Nabla^2 Y * fhat ):" << endl
       << nfsoftPlan.evalNabla2YMatrixMultiplication( fhat ) << endl << endl;

  // creats an array of function values
  Eigen::ArrayXcd f = Eigen::ArrayXcd::Random(M);
  // prints the adjoint transformations provided by the NFSOFTplan
  cout << "Random (complex) Function values, f:" << endl << f << endl;
  cout << "nfsoft Adjoint Trafo (Y' * f): " << endl 
       << nfsoftPlan.evalYMatrixAdjointMultiplication( f ) << endl << endl;
  // creats an array of nabla function values
  Eigen::Array3Xd f2 = Eigen::Array3Xd::Random(3,M);
  cout << "Random (real) Nabla values, f2:" << endl << f2 << endl;
  cout << "nfsoft Adjoint Nabla Trafo ( (Nabla Y)' * f2 ):" << endl
       << nfsoftPlan.evalNablaYMatrixAdjointMultiplication( f2 ) << endl;

  return 0;
}
