// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroup/EulerAngles.hpp"
#include "../src/Manifolds/RotationGroup/RotationGroupPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/NFFTwrapper/NFSOFT.hpp"
#include "../src/ObjectiveFunctions/PolynomialSO3Vectorized.hpp"
#include "../src/Utils/TestUtils.hpp"
#include "../src/OptimizationMethods/ArmijoMethodVectorized.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethodVectorized.hpp"

/** \file test_polynomialSO3Vectorized.cpp
   *
   *  \brief A simple example demonstrating the vectorized optimization of a
   *  polynomial on the rotation group SO(3) by using the nfsoft and the
   *  nonlinear CG method.
   *
   */

int main () {

  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  using namespace NFFT;
  
  // creates new instance of a polynomial test functions which is computed by
  // means of the nfsoft
  int M = 6;
  int N = 2;
  PolynomialSO3Vectorized polyVectorized(M, N);

  // creates a shared pointer to a new instance of an array of M points on SO(3)
  auto sp_points = std::make_shared<RotationGroup::RotationGroupPointArray<double> > 
    ( polyVectorized.getEulerAngles().getRotationGroup() );
  // registers the point array to the test function
  polyVectorized.setPointArray( sp_points );

  // generates uniformly distributed random points on SO(3) in R^(3x3) by
  // projecting uniformly distributed random points on S^8 onto SO(3)
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( 
    Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 8, M) );

  // prints the point coordinates, funtion values and gradients
  cout << "Point Coordinates: " << endl 
       << sp_points->getCoordinates() << endl;
  cout << "Function Values: " << endl
       << polyVectorized.evalFunction() << endl;
  cout << "Function Gradients: " << endl 
       << polyVectorized.evalGradientArray().getCoordinates() << endl;

  // creates an array of tangent vectors associated to the array of points 
  RotationGroup::RotationGroupTangentVectorArray<double> tangents ( sp_points );
  // generates uniformly distributed random tangent vectors by projecting
  // uniformly distrubted points on S^8 in R^(3x3) onto the corresponding
  // tangent spaces at the points
  tangents.setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<double>
			   ( &randomGenerator, 8, M) );
  
  // outputs a data file with 1000 equidistant sampling values of the vectorized
  // function and its quadratic approximation along the geodesics starting at
  // the points in direction of the tangent vectors
  evalLineFunctionVectorizedWithQuadraticApproximation ( 
     polyVectorized, *sp_points, tangents, 0.1, 1000, 
     "test_PolynomialSO3VectorizedQuadraticApproximation.dat" );

  // creates an shared pointer to a new instance of ArmijoMethodVectorized for
  // the vectorized line search in the optimization method
  auto sp_lineSearchVectorized = std::make_shared<ArmijoMethodVectorized<double> > ();
  // creates an instance of the vectorized conjugate gradient method with the
  // ArmijoMethodVectorized
  ConjugateGradientMethodVectorized<double> cgMethod( sp_lineSearchVectorized );
  
  // runs the optimization method with initial guess given by the points on SO(3)
  cgMethod.run( &(*sp_points), polyVectorized );
  
  // prints the point coordinates, funtion values and gradients
  cout << endl << "Computed Minimizers" << endl;
  cout << "Point Coordinates: " << endl 
       << sp_points->getCoordinates() << endl;
  cout << "Function Values: " << endl
       << polyVectorized.evalFunction() << endl;
  cout << "Function Gradients: " << endl 
       << polyVectorized.evalGradientArray().getCoordinates() << endl;

  return 0;

}
