// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/ObjectiveFunctions/PolynomialT2Vectorized.hpp"
#include "../src/Manifolds/Torus/TorusPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/Manifolds/Torus/TorusTangentVectorArray.hpp"
#include "../src/Utils/TestUtils.hpp"
#include "../src/OptimizationMethods/ArmijoMethodVectorized.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethodVectorized.hpp"

/** \file test_polynomialT2Vectorized.cpp
   *
   *  \brief A simple example demonstrating the vectorized optimization of a
   *  trigonometric polynomial on the torus T^2 by using the 2-dimensional nfft
   *  and the nonlinear CG method.
   *
   */

int main() {
  
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  using namespace NFFT;
  
  // creates new instance of a trigonometric test functions which is computed by
  // means of the nfft
  int M = 6;
  Eigen::Array2i degrees;
  degrees << 5,3;
  PolynomialT2Vectorized polyVectorized(M, degrees);
  
  // creates a shared pointer to a new instance of an array of M points on T^2
  auto sp_points = std::make_shared<Torus::TorusPointArray<double> > 
    ( polyVectorized.getTorusParameterization().getTorus() );
  // registers the point array to the test function
  polyVectorized.setPointArray( sp_points );

  // generates uniformly distributed random points on T^2 by projecting points
  // on S^3 onto T^2
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( 
    Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 3, M) );

  // prints the point coordinates, funtion values and gradients
  cout << "Point Coordinates: " << endl 
       << sp_points->getCoordinates() << endl;
  cout << "Function Values: " << endl
       << polyVectorized.evalFunction() << endl;
  cout << "Function Gradients: " << endl 
       << polyVectorized.evalGradientArray().getCoordinates() << endl;

  // creates an array of tangent vectors associated to the array of points 
  Torus::TorusTangentVectorArray<double> tangents ( sp_points );
  // generates uniformly distributed random tangent vectors by projecting
  // uniformly distrubted points on S^3 onto the corresponding tangent spaces at
  // the points
  tangents.setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<double>
			   ( &randomGenerator, 3, M) );

  // outputs a data file with 1000 equidistant sampling values of the vectorized
  // function and its quadratic approximation along the geodesics starting at
  // the points in direction of the tangent vectors
  evalLineFunctionVectorizedWithQuadraticApproximation ( 
     polyVectorized, *sp_points, tangents, 1.0, 1000, 
     "test_PolynomialT2VectorizedQuadraticApproximation.dat" );

  // creates an shared pointer to a new instance of ArmijoMethodVectorized for
  // the vectorized line search in the optimization method
  auto sp_lineSearchVectorized = std::make_shared<ArmijoMethodVectorized<double> > ();
  // creates an instance of the vectorized conjugate gradient method with the
  // ArmijoMethodVectorized
  ConjugateGradientMethodVectorized<double> cgMethod( sp_lineSearchVectorized );
  
  // runs the optimization method with initial guess given by the points on T^2
  cgMethod.run( &(*sp_points), polyVectorized );
  
  // prints the point coordinates, funtion values and gradients
  cout << endl << "Computed Minimizers" << endl;
  cout << "Point Coordinates: " << endl 
       << sp_points->getCoordinates() << endl;
  cout << "Function Values: " << endl
       << polyVectorized.evalFunction() << endl;
  cout << "Function Gradients: " << endl 
       << polyVectorized.evalGradientArray().getCoordinates() << endl;

  return 0;
}
