// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/SphereTimesR/SphereTimesRPointArray.hpp"
#include "../src/ObjectiveFunctions/QuadratureErrorS2R.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/SteepestDescentMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_quadratureS2R.cpp
   *
   *  \brief A simple example demonstrating the computation of optimal
   *  quadrature points and weights on the sphere S^2 in double precision by
   *  LORM::ObjectiveFunctions::QuadratureErrorS2R.
   *
   */

int main () 
{
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  
  // creates a new instance of the quadrature error function on the sphere S^2
  // for polynomial degree N and M points with weights
  int M=6, N=3;
  QuadratureErrorS2R error(M, N);
  
  // creates a shared pointer to a new instance of an array of M points on S^2 x R
  auto sp_points = std::make_shared<Sphere::SphereTimesRPointArray<double> > 
    ( error.getSphericalCoordinatesTimesR().getSphereTimesR() );
  // registers the point array to the error function
  error.setPointArray( sp_points );
    
  // generates uniformly distributed random points on S^2 x R by projection of
  // uniformly distributed random points on S^3 onto S^2 x R 
  std::default_random_engine randomGenerator;
  auto coords = Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 3, M);
  sp_points->setCoordinates( coords );

  // prints the quadrature error and the gradients
  cout << "error: " << error.evalFunction() << endl;
  cout << "grad:  " << endl << error.evalGradientArray().getCoordinates() << endl;

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<double> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<double> cgMethod( sp_lineSearch );

  // runs the optimization method with initial guess given by the points on S2 x R
  cgMethod.run( &(*sp_points), error );
  
  return 0;
}
