// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroup/RotationGroupPointArray.hpp"
#include "../src/ObjectiveFunctions/QuadratureErrorSO3.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_quadratureSO3.cpp
   *
   *  \brief A simple example demonstrating the computation of optimal
   *  quadrature points with equal weights on the rotation group SO(3) in double
   *  precision by LORM::ObjectiveFunctions::QuadratureErrorSO3.
   *
   */


int main () 
{
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;

  // creates a new instance of the quadrature error function on the rotation
  // group SO(3) for polynomial degree N and M points
  int M=24, N=3;
  QuadratureErrorSO3 error(M, N);

  // creates a shared pointer to a new instance of an array of M points on SO(3)
  auto sp_points = std::make_shared<RotationGroup::RotationGroupPointArray<double> > 
    ( error.getEulerAngles().getRotationGroup() );
  // registers the point array to the error function
  error.setPointArray( sp_points );
   
  // generates uniformly distributed random points on SO(3) in R^(3x3) by
  // projecting uniformly distributed random points on S^8 onto SO(3)
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( 
    Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 8, M) );

// creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<double> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<double> cgMethod( sp_lineSearch );

  // runs the optimization method with initial guess given by the points on SO(3)
  cgMethod.run( &(*sp_points), error );

  return 0;
}
