// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroupTimesR/RotationGroupTimesRPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/ObjectiveFunctions/QuadratureErrorSO3R.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_quadratureSO3R.cpp
   *
   *  \brief A simple example demonstrating the computation of optimal
   *  quadrature points and weights on the rotation group SO(3) in double
   *  precision by LORM::ObjectiveFunctions::QuadratureErrorSO3R.
   *
   */

int main () 
{
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  
  // creates a new instance of the quadrature error function on the rotation
  // group SO(3) for polynomial degree N and M points with weights
  int M=4, N=1;
  QuadratureErrorSO3R error(M, N);
  
  // creates a shared pointer to a new instance of an array of M points on SO(3) x R
  auto sp_points = std::make_shared<RotationGroup::RotationGroupTimesRPointArray<double> > 
    ( error.getEulerAnglesTimesR().getRotationGroupTimesR() );
  // registers the point array to the error function
  error.setPointArray( sp_points );
    
  // generates uniformly distributed random points on SO(3) x R in R^(3x3) x R
  // by projection of uniformly distributed random points on S^9 onto SO(3) x R
  std::default_random_engine randomGenerator;
  auto coords = Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 9, M);
  sp_points->setCoordinates( coords );

  // prints the quadrature error and the gradients
  cout << "error: " << error.evalFunction() << endl;
  cout << "grad:  " << endl << error.evalGradientArray().getCoordinates() << endl;

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<double> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<double> cgMethod( sp_lineSearch );

  // runs the optimization method with initial guess given by the points on SO(3) x R
  cgMethod.run( &(*sp_points), error );

  // prints the coordinates of the local minimizer
  cout << endl 
       << "Coordinates of the local Minimizer:" << endl 
       << sp_points->getCoordinates() << endl;

  return 0;
}
