// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Torus/TorusPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/ObjectiveFunctions/QuadratureErrorT2.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_quadratureT2.cpp
   *
   *  \brief A simple example demonstrating the computation of optimal
   *  quadrature points with equal weights on the torus T^2 in double precision
   *  by LORM::ObjectiveFunctions::QuadratureErrorT2.
   *
   */

int main () 
{
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;

  // creates a new instance of the quadrature error function on the torus T^2
  // for polynomial degrees (N1,N2) and M points
  int M=60;
  Eigen::Array2i degrees;
  degrees << 5,5; // bug?! degrees must be greater than 2 or so?
  QuadratureErrorT2 error(M, degrees);

  // creates a shared pointer to a new instance of an array of M points on T^2
  auto sp_points = std::make_shared<Torus::TorusPointArray<double> > 
    ( error.getTorusParameterization().getTorus() );
  // registers the point array to the error function
  error.setPointArray( sp_points );
  
  // generates uniformly distributed random points on T^2 by projecting points
  // on S^3 onto T^2 = S^1 x S^1
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( 
    Sphere::generateRandomSpherePointCoordinatesArray<double> 
    (&randomGenerator, 3, M) );

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<double> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<double> cgMethod( sp_lineSearch );

  // runs the optimization method with initial guess given by the points on T^2
  cgMethod.run( &(*sp_points), error );

  return 0;
}
