// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/RotationGroup/RotationGroupPointArray.hpp"
#include "../src/Manifolds/Sphere/Sphere.hpp"
#include "../src/ObjectiveFunctions/CoulombEnergy.hpp"
#include "../src/Utils/TestUtils.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_rotationGroup.cpp
   *
   *  \brief A simple example demonstrating the use of several functions related
   *  to the rotation group SO(n). Ranging from the initialization of uniformly
   *  random distributed points to the optimization of the Coulomb energy by the
   *  nonlinear CG method.
   *
   */

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
typedef LORM_INSTANTIATE_SCALAR_TYPE T_scalar;
#else
typedef double T_scalar;
#endif

int main() {
  
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;

  const int n=2; // parameter n for SO(n)
  const int M=5; // number of points 
  
  // creates a shared pointer to a new instance of SO(n)
  auto sp_rotationGroup = make_shared<RotationGroup::RotationGroup>(n);

  // prints general informations of the manifold object
  cout << "Manifold name: " << sp_rotationGroup->getName() << endl;
  cout << "Manifold dimension: " << sp_rotationGroup->getDimension() << endl;
  cout << "Manifold ambient space dimension: " 
       << sp_rotationGroup->getAmbientSpaceDimension() << endl << endl;

  // creates a shared pointer to a new instance of an array of M points on SO(n)
  auto sp_points = make_shared<RotationGroup::RotationGroupPointArray<T_scalar> >
    ( sp_rotationGroup );
  // generates uniformly distributed random points on SO(n) by projecting
  // uniformly distrubted points on S^(n^2-1) in R^(n x n) onto SO(n)
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<T_scalar>
			     ( &randomGenerator, n*n-1, M) );

  // prints the Cartesian point coordinates of the points
  cout << "Point Coords: " << endl << sp_points->getCoordinates() << endl;
  
  // checks the orthogonality of the first point
  Eigen::Map<const Eigen::Matrix<T_scalar, Eigen::Dynamic, Eigen::Dynamic> > map 
    ( sp_points->getCoordinates().data(), n, n );
  cout << "Orthogonality test of the first point: " << endl 
       << map * map.transpose() << endl << endl;
  
  // creates the function object for the Coulomb energy of the points, by
  // registering the point array
  CoulombEnergy<T_scalar> energy( sp_points );

  // prints the energy of the points
  cout << "Coulomb energy: " << endl 
       << energy.evalFunction() << endl;
  // prints the force of every single point as a vector in Cartesian coordinates
  cout << "Coulomb forces: " << endl 
       << energy.evalGradientArray().getCoordinates() << endl << endl;

  // creates an array of tangent vectors associated to the array of points 
  RotationGroup::RotationGroupTangentVectorArray<T_scalar> tangents ( sp_points );
  // generates uniformly distributed random tangent vectors by projecting
  // uniformly distrubted points on S^(n^2-1) in R^(n x n) onto the
  // corresponding tangent spaces at the points on SO(n)
  tangents.setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<T_scalar>
			   ( &randomGenerator, n*n-1, M) );
  
  // outputs a data file with 1000 equidistant sampling values of the Coulomb
  // energy and its quadratic approximation along a geodesic starting at the
  // points in direction of the tangent vectors
  evalLineFunctionWithQuadraticApproximation ( 
      energy, *sp_points, tangents, (T_scalar) 10., 1000, 
      "test_rotationGroup_CoulombEnergyQuadraticApproximation.dat" );

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method.
  auto sp_lineSearch = std::make_shared<ArmijoMethod<T_scalar> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<T_scalar> cgMethod( sp_lineSearch );
  
  // runs the optimization method with initial guess given by the points on SO(n)
  cgMethod.run( &(*sp_points), energy );

  // plots the rotation angles of the minimizers if n==2, which should be equidistant
  if (n==2) {
    cout << "Normalized Rotation angles of the points at the minimum:" << endl;
    T_scalar alpha_min = M_PI;
    Eigen::Array<T_scalar, 1, M> alphaArray;
    for ( int i=0; i<M; ++i ) {
      alphaArray(i) = 2 * atan(  sp_points->getCoordinates()(3,i) / 
				(sp_points->getCoordinates()(2,i) - 1) ) + M_PI;
      if( alphaArray(i) < alpha_min )
	alpha_min = alphaArray(i);
    }
    for ( int i=0; i<M; ++i ) 
      cout << "x(" << i+1 << "): " << (alphaArray(i) - alpha_min) / (2*M_PI)  << endl;
  }
    
  return 0;
}
