// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Sphere/SpherePointArray.hpp"
#include "../src/ObjectiveFunctions/CoulombEnergy.hpp"
#include "../src/Utils/TestUtils.hpp"
#include "../src/OptimizationMethods/ArmijoMethod.hpp"
#include "../src/OptimizationMethods/ConjugateGradientMethod.hpp"

/** \file test_sphere.cpp
   *
   *  \brief A simple example demonstrating the use of several functions related
   *  to the sphere S^d. Ranging from the initialization of uniformly random
   *  distributed points to the optimization of the Coulomb energy by the
   *  nonlinear CG method.
   *
   */

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
typedef LORM_INSTANTIATE_SCALAR_TYPE T_scalar;
#else
typedef double T_scalar;
#endif

int main() {
  
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;

  const int d=1; // dimension d of the sphere S^d
  const int M=5; // number of points 
  
  // creates a shared pointer to a new instance of S^d
  auto sp_sphere = make_shared<Sphere::Sphere>(d);

  // prints general informations of the manifold object
  cout << "Manifold name: " << sp_sphere->getName() << endl;
  cout << "Manifold dimension: " << sp_sphere->getDimension() << endl;
  cout << "Manifold ambient space dimension: " 
       << sp_sphere->getAmbientSpaceDimension() << endl << endl;

  // creates a shared pointer to a new instance of an array of M points on S^d
  auto sp_points = make_shared<Sphere::SpherePointArray<T_scalar> >
    ( sp_sphere );
  // generates uniformly distributed random points on S^d
  std::default_random_engine randomGenerator;
  sp_points->setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<T_scalar>
			     ( &randomGenerator, d, M) );

  // prints the Cartesian point coordinates of the points
  cout << "Point Coords: " << endl << sp_points->getCoordinates() << endl;
  
  // checks the norm of the points
  cout << "Norm test of the points: " << endl 
       << sp_points->getCoordinates().square().colwise().sum() << endl << endl;
  
  // creates the function object for the Coulomb energy of the points, by
  // registering the point array
  CoulombEnergy<T_scalar> energy( sp_points );

  // prints the energy of the points
  cout << "Coulomb energy: " << endl 
       << energy.evalFunction() << endl;
  // prints the force of every single point as a vector in Cartesian coordinates
  cout << "Coulomb forces: " << endl 
       << energy.evalGradientArray().getCoordinates() << endl << endl;

  // creates an array of tangent vectors associated to the array of points 
  Sphere::SphereTangentVectorArray<T_scalar> tangents ( sp_points );
  // generates uniformly distributed random tangent vectors by projecting
  // uniformly distrubted points on S^d onto the corresponding tangent spaces at
  // the points
  tangents.setCoordinates( Sphere::generateRandomSpherePointCoordinatesArray<T_scalar>
			   ( &randomGenerator, d, M) );
  
  // outputs a data file with 1000 equidistant sampling values of the Coulomb
  // energy and its quadratic approximation along a geodesic starting at the
  // points in direction of the tangent vectors
  evalLineFunctionWithQuadraticApproximation ( 
      energy, *sp_points, tangents, (T_scalar) 10., 1000, 
      "test_sphere_CoulombEnergyQuadraticApproximation.dat" );

  // creates an shared pointer to a new instance of ArmijoMethod for the line
  // search in the optimization method
  auto sp_lineSearch = std::make_shared<ArmijoMethod<T_scalar> > ();
  // creates an instance of the conjugate gradient method with the ArmijoMethod
  ConjugateGradientMethod<T_scalar> cgMethod( sp_lineSearch );
  
  // runs the optimization method with initial guess given by the points on S^d
  cgMethod.run( &(*sp_points), energy );

  // plots the rotation angles of the minimizers if d==1, which should be equidistant
  if (d==1) {
    cout << "Normalized angles of the points at the minimum:" << endl;
    T_scalar alpha_min = M_PI;
    Eigen::Array<T_scalar, 1, M> alphaArray;
    for ( int i=0; i<M; ++i ) {
      alphaArray(i) = 2 * atan(  sp_points->getCoordinates()(1,i) / 
				(sp_points->getCoordinates()(0,i) - 1) ) + M_PI;
      if( alphaArray(i) < alpha_min )
	alpha_min = alphaArray(i);
    }
    for ( int i=0; i<M; ++i ) 
      cout << "x(" << i+1 << "): " << (alphaArray(i) - alpha_min) / (2*M_PI)  << endl;
  }
    
  return 0;
}
