// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Sphere/SphericalCoordinates.hpp"
#include "../src/ObjectiveFunctions/TestFunctions.hpp"
#include "../src/Manifolds/Sphere/SpherePointArray.hpp"
#include "../src/Utils/TestUtils.hpp"

/** \file test_sphericalCoordinates.cpp
   *
   *  \brief A simple example demonstrating the parameterization in spherical
   *  coordinates by LORM::ObjectiveFunctions::TestFunctionSphereS2.
   *
   */

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
typedef LORM_INSTANTIATE_SCALAR_TYPE T_scalar;
#else
typedef double T_scalar;
#endif

int main() {
  
  using namespace std;
  using namespace LORM;
  using namespace ObjectiveFunctions;
  
  // creates an instance of an test function which makes use of spherical
  // coordinates
  TestFunctionSphereS2<T_scalar> fun;
    
  // gets the associated spherical coordinates parameterization
  auto & paramS2 = fun.getSphericalCoordinates();
  // gets the associated S^2 manifold
  auto & sp_sphere = paramS2.getSphere();
  
  // prints general informations of the manifold object
  cout << "Manifold name: " << sp_sphere->getName() << endl;
  cout << "Manifold dimension: " << sp_sphere->getDimension() << endl;
  cout << "Manifold ambient space dimension: " 
       << sp_sphere->getAmbientSpaceDimension() << endl << endl;
  
  // generates random spherical Coordinates of a single point
  Eigen::Array<T_scalar, 2, 1> localCoords;
  localCoords.setRandom();
  // computes the cooresponding Cartesian coordinates
  auto pointCoords = paramS2.evalParameterization( localCoords );

  // prints the results of the conversion between local and Cartesian coordinates
  cout << "Spherical coordinates: " << endl << localCoords << endl;
  cout << "Cartesian coordinates: " << endl << pointCoords << endl;
  cout << "Computed spherical coordinates: " << endl 
       << paramS2.evalInverseParameterization( pointCoords ) << endl;
  pointCoords = paramS2.evalParameterization( 
	           paramS2.evalInverseParameterization( pointCoords ) );
  cout << "Recomputed Cartesian coordinates: " << endl <<  pointCoords << endl << endl;

  // prints several matrices needed for Riemannian geometry computations
  cout << "Jacobi Matrix: " << endl 
       << paramS2.evalJacobiMatrix( localCoords ) << endl;
  cout << "inverse Riemann Matrix: " << endl 
       << paramS2.evalInverseRiemannianMatrix( localCoords ) << endl;
  // generates random local coefficients for the Christoffel matrix
  Eigen::Array<T_scalar, 2, 1> localCoeffs;
  localCoeffs.setRandom();
  cout << "Christoffel Matrix linear combination: " << endl
       << paramS2.evalChristoffelMatrixLinearCombination( localCoeffs, localCoords ) 
       << endl << endl;

  // creates a shared pointer to new instance of an array of points on S^2 of size M
  const int M = 10;
  auto sp_points = make_shared<Sphere::SpherePointArray<T_scalar> >
    ( sp_sphere );
  // creates an associated array of tangent vectors 
  Sphere::SphereTangentVectorArray<T_scalar> tangents( sp_points);
  // sets random points and random tangent vectors
  sp_points->setCoordinates( Eigen::Array<T_scalar, 3, M>::Random() );
  tangents.setCoordinates( Eigen::Array<T_scalar, 3, M>::Random() );
    
  cout << "Test function in spherical coordinates parameterization:" << endl
       << " number of Points: " << M << endl;  

  // outputs a data file with 1000 equidistant sampling values of the the test
  // function and its quadratic approximation along a geodesic starting at the
  // points in direction of the tangent vectors
  evalLineFunctionWithQuadraticApproximation ( 
     fun, *sp_points, tangents, (T_scalar) 1., 1000, 
     "test_sphericalCoordinates_TestFunctionQuadraticApproximation.dat");

  return 0;
}
