// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "../src/Manifolds/Torus/TorusParameterization.hpp"
#include "../src/Manifolds/Torus/TorusPointArray.hpp"

/** \file test_torusParameterization.cpp
   *
   *  \brief A simple example demonstrating the parameterization of the torus
   *  T^d.
   *
   */

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
typedef LORM_INSTANTIATE_SCALAR_TYPE T_scalar;
#else
typedef double T_scalar;
#endif

int main() {
  
  using namespace std;
  using namespace LORM;
  
  int d=2;
  
  // creates the parameterization
  Torus::TorusParameterization<T_scalar> paramTd(d);
  // gets the associated T^d manifold
  auto & sp_torus = paramTd.getTorus();
  
  // prints general informations of the manifold object
  cout << "Manifold name: " << sp_torus->getName() << endl;
  cout << "Manifold dimension: " << sp_torus->getDimension() << endl;
  cout << "Manifold ambient space dimension: " 
       << sp_torus->getAmbientSpaceDimension() << endl << endl;
  
  // generates random local coordinates of a single point
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> localCoords;
  localCoords.setRandom(d);
  // computes the cooresponding Cartesian coordinates
  auto pointCoords = paramTd.evalParameterization( localCoords );

  // prints the results of the conversion between local and Cartesian coordinates
  cout << "Local coordinates: " << endl << localCoords << endl;
  cout << "Cartesian coordinates: " << endl << pointCoords << endl;
  cout << "Computed local coordinates:" << endl 
       << paramTd.evalInverseParameterization( pointCoords ) << endl;
  pointCoords = paramTd.evalParameterization( 
	           paramTd.evalInverseParameterization( pointCoords ) );
  cout << "Recomputed Cartesian coordinates: " << endl <<  pointCoords << endl << endl;

  // prints several matrices needed for Riemannian geometry computations
  cout << "Jacobi Matrix: " << endl 
       << paramTd.evalJacobiMatrix( localCoords ) << endl;
  cout << "inverse Riemann Matrix: " << endl 
       << paramTd.evalInverseRiemannianMatrix( localCoords ) << endl;
  // generates random local coefficients for the Christoffel matrix
  Eigen::Array<T_scalar, Eigen::Dynamic, 1> localCoeffs;
  localCoeffs.setRandom(d);
  cout << "Christoffel Matrix linear combination: " << endl
       << paramTd.evalChristoffelMatrixLinearCombination( localCoeffs, localCoords ) 
       << endl << endl;
    
  return 0;
}
