// This file is part of LORM, a  C++ library for optimization on 
// Riemannian manifolds.
//
// Copyright (C) 2013 Manuel Gräf <m.graef@mathematik.tu-chemnitz.de>
//
// This source code form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
#include "ManifoldFunctionSecondOrderParameterized.hpp"

using namespace LORM;

template<typename T_scalar, typename T_return>
ManifoldFunctionSecondOrderParameterized<T_scalar, T_return>::ManifoldFunctionSecondOrderParameterized(
  const std::shared_ptr<const ManifoldParameterization<T_scalar> > & sp_parameterization ) :
  ManifoldFunctionSecondOrder<T_scalar, T_return> (),
  sp_parameterization_( sp_parameterization )
{}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrderParameterized<T_scalar, T_return>::computeFx( 
  T_return * p_fx, const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const
{
  computeLocalPointCoordsArray( pointCoordsArray );
  computeFxParameterized( p_fx, localPointCoordsArray_ );
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrderParameterized<T_scalar, T_return>::computeNablaFxArray( 
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nablaFxArray,
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray) const
{
  computeLocalPointCoordsArray( pointCoordsArray );
  nablaFxParameterizedArray_.resize( localPointCoordsArray_.rows(), localPointCoordsArray_.cols() );
  computeNablaFxParameterizedArray( &nablaFxParameterizedArray_, localPointCoordsArray_ );
  
  for ( int i=0; i< localPointCoordsArray_.cols(); ++i ) {
    auto invRiemannMatrix = sp_parameterization_->evalInverseRiemannianMatrix( localPointCoordsArray_.col(i) );
    auto jacobiMatrix = sp_parameterization_->evalJacobiMatrix( localPointCoordsArray_.col(i) );
    // nabla f = D_h * G_h^(-1) * nabla_h f
    p_nablaFxArray->col(i) = jacobiMatrix * invRiemannMatrix * nablaFxParameterizedArray_.col(i).matrix();
  }
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrderParameterized<T_scalar, T_return>::computeNabla2FxMultiplication( 
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> * p_nabla2FxMultTangentArray,
  const Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & pointCoordsArray ) const
{
  // NOTE: computeNablaFxArray() has been already executed (see evalNabla2FxMultiplication), i.e., 
  //       nablaFxArray_ is already computed
  // However, we may need to recompute the local coordinates!
  computeLocalPointCoordsArray( pointCoordsArray );
  
  // the following is dirty!!
  // we exploit the temporary array so that we save the creation of another array
  Eigen::Array<T_scalar, Eigen::Dynamic, Eigen::Dynamic> & nabla2FxParameterizedMultTangentArray
    = nablaFxParameterizedArray_; 
  nabla2FxParameterizedMultTangentArray.resize( localPointCoordsArray_.rows(), localPointCoordsArray_.cols() );
  
  // The following loop computes the matrix multiplication with the Christoffel matrix
  for ( int i=0; i<localPointCoordsArray_.cols(); ++i ) {
    // compute Jacobian and Riemannian Matrix
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords = localPointCoordsArray_.col(i);
    auto jacobiMatrix = sp_parameterization_->evalJacobiMatrix( localPointCoords );
    auto invRiemannMatrix =	sp_parameterization_->evalInverseRiemannianMatrix( localPointCoords );
    
    // compute gradient of parameterized function
    // nabla_h f = D_h^T * (D_h * G_h^(-1) * nabla_h f) = D_h^T * nabla f 
    Eigen::Array<T_scalar, Eigen::Dynamic, 1> nablaFxParameterized = 
      jacobiMatrix.transpose() * this->getNablaFxArray().col(i).matrix();
    
    // compute Christoffel matrix 
    auto christoffelMatrix = sp_parameterization_->evalChristoffelMatrixLinearCombination( nablaFxParameterized, localPointCoords );
    
    // compute local coordinates of nabla2FxMultTangentArray
    // v_h = G_h^(-1) * D_h^T * v
    nabla2FxParameterizedMultTangentArray.col(i) =  invRiemannMatrix * jacobiMatrix.transpose() 
      * p_nabla2FxMultTangentArray->col(i).matrix();
    
    // set tangent vector coordinates to negative Christoffel matrix multiplication
    p_nabla2FxMultTangentArray->col(i) = - jacobiMatrix * invRiemannMatrix * christoffelMatrix * 
      nabla2FxParameterizedMultTangentArray.col(i).matrix();
  }
  
  // apply Hessian matrix multiplication for local tangent vector coordinates
  computeNabla2FxParameterizedMultiplication( &nabla2FxParameterizedMultTangentArray, localPointCoordsArray_ );
  
  // the following loop adds the multiplication with the Hessian matrix
  for ( int i=0; i<localPointCoordsArray_.cols(); ++i ) {
    // compute Jacobian and Riemannian Matrix
    const Eigen::Array<T_scalar, Eigen::Dynamic, 1> & localPointCoords = localPointCoordsArray_.col(i);
    auto jacobiMatrix = sp_parameterization_->evalJacobiMatrix( localPointCoords );
    auto invRiemannMatrix = sp_parameterization_->evalInverseRiemannianMatrix( localPointCoords );
    
    // add Hessian matrix multiplication to tangent vector coordinates
    p_nabla2FxMultTangentArray->col(i).matrix() += jacobiMatrix * invRiemannMatrix 
      * nabla2FxParameterizedMultTangentArray.col(i).matrix();
  }
}

template<typename T_scalar, typename T_return>
void ManifoldFunctionSecondOrderParameterized<T_scalar, T_return>::computeLocalPointCoordsArray ( 
  const Eigen::Array<T_scalar, Eigen::Dynamic,Eigen::Dynamic> & pointCoordsArray ) const
{
  localPointCoordsArray_.swap( computeLocalCoordinatesArrayFromPointCoordinatesArray( pointCoordsArray, *sp_parameterization_ ) ); 
}

//Macro for explicit instantiations of ManifoldFunctionSecondOrderParameterized<T_scalar>
#define LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDERPARAMETERIZED(T_scalar)	\
  template class ManifoldFunctionSecondOrderParameterized<T_scalar>; \
  template class ManifoldFunctionSecondOrderParameterized<T_scalar, Eigen::Array<T_scalar, 1, Eigen::Dynamic> >; 

#if defined( LORM_INSTANTIATE_SCALAR_TYPE )
  LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDERPARAMETERIZED( LORM_INSTANTIATE_SCALAR_TYPE )
#else
  LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDERPARAMETERIZED(double)
#endif

#undef LORM_INSTANTIATE_MANIFOLDFUNCTIONSECONDORDERPARAMETERIZED
